! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module edetect_io_ml
  
  use errorHandling
  use param
  use dal

  implicit none

  private

  public cp_head, get_files, co_image, co_bgima, co_expima, co_mask, rawcoords, impix2equ
  
! MAX_IMA              max number of images
! MAX_INS              max number of different instruments 
! MAX_BANDS            max number of energy bands

  integer, parameter, public :: max_bands = 6
  integer, parameter, public :: max_ins = 105  !! 60 !! IT 2014/10
  integer, parameter, public :: max_ima = 525  !!240 !! IT 2014/10 (max_ins*bands) vs gfortran limit (2GB arrays)

      integer, parameter,public :: max_fit = 10        
      integer, parameter,public :: max_par = max_fit*max_ima+max_fit*3

! NUM_INSTR            number of instruments
! NUM_IMAGE(MAX_INS)   number of images per instrumet
! NIM                  total number of images

  integer, public :: num_instr, nim
  integer, dimension(max_ins), public :: num_image       
  integer, dimension(max_ima), public :: pimin, pimax

  integer, dimension(:), allocatable, public :: ixb0, iyb0, ixb1, iyb1, &
                                                xbsize, ybsize
  integer, public :: ixsize, iysize    !!! IT 2016/11
  integer, public :: imbuff                             !(JR)

  !!--JR   /   IT 2016/08 (moved)
  logical, public :: withimbuff

! IMAGE                input images
! BG_IMAGE             background images
! EXP_IMAGE            exposure images

  integer, dimension(:,:,:), allocatable, public  :: image
  real, dimension (:,:,:), allocatable, public :: exp_image, bg_image
  logical, dimension(:,:,:), allocatable, public :: mask_image
  logical, dimension(max_ima), public :: OotCorrDone  
  

  integer, dimension(2), public :: npix

! IP(MAX_INS,MAX_BANDS)pointer field to manage images

  integer, dimension(max_ins, max_bands), public :: ip
  integer, dimension(max_ins), public :: expip

  real(kind=single), dimension(max_ima), public :: ecf

  real(kind=single), public :: xcen1, ycen1
  real(kind=double), dimension(max_ins), public :: ra_pnt, dec_pnt, pa_pnt
  integer, dimension(max_ins), public :: expind
  real(kind=double), public :: crpix1,crpix2,crval1,crval2, &
                               crdelt1,crdelt2,crota2
  real(kind=single), dimension(3), public :: xoff, yoff
  character(len=30), public :: submode
  character(len=20), public :: datetime
  character(8) :: m1filt,m2filt,pnfilt
  character(30) :: m1sub,m2sub,pnsub
  real(kind=single) :: pnexpos,m1expos,m2expos

  real(kind=single), dimension(12,max_ima), public :: ontime


  real(kind=double), parameter :: pi=3.14159265358979323846d0
  real(kind=double), parameter :: pi180=pi/180d0, overpi180=180d0/pi

!!! obsolete  public ixb, iyb                       !(JR)

contains

!-----------------------------------------------------------------------
!---------- (JR) -------------------------------------------------------

!!!!!! IT 2016/12 - replaced
!!!  !---------------------------------------------------------------------
!!!  integer function ixb(ix,im)
!!!  !---------------------------------------------------------------------
!!!
!!!  integer :: ix, im
!!!
!!!  ixb = max(min(ix-ixb0(im),imbuff),1)
!!!
!!!  end function ixb
!!!
!!!  !---------------------------------------------------------------------
!!!  integer function iyb(iy,im)
!!!  !---------------------------------------------------------------------
!!!
!!!  integer :: iy, im
!!!
!!!  iyb = max(min(iy-iyb0(im),imbuff),1)
!!!
!!!  end function iyb

!-----------------------------------------------------------------------
!-----------------------------------------------------------------------



  !---------------------------------------------------------------------
  subroutine cp_attr (set1, set2, arr, attr)
  !---------------------------------------------------------------------
    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT), intent(IN) :: arr
    character(len=*), dimension(:), intent(IN) :: attr

    integer i

    do i=1,size(attr)
      if (hasAttribute(set1,attr(i))) then
        call setAttribute(set2,attribute(set1,attr(i)))
      else 
        if (hasAttribute(arr,attr(i))) then
          call setAttribute(set2,attribute(arr,attr(i)))
        end if
      end if
    end do

  end subroutine cp_attr

  !---------------------------------------------------------------------
  subroutine cp_head (set1, set2, content, xid_arr, mod)
  !---------------------------------------------------------------------

    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT) :: arr,arr2
    character(len=*), intent (IN) :: content
    integer, dimension(:), intent(in) :: xid_arr
    integer, intent(IN) :: mod
    character(len=2), dimension(3) :: strInst
    character(len=8)  :: strEmin,strEmax 
    character(len=1)  :: strBand 
    character(len=80) :: strComment
    integer i,j,j0,jj,ipoint
    integer :: num_instr0, n_instr
    integer, dimension(max_ins) :: num_image0
    logical :: xidflg    

    data strInst /"PN","M1","M2"/

    arr = array (set1,"PRIMARY",read)
    
    !!! IT 2014/05
    !!! copy full dataset header and array header of images
    if (mod <= 2) then
       call setAttributes( attributable(set2), attributable(set1) )
       arr2 = array(set2,0)
       call setAttributes( attributable(arr2), attributable(arr) )
       call release(arr2)

!!!    if (mod >= 1 .and. mod <=4 ) then
!!!      call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
!!!    end if
!!!
!!!    if (mod == 2) then
!!!      call cp_attr(set1,set2,arr,(/"BAND"/) )
!!!    end if
!!!
!!!    !if (mod == 3) then
!!!    !  call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
!!!    !end if
!!!    
!!!    if (mod < 3) then  
!!!      call cp_attr(set1,set2,arr, &
!!!      (/ "TELESCOP","DATAMODE","FILTER  ","OBS_MODE","OBS_ID  ","EXP_ID  ", &
!!!         "EXPIDSTR","ORIGIN  ","OBJECT  ","RA_PNT  ","DEC_PNT ","PA_PNT  ", &
!!!         "CRPIX1  ","CRPIX2  ","CRVAL1  ","CRVAL2  ","CDELT1  ", &
!!!         "CDELT2  ","CTYPE1  ","CTYPE2  ","SUBMODE ","RADECSYS","EQUINOX ", &
!!!         "EXPOSURE","DATE-OBS","DATE-END","OBSERVER","ATT_SRC "/) )
    else
       if (mod <= 4) call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
       call cp_attr(set1,set2,arr, &
         (/ "TELESCOP","INSTRUME","DETNAME ","DATAMODE","FILTER  ","OBS_MODE", &
            "OBS_ID  ","EXP_ID  ","EXPIDSTR","REVOLUT ","ORIGIN  ","OBJECT  ", &
            "RA_PNT  ","DEC_PNT ","PA_PNT  ","RADECSYS","EQUINOX ","EXPOSURE", &
            "DATE-OBS","DATE-END","OBSERVER","ATT_SRC ","SUBMODE " /) )
       !for tables, rename WCS keywords:
       call cp_wcs2table(set1,set2,"real32",arr, &
          (/ "CRPIX1  ","CRPIX2  ","CRVAL1  ","CRVAL2  ","CDELT1  ", &
             "CDELT2  "/) )
       call cp_wcs2table(set1,set2,"string",arr, (/ "CTYPE1  ","CTYPE2  ", &
            "CUNIT1  ","CUNIT2  "/) )             !!! IT 2014/05: CUNIT1,2 added

       num_instr0=0
       num_image0=0
       do j=1,num_instr
          j0=expip(j)
          num_instr0=max(j0,num_instr0) 
          num_image0(j0)=num_image(j)
       end do
       n_instr=0 
       do j0=1,num_instr0
          if (num_image0(j0)>0) n_instr=n_instr+1
       end do
  
       call setAttribute(set2,"N_INSTR",n_instr," ","Number of epic cameras")
       call setAttribute (set2,"PN_BNDS",num_image0(1), " ", &
                                               "Number of PN energy bands")
       call setAttribute (set2,"M1_BNDS",num_image0(2), " ", &
                                               "Number of MOS1 energy bands")
       call setAttribute (set2,"M2_BNDS",num_image0(3), " ", &
                                               "Number of MOS2 energy bands")

       if (mod == 4) then
          xidflg=.false.
          do j=1,num_instr
             if (xid_arr(j) > 0) then 
                xidflg=.true.
                exit
             end if
          end do
          if (xidflg) then
             call setAttribute (set2,"XID_BND",1," ", &
                                         "1 if XID band output is present")
          else
             call setAttribute (set2,"XID_BND",0," ", &
                                         "1 if XID band output is present")
          end if
       end if


       call setAttribute(set2,"FILT_PN",pnfilt,"EPN FILTER ID")
       call setAttribute(set2,"FILT_M1",m1filt,"EMOS1 FILTER ID")
       call setAttribute(set2,"FILT_M2",m2filt,"EMOS2 FILTER ID")
       call setAttribute(set2,"SUBM_PN",pnsub,"EPN SUBMODE")
       call setAttribute(set2,"SUBM_M1",m1sub,"EMOS1 SUBMODE")
       call setAttribute(set2,"SUBM_M2",m2sub,"EMOS2 SUBMODE")
       call setAttribute(set2,"EXPOS_PN",pnexpos,"EPN EXPOSURE")
       call setAttribute(set2,"EXPOS_M1",m1expos,"EMOS1 EXPOSURE")
       call setAttribute(set2,"EXPOS_M2",m2expos,"EMOS2 EXPOSURE")


       do i=1,num_instr
          do j=1,num_image(i)
             jj=j
             if (mod == 4) then
                if (xid_arr(i) == j) jj=9
             end if
             ipoint=ip(i,j)
             if (ipoint > 0) then
                write(strBand,'(I1)') jj 
                strEmin = strInst(expip(i))//"_"//strBand//"_ELO"
                strEmax = strInst(expip(i))//"_"//strBand//"_EHI"
                strComment = "E_min for Instr "//strInst(expip(i))//" , band "//strBand   
                call setAttribute(set2,strEmin,pimin(ipoint),"PI chan",strComment)
                strComment = "E_max for Instr "//strInst(expip(i))//" , band "//strBand   
                call setAttribute(set2,strEmax,pimax(ipoint),"PI chan",strComment)
             endif
          end do
       end do

      

    end if   !! end of "if (mod <= 2) then ... else"

    call setAttribute(set2,"CONTENT", content, "Content of file")

    call release(arr)

  end subroutine cp_head

  !---------------------------------------------------------------------
  subroutine cp_wcs2table (set1, set2, dtype, arr, attr)
  !---------------------------------------------------------------------
    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT), intent(IN) :: arr
    character(len=*), dimension(:), intent(IN) :: attr
    character(len=*), intent(IN) :: dtype
    character(len=8) attname
    real value
    character(len=80) svalue,comment 
    integer i

    do i=1,size(attr)
      comment="copy of WCS keyword "//attr(i)//" from input image" 
      if (dtype == "real32") then 
        if (hasAttribute(set1,attr(i))) then
          value=real32Attribute(set1,attr(i))
          attname = "T"//attr(i)
          call setAttribute(set2,attname,value,"",trim(comment))
        else 
          if (hasAttribute(arr,attr(i))) then
            value=real32Attribute(arr,attr(i))
            attname = "T"//attr(i)
            call setAttribute(set2,attname,value,"",trim(comment))
          end if
        end if
      elseif (dtype == "string") then
        if (hasAttribute(set1,attr(i))) then
          svalue=stringAttribute(set1,attr(i))
          attname = "T"//attr(i)
          call setAttribute(set2,attname,trim(svalue),trim(comment))
        else 
          if (hasAttribute(arr,attr(i))) then
            svalue=stringAttribute(arr,attr(i))
            attname = "T"//attr(i)
            call setAttribute(set2,attname,trim(svalue),trim(comment))
          end if
        end if 
      end if 
    end do

  end subroutine cp_wcs2table

  !--------------------------------------------------------------------
  subroutine get_files (param_name, imag, instrume, band, exp_id, first)
  !---------------------------------------------------------------------
 
    character(len=*), intent(IN) :: param_name
    logical, intent(IN) :: first
    character(len=1024), dimension(*), intent(OUT) :: imag
    character(len=21), dimension(*), intent(OUT) :: instrume
    character(len=5) :: instru
    integer, dimension(*), intent(OUT) :: band
!!! IT 2014/05
    character(len=13), dimension(*), intent(OUT) :: exp_id        
    character(len=13), dimension(max_ima) :: exp_idtmp
    character(len=13) :: expidtmp
!!!
    character(len=80) errMess
    character(len=8) :: filter

    integer :: i, j, ib   !!!, ban
    integer :: nimage, ipoint
!    character(len=7) :: teles, instru, expid  
!    character(len=16),dimension(max_ima) :: ins
    character(len=3) :: teles
    character(len=21),dimension(max_ima) :: ins
    logical, dimension(max_ima) :: same_inst

    type(DataSetT) :: set
    type(ArrayT) :: arr

    nimage = parameterCount(param_name) 
    if (nimage == 0) then
      call error("MissingParameter", "Missing input file name")
    end if

 
    if (first) then
      pnfilt="UNDEF"
      m1filt="UNDEF"    
      m2filt="UNDEF" 
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)

        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if


        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if


        if (hasAttribute(set,"FILTER")) then
          filter = stringAttribute(set,"FILTER")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"FILTER")) then
            filter = stringAttribute(arr,"FILTER")
          else
            call warning("MissingAttribute","Keyword FILTER is missing")
          end if
          call release(arr)
        end if


        if (index(instru,"PN") > 0 .or. &
            index(instru,"pn") > 0) then 
          pnfilt=filter
        else if (index(instru,"MOS1") > 0 .or. &
                 index(instru,"mos1") > 0) then
          m1filt=filter
        else if (index(instru,"MOS2") > 0 .or. &
                 index(instru,"mos2") > 0) then
          m2filt=filter
        end if
 
        


        if (hasAttribute(set,"EXP_ID")) then
          expidtmp = stringAttribute(set,"EXP_ID")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"EXP_ID")) then
            expidtmp = stringAttribute(arr,"EXP_ID")
          else
            expidtmp = "---"
            call warning("MissingAttribute","Keyword EXP_ID is missing")
          end if
          call release(arr)
        end if
        instrume(i) = trim(teles) // trim(instru) // trim(expidtmp)
        exp_idtmp(i) = trim(expidtmp)                             !!! IT 2014/05

        if (hasAttribute(set,"BAND")) then
          band(i) = int32Attribute(set, "BAND")
        else
          band(i) = 7
        end if
        call release(set)
      end do

      num_instr = 0
      same_inst = .false.    
      ip = -1
      do i=1,nimage      
        if (.not. same_inst(i)) then
           num_instr = num_instr+1
!!! IT 2016/05 Throw error, if number of images exceeds max_ima.
           if (num_instr.gt.max_ins) then
              write(errstr,*) &
                   "Up to ",max_ins," instruments*pointings are allowed."
              call error("TooManyInstruments", errstr)
           end if
           exp_id(num_instr) = exp_idtmp(i)                       !!! IT 2014/05
        end if
        ib = 1 
        do while (ip(num_instr,ib) /= -1)
          ib = ib+1
          if (ib.gt.max_bands) call error("TooManyBands", &       !!! IT 2016/04
               "Up to six energy bands are allowed per instrument.")
        end do
        ip(num_instr,ib) = i
        num_image(num_instr) = ib
        do j=i+1,nimage
          if (instrume(i) == instrume(j)) same_inst(j) = .true.
       end do
      end do

      do i=1,num_instr
        ipoint=ip(i,1) 
        if (index(instrume(ipoint),"PN") > 0 .or. &
            index(instrume(ipoint),"pn") > 0) then 
          expip(i) = 1
        else if (index(instrume(ipoint),"MOS1") > 0 .or. &
                 index(instrume(ipoint),"mos1") > 0) then
          expip(i) = 2
        else if (index(instrume(ipoint),"MOS2") > 0 .or. &
                 index(instrume(ipoint),"mos2") > 0) then
          expip(i) = 3
        else
          expip(i) = 4
          call error("WrongInst","Unknown instrument")
        end if
      end do
        
    else                     ! end "first"
      
      if (param_name(1:3) == "det") then
        if (nimage /= num_instr) then
           call error("FileMismatch", &
             "# detector masks /= # instruments")
        end if
      else 
        if (nimage /= nim) then
          call error("FileMismatch","Inconsistent number of input images")
        end if
      end if
 
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)

        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"EXP_ID")) then
          expidtmp = stringAttribute(set,"EXP_ID")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"EXP_ID")) then
            expidtmp = stringAttribute(arr,"EXP_ID")
          else
            expidtmp = "---"
            call warning("MissingAttribute","Keyword EXP_ID is missing")
          end if
          call release(arr)
        end if

        ins(i) = trim(teles) // trim(instru) // trim(expidtmp)


        call release(set)

      end do

      if (param_name(1:3) == "det") then 
         do i=1,num_instr
           do j=1,num_image(i)
             if (ins(i) /= instrume(ip(i,j))) then
               errMess = ": Wrong detector mask" 
               call error ("FileMismatch",errMess) 
             end if
           end do
         end do
      end if 

    end if

    nim = ip(num_instr,num_image(num_instr))

    

  end subroutine get_files 

  !---------------------------------------------------------------------
  subroutine co_image (fname, start, step, npix, exposure, xcen, ycen, &
                       ra_point, dec_point, pa_point, datetime, &
                       crpix1, crpix2, crval1, crval2, &
                       crdelt1,crdelt2,crota2, &
                       xoff,yoff,inst,ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix
    real(kind=double), dimension(2), intent(out) :: start, step 
    real(kind=double), intent(inout) :: ra_point, dec_point, pa_point, &
                       crpix1, crpix2, crval1, crval2, &
                       crdelt1, crdelt2, crota2 
    character(len=20), intent(out) :: datetime
    character(len=8) :: keyname
    character(len=2) :: istring
    real(kind=double) ::cos_pa, sin_pa, rr, dr, r0r, d0r, &
                       dra, ddec, xr, yr, dist1, dist2, del, &
                       crpix1old,crpix2old,crval1old,crval2old, &
                       crdelt1old, crdelt2old
 
    real(kind=single) :: exposure, xcen, ycen, r4huge !!!, ran
    real(kind=double) :: r8huge
!!!    integer, dimension(3) :: ixoff, iyoff
    integer :: iix, iiy, atype, ichip  !!! , itot, ix, iy, i, nev
    integer :: ix0, ix1, iy0, iy1                           !(JR)
    integer, dimension(3) :: nchips=(/12,7,7/)
    integer, dimension(2), save :: npix_save
    logical, save :: first = .true.

!!!    logical :: empty                                        !(JR)

!!!    integer, save :: iseed

    integer(kind=int32), dimension(:,:), pointer :: i4arr
    integer(kind=int16), dimension(:,:), pointer :: i2arr
    integer(kind=int8), dimension(:,:), pointer :: i1arr
    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=double), dimension(:,:), pointer :: r8arr
    integer, dimension(:), pointer :: shape

    integer(kind=int32), dimension(:), allocatable :: ipx, ipy  !(JR)
    character(len=10) :: cdummy   !!! IT 2016/12

    type(datasett) :: set
    type(arrayt) :: arr

    integer :: astat

    nullify (i1arr, i2arr, i4arr, r4arr, r8arr, shape)    !!! IT 2016

    !!! IT 2016/08: more concise version of JR's withimbuff and imbuff checks,
    !!! to be run only if "first" (below)

    r4huge=huge(1.0)
    r8huge=huge(1d0)

!    del = 1d-5
     del = 1d-3


    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)

    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    if (first) then
       
       !!-- JR ------   /   !!! IT 2016/08
       withimbuff = booleanParameter("withimagebuffersize")
       if (withimbuff) then
          imbuff = intParameter("imagebuffersize")
       else
          imbuff = max(npix(1),npix(2))
       endif
       !!------------
   
      pnsub='UNDEFINED'
      m1sub='UNDEFINED'
      m2sub='UNDEFINED'
      pnexpos=0.0
      m1expos=0.0
      m2expos=0.0
      npix_save=npix
      ixsize = min(npix(1),imbuff)   !!! IT 2016/12
      iysize = min(npix(2),imbuff)
      allocate (image(ixsize,iysize,nim), stat=astat)
      if (astat /= 0) call error("notEnoughMemory", &
           "Not enough memory for array 'image'")
      image(:,:,:) = 0
      allocate (ixb0(nim))
      allocate (ixb1(nim))
      allocate (iyb0(nim))
      allocate (iyb1(nim))
      allocate (xbsize(nim))
      allocate (ybsize(nim))
      first=.false.
    end if

    allocate (ipx(0:npix(1)-1))                         !(JR)
    allocate (ipy(0:npix(2)-1))                         !(JR)

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

!!!    ixoff(1:3)=nint(xoff(1:3))
!!!    iyoff(1:3)=nint(yoff(1:3))


    !!! IT 2016/12 - new code to determine the coordinates of the subimage;
    !!! partial re-write of the redundant code parts

!!!obsolete    itot=0
    ! image type
    atype = arrayDataType(arr)
    ! get image and total count numbers in x and y direction
    select case (atype)
       case (Integer8)
          i1arr => int8array2data(arr)
          ipx=sum(i1arr,2)
          ipy=sum(i1arr,1)
       case (Integer16)
          i2arr => int16array2data(arr)
          ipx=sum(i2arr,2)
          ipy=sum(i2arr,1)
       case (Integer32)
          i4arr => int32array2data(arr)
          ipx=sum(i4arr,2)
          ipy=sum(i4arr,1)
       case (Real32)
          r4arr => Real32array2data(arr)
          ipx=sum(r4arr,2)
          ipy=sum(r4arr,1)
       case (Real64)
          r8arr => Real64array2data(arr)
          ipx=sum(r8arr,2)
          ipy=sum(r8arr,1)
       case default
          call error("WrongType","Input image has wrong type")
    end select

    ! subimage coordinates - start values
    ix0=0
    iy0=0
    ix1=1
    iy1=1

    ! determine coordinates of the subimage:
    ! - look for the first column with a non-zero entry (i.e.: along x),
    !   bottom-up
    do iix=0,npix(1)-1
       if (ipx(iix)>0) then
          ix0=iix
          exit
       end if
    end do
    ! - look for the first row    with a non-zero entry (i.e.: along y),
    !   bottom-up
    do iiy=0,npix(2)-1
       if (ipy(iiy)>0) then
          iy0=iiy
          exit
       end if
    end do
    ! - look for the last  column with a non-zero entry (i.e.: along x),
    !   top-down
    do iix=npix(1)-1,0,-1
       if (ipx(iix)>0) then
          ix1=iix
          exit
       end if
    end do
    ! - look for the last  row    with a non-zero entry (i.e.: along y),
    !   top-down
    do iiy=npix(2)-1,0,-1
       if (ipy(iiy)>0) then
          iy1=iiy
          exit
       end if
    end do

    ! Add some extra space and round to multiples of 16 by means of integer
    ! division in order to be more than sure that coordinates aren't shifted
    ! during the compression steps (maximum compression factor is 8, and we
    ! want consistent to have pixel edges for all subimages, irrespective of
    ! their offsets). eident offsets (xoff, yoff) are ignored here - eident
    ! was ceased in 2003.
    ! Our image indices within Fortran start from 1; those of the "raw" input
    ! images read via C routines start from 0.
    ixb0(ipoint) = max( (ix0-16)/16*16, 0 )
    iyb0(ipoint) = max( (iy0-16)/16*16, 0 )
    ixb1(ipoint) = min( (ix1+16)/16*16, npix(1)-1 )
    iyb1(ipoint) = min( (iy1+16)/16*16, npix(2)-1 )
    xbsize(ipoint) = ixb1(ipoint)-ixb0(ipoint)+1
    ybsize(ipoint) = iyb1(ipoint)-iyb0(ipoint)+1

    ! We'll run into trouble, if the subimage is larger than imagebuffersize
    if (xbsize(ipoint)>imbuff .or. ybsize(ipoint)>imbuff ) then
       write(cdummy, '(i10)') max(xbsize(ipoint), ybsize(ipoint))
       call error("SizeMismatch", &
            "Increase ImageBufferSize to at least "//trim(cdummy))
    end if

    ! Copy valid image ranges. Formerly, ixoff(inst) and iyoff(inst) were
    ! applied. Since the offsets were produced by eident, which was ceased in
    ! 2003, support for them is dropped (as of 2016).

    ! removed itot, which is obsolete - syntax was
    ! itot = sum(i1arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint)))
    ! etc

    image(:,:,ipoint) = 0.0

    select case (atype)

       case (Integer8)
          image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) &
               = i1arr(ixb0(ipoint):ixb1(ipoint), &
                       iyb0(ipoint):iyb1(ipoint))

       case (Integer16)
          image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) &
               = i2arr(ixb0(ipoint):ixb1(ipoint), &
                       iyb0(ipoint):iyb1(ipoint))

       case (Integer32)
          image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) &
               = i4arr(ixb0(ipoint):ixb1(ipoint), &
                       iyb0(ipoint):iyb1(ipoint))

       case (Real32)
          image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) &
               = nint(r4arr(ixb0(ipoint):ixb1(ipoint), &
                           iyb0(ipoint):iyb1(ipoint)))
          where(r4arr(ixb0(ipoint):ixb1(ipoint), &
                      iyb0(ipoint):iyb1(ipoint)) .gt. r4huge) 
             image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) = r4huge  
          end where

       case (Real64)
          image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) &
               = nint(r8arr(ixb0(ipoint):ixb1(ipoint), &
                           iyb0(ipoint):iyb1(ipoint)))
          where(r8arr(ixb0(ipoint):ixb1(ipoint), &
                      iyb0(ipoint):iyb1(ipoint)) .gt. r8huge) 
             image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) = r8huge  
          end where

       case default
          call error("WrongType","Input image has wrong type")

    end select

    !!!---- end re-write 2016/12



    start(1) = 1d0
    start(2) = 1d0
    step(1) = 1d0
    step(2) = 1d0

    if (hasAttribute(set,"EXPOSURE")) then
      exposure = real32Attribute(set,"EXPOSURE")
      if (inst == 1) pnexpos=exposure
      if (inst == 2) m1expos=exposure
      if (inst == 3) m2expos=exposure  
    else if (hasAttribute(arr,"EXPOSURE")) then
      exposure = real32Attribute(arr,"EXPOSURE")
      if (inst == 1) pnexpos=exposure
      if (inst == 2) m1expos=exposure
      if (inst == 3) m2expos=exposure           
    else
      exposure = 1.0
      call warning("MissingAttribute", &
                   "No exposure keyword present: set to 1.0")
    end if 

    if (hasAttribute(set,"SUBMODE")) then
      submode = stringAttribute(set,"SUBMODE")
    else if (hasAttribute(arr,"SUBMODE")) then
      submode = stringAttribute(arr,"SUBMODE")
    else 
      submode="PrimeFullWindow"
    end if 
    if (inst == 1) pnsub=submode
    if (inst == 2) m1sub=submode
    if (inst == 3) m2sub=submode     




 
    if (hasAttribute(set,"DATE-OBS")) then
      datetime = stringAttribute(set,"DATE-OBS")
    else if (hasAttribute(arr,"DATE-OBS")) then
      datetime = stringAttribute(arr,"DATE-OBS")
    else 
      datetime=" "
    end if 

    if (hasAttribute(set,"RA_PNT")) then
      ra_point = real64Attribute(set,"RA_PNT")
    else  
      if (hasAttribute(arr,"RA_PNT")) then
        ra_point = real64Attribute(arr,"RA_PNT")
      else
        ra_point = 0d0
        call warning("MissingAttribute","Keyword RA_PNT is missing")
      end if
    end if

    if (hasAttribute(set,"DEC_PNT")) then
      dec_point = real64Attribute(set,"DEC_PNT")
    else  
      if (hasAttribute(arr,"DEC_PNT")) then
        dec_point = real64Attribute(arr,"DEC_PNT")
      else
        dec_point = 0d0
        call warning("MissingAttribute","Keyword DEC_PNT is missing")
      end if
    end if

    if (hasAttribute(set,"PA_PNT")) then
      pa_point = real64Attribute(set,"PA_PNT")
    else  
      if (hasAttribute(arr,"PA_PNT")) then
        pa_point = real64Attribute(arr,"PA_PNT")
      else
        pa_point = 0d0
        call warning("MissingAttribute","Keyword PA_PNT is missing")
      end if
    end if

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if
 
    if (hasAttribute(arr,"CROTA2")) then
       crota2 = real64Attribute(arr,"CROTA2")
    else
      if (hasAttribute(set,"CROTA2")) then
        crota2 = real64Attribute(set,"CROTA2")
      else
        crota2 = 0d0
        !call warning("MissingAttribute","Keyword CROTA2 is missing")
      end if
    end if

    if (crdelt1 > 0.0 .and. crdelt2 > 0.0) then
      dist1=0.36d0/abs(crdelt1)
      dist2=0.36d0/abs(crdelt2)
    else
      dist1=320.0d0
      dist2=320.0d0
    end if
    if (crpix1 < -dist1 .or. crpix1 > npix(1)+dist2 .or. &
        crpix2 < -dist2 .or. crpix2 > npix(2)+dist2) then
      call error("WrongRefPix","Reference pixel is outside of FOV")
    end if

    cos_pa = dcos(crota2 * pi180)
    sin_pa = dsin(crota2 * pi180)
    rr = ra_point * pi180
    dr = dec_point * pi180  
    r0r = crval1 * pi180
    d0r = crval2 * pi180
    call spher2(rr, dr, r0r, d0r, dra, ddec)
    xr =   dra * cos_pa + ddec * sin_pa ! Vorzeichen ??
    yr = - dra * sin_pa + ddec * cos_pa ! 
    xcen = real( xr * overpi180 / crdelt1 + crpix1 )
    ycen = real( yr * overpi180 / crdelt2 + crpix2 )

    if (xcen < 1.0 .or. xcen > npix(1) .or. &
        ycen < 1.0 .or. ycen > npix(2)) then
      xcen=min(max(xcen,1.0),real(npix(1)))
      ycen=min(max(ycen,1.0),real(npix(2)))
      call warning("WrongPointDir","Pointing direction is outside of image")
    end if

! read ontime values from image header:
    
    do ichip=1,nchips(inst)
      write(istring,'(I2.2)') ichip
      keyname="ONTIME"//istring
      if (hasAttribute(arr,keyname)) then
        ontime(ichip,ipoint) = real64Attribute(arr,keyname)  
      else
        if (hasAttribute(set,keyname)) then
          ontime(ichip,ipoint) = real64Attribute(set,keyname) 
        endif          
      endif  
  
    enddo
    

    call release(arr)
    call release(set)

  end subroutine co_image

  !---------------------------------------------------------------------
  subroutine co_bgima (fname, npix, &
                       crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                       xoff, yoff, inst, ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix 
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 
    integer, dimension(2), save :: npix_save
!!!    integer, dimension(3) :: ixoff, iyoff
!!!    integer :: ix, iy, iix, iiy   !!!, i, l
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=single) :: imax  !!!, bmax   !!!  tot  -  2018/04
    integer, dimension(:), pointer :: shape
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del

    type(datasett) :: set
    type(arrayt) :: arr

    integer :: astat

!!! IT 2016/08 - redundant
!!!    !!--JR
!!!    logical :: withimbuff
!!!    !!-- JR ------
!!!    withimbuff = booleanParameter("withimagebuffersize")
!!!    !!------------
!!!    !!-- JR ------
!!!    if (withimbuff) then
!!!      imbuff = intParameter("imagebuffersize")
!!!      withimbuff = .true.
!!!      else
!!!      imbuff = 640
!!!      withimbuff = .false.
!!!    endif
!!!    !!------------

    nullify (r4arr, shape)  !!! IT 2016

    del = 1d-5

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)
    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

!!! IT 2016/08 - redundant
!!!    !! -- JR -----------------------------
!!!!!!    if (withimbuff) then
!!!!!!      imbuff = imbuff
!!!!!!      else
!!!    if (.not. withimbuff) then
!!!      imbuff = max(npix(1),npix(2))
!!!    end if
!!!    !! -----------------------------------

    if (first) then
      npix_save=npix
      allocate (bg_image(ixsize,iysize,nim), stat=astat)
      if (astat /= 0) call error("notEnoughMemory", &
           "Not enough memory for array 'bg_image'")
      bg_image(:,:,:) = 0.0
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

!!!    ixoff(1:3)=nint(xoff(1:3))
!!!    iyoff(1:3)=nint(yoff(1:3))

    bg_image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) &   !!! IT 2016/12
            = r4arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint))
    imax = maxval(r4arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint)))
!!! IT 2016/12 --- obsolete, since the code below has been deactivated ...
!!!    bmax = maxval(r4arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint)))

    !!! update 2018/04: maxval (empty pixels are set to 1e-5 by esplinemap)
    if (imax<=2e-5) call error("EmptyBgMap","Background map is empty") 

     !do iy = 1,npix(2)
     !  do ix = 1,npix(1)
     !    if (bg_image(ix,iy,ipoint)<2e-5) bg_image(ix,iy,ipoint)=bmax
     !  end do
     !end do

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if

    call release(arr)
    call release(set)
  
  end subroutine co_bgima

  !---------------------------------------------------------------------
  subroutine co_expima (fname, npix, &
                        crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                        xoff, yoff, inst, ipoint, &
                        xcen, ycen, exp_center, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim
    real(kind=single), intent(inout) :: xcen, ycen, exp_center
    real(kind=single), dimension(3), intent(in) :: xoff, yoff   
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 
    integer, dimension(2), intent(out) :: npix 

    integer, dimension(2), save :: npix_save
!!!    integer, dimension(3) :: ixoff, iyoff
!!!    integer :: ix, iy, iix, iiy   !!!, i
    real(kind=single) :: expmax   !!!  tot  -  2018/04
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    integer, dimension(:), pointer :: shape
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del

    type(datasett) :: set
    type(arrayt) :: arr

    integer :: astat

!!! IT 2016/08 - redundant
!!!    !!--JR
!!!    logical :: withimbuff
!!!    !!-- JR ------
!!!    withimbuff = booleanParameter("withimagebuffersize")
!!!    !!------------
!!!    !!-- JR ------
!!!    if (withimbuff) then
!!!      imbuff = intParameter("imagebuffersize")
!!!      withimbuff = .true.
!!!      else
!!!      imbuff = 640
!!!      withimbuff = .false.
!!!    endif
!!!    !!------------

    nullify (r4arr, shape)  !!! IT 2016

    del = 1d-5

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)

    if (hasAttribute(set,"XCEN")) then
      xcen = real32Attribute(set,"XCEN")
    else if (hasAttribute(arr,"XCEN")) then
      xcen = real32Attribute(arr,"XCEN")
    end if
    if (hasAttribute(set,"YCEN")) then
      ycen = real32Attribute(set,"YCEN")
    else if (hasAttribute(arr,"YCEN")) then
      ycen = real32Attribute(arr,"YCEN")
    end if
    exp_center=0.
    if (hasAttribute(set,"EXPOSURE")) then
      exp_center = real32Attribute(set,"EXPOSURE")
    else if (hasAttribute(arr,"EXPOSURE")) then
      exp_center = real32Attribute(arr,"EXPOSURE")
    end if 

    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

!!! IT 2016/08 - redundant
!!!    !! -- JR -----------------------------
!!!!!!    if (withimbuff) then
!!!!!!      imbuff = imbuff
!!!!!!      else
!!!    if (.not. withimbuff) then
!!!      imbuff = max(npix(1),npix(2))
!!!    end if
!!!    !! -----------------------------------

    if (first) then
!     print *,'allocate exp image'
      npix_save=npix
      allocate (exp_image(ixsize,iysize,nim), stat=astat)
      if (astat /= 0) call error("notEnoughMemory", &
           "Not enough memory for array 'exp_image'")
      exp_image(:,:,:) = 0.0
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

!!!    ixoff(1:3)=nint(xoff(1:3))
!!!    iyoff(1:3)=nint(yoff(1:3))

    exp_image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint) &   !!! IT 2016/12
            = r4arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint))
!!!    tot = sum(r4arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint)))
    expmax = maxval(r4arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint)))
      
    if (expmax<=1e-5) call error("EmptyExpMap","Exposure map is empty")
    if (exp_center<=0.) exp_center=expmax

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if
    OotCorrDone(ipoint)=.false.
    if (hasAttribute(arr,"OOTCORR")) then
       OotCorrDone(ipoint)=booleanAttribute(arr,"OOTCORR")
    else
      if (hasAttribute(set,"OOTCORR")) then
        OotCorrDone(ipoint)=booleanAttribute(set,"OOTCORR")
      end if
    end if     


    call release(arr)
    call release(set)

  end subroutine co_expima

  !---------------------------------------------------------------------
  subroutine co_mask (fname, npix, &
                      crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                      xoff, yoff, inst, ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 
    integer, dimension(2), intent(out) :: npix

    integer, dimension(2), save :: npix_save
!!!    integer, dimension(3) :: ixoff, iyoff
    integer :: ix, iy, iix, iiy, atype  !!!  i, itot, 
    !!! 2018/04 replace itot by logical "testempty"
    logical, save :: first = .true., testempty = .true.

    integer, dimension(:,:), pointer :: i4arr
    integer, dimension(:), pointer :: shape
    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del

    type(datasett) :: set
    type(arrayt) :: arr

    integer :: astat

!!! IT 2016/08 - redundant
!!!    !!--JR
!!!    logical :: withimbuff
!!!    !!-- JR ------
!!!    withimbuff = booleanParameter("withimagebuffersize")
!!!    !!------------
!!!    !!-- JR ------
!!!    if (withimbuff) then
!!!      imbuff = intParameter("imagebuffersize")
!!!      withimbuff = .true.
!!!      else
!!!      imbuff = 640
!!!      withimbuff = .false.
!!!    endif
!!!    !!------------

    nullify (i4arr, r4arr, shape)  !!! IT 2016

    del = 1d-5

    set = dataset (fname, read)

    !! The mask extension should be by default in an extension called
    !! MASK, if not then look in the PRIMARY array
    if ( hasBlock( set, "MASK" ) ) then
        arr = array(set, "MASK", read)
    else
        arr = array(set, "PRIMARY", read)
    endif

    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

!!! IT 2016/08 - redundant
!!!    !! -- JR -----------------------------
!!!    if (withimbuff) then
!!!      imbuff = imbuff
!!!      else
!!!      imbuff = max(npix(1),npix(2))
!!!    end if
!!!    !! -----------------------------------

    if (first) then
!     print *,'allocate mask image:', nim,npix
      npix_save=npix
      allocate (mask_image(ixsize,iysize,nim), stat=astat)
      if (astat /= 0) call error("notEnoughMemory", &
           "Not enough memory for array 'mask_image'")
      mask_image(:,:,:) = .false.
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

!!!    ixoff(1:3)=nint(xoff(1:3))
!!!    iyoff(1:3)=nint(yoff(1:3))

!!!    itot=0
    atype = arrayDataType(arr)

    if (atype == Integer32) then
     i4arr => int32array2data(arr)
    iiy = iyb0(ipoint) ! coordinates of subimage within large image  !!! IT 2016
    do iy = 1,ybsize(ipoint)    ! subimage coordinates
       iix = ixb0(ipoint)
       do ix = 1,xbsize(ipoint)
          if (i4arr(iix, iiy) == 0 ) then
             mask_image(ix, iy, ipoint) = .false.
          else
             mask_image(ix, iy, ipoint) = .true.
             !!! 2017/12: moved from above - and finally removed 2018/04
!!!             itot=itot+1
             testempty = .false.
          end if
          iix = iix+1
       end do  ! ix
       iiy = iiy+1
    end do  ! iy
    end if

    if (atype == Real32) then
     r4arr => Real32array2data(arr)
    iiy = iyb0(ipoint) ! coordinates of subimage within large image  !!! IT 2016
    do iy = 1,ybsize(ipoint)    ! subimage coordinates
       iix = ixb0(ipoint)
       do ix = 1,xbsize(ipoint)
          if (r4arr(iix, iiy) < 1e-6 ) then
             mask_image(ix, iy, ipoint) = .false.
          else
             mask_image(ix, iy, ipoint) = .true.
             !!! 2017/12: moved from above - and finally removed 2018/04
!!!             itot=itot+1
             testempty = .false.
          end if
          iix = iix+1
       end do  ! ix
       iiy = iiy+1
    end do  ! iy
    end if

    if (testempty) call error("EmptyMask","Detection mask is empty") 

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        crval1 = 0d0
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if

    call release(arr)
    call release(set)

  end subroutine co_mask

!---------------------------------------------------------------------------------------

    subroutine impix2equ(xima,yima,ra,dec)

    real, intent(in) :: xima,yima
    real(kind=double), intent(out) :: ra,dec
    real(kind=double) :: xr,yr,dra,ddec,r0r,d0r,rr,dr,sin_pa,cos_pa
    real(kind=double), parameter :: pi=3.14159265358979323846d0
    real(kind=double), parameter :: pi180=pi/180d0, overpi180=180d0/pi

        cos_pa = dcos(crota2 * pi180)
        sin_pa = dsin(crota2 * pi180)
        xr = (xima-crpix1) * crdelt1 * pi180
        yr = (yima-crpix2) * crdelt2 * pi180
        dra =  xr * cos_pa - yr * sin_pa
        ddec = xr * sin_pa + yr * cos_pa
        r0r = crval1 * pi180
        d0r = crval2 * pi180
        call antisp(dra,ddec,r0r,d0r,rr,dr)
        ra = rr * overpi180
        dec = dr *overpi180

    
    end subroutine impix2equ


  !----------------------------------------------------------------------
  subroutine rawcoords(ra,dec,nsrc,datetime, &
                       chipix,chipiy,ccdinr,offax,azim)
  !----------------------------------------------------------------------

    use cal

    real(kind=double), dimension(max_fit), intent(in) :: ra, dec
    character(len=20), intent(in) :: datetime
    integer, intent(in) :: nsrc 
    integer, dimension(max_ins,max_fit), intent(out) :: & 
             chipix, chipiy, ccdinr
    
!!!    real(kind=single), dimension(:), pointer :: theta, phi, &
    real(kind=double), dimension(:), pointer :: theta, phi, &
                       cc2x, cc2y, cc2z, cc1x, cc1y, cc1z      
    real(kind=double) :: timestamp
    integer(int16), dimension(:), pointer :: chipX, chipY
    integer(int8), dimension(:), pointer :: ccdNr  
    integer :: i, j, j0

    type(SpacecraftAttitudeType) :: myattitude  
    type (BoreSightT) :: boresight

    type(EquatorialDirectionType), allocatable :: srcDirs(:)
    real(kind=double), dimension(max_ins,max_fit), intent(out) :: offax,azim 


    allocate(srcDirs(nsrc))

    nullify(chipX,chipY)
    nullify(theta,phi)
    nullify(cc1x,cc1y,cc1z)
    nullify(cc2x,cc2y,cc2z)
    nullify(ccdNr)

    srcDirs(1:nsrc)%ra = ra(1:nsrc)*pi180
    srcDirs(1:nsrc)%dec = dec(1:nsrc)*pi180

    timeStamp = toTimeTag(datetime) ! in caloalutils/caloaldefs.f90        
!    boresight = CAL_getBoresight(TimeStamp)

    do j=1,num_instr
      myattitude%ra=ra_pnt(j)*pi180
      myattitude%dec=dec_pnt(j)*pi180
      myattitude%apos=pa_pnt(j)*pi180       
      j0=expip(j)
      if (j0==1) call CAL_setState(EPN)
      if (j0==2) call CAL_setState(EMOS1)
      if (j0==3) call CAL_setState(EMOS2) 
      boresight = CAL_getBoresight(TimeStamp)
      call CAL_withSacCoordOrientation(boresight, myAttitude)
      call CAL_projectOntoTelCoord(boresight, srcDirs, theta, phi)
      call CAL_telCoordToCamCoord2(theta, phi, cc2x, cc2y, cc2z)
      call CAL_camCoord2ToCamCoord1(cc2x, cc2y, cc2z, cc1x, cc1y, cc1z)
      call CAL_camCoord1ToChipCoord(cc1x, cc1y, cc1z, chipX, chipY, ccdNr)
!!!      do i=1,nsrc
        chipix(j,1:nsrc)=chipX(0:nsrc-1)
        chipiy(j,1:nsrc)=chipY(0:nsrc-1)
        ccdinr(j,1:nsrc)=ccdNr(0:nsrc-1)
        offax(j,1:nsrc)=theta(0:nsrc-1)
        azim(j,1:nsrc)=phi(0:nsrc-1)
!--------------------------------------------------------------
        !write(*,*)ra(i),dec(i),azim(j,i),offax(j,i)
!--------------------------------------------------------------
!!!      end do
    end do
     
    call CAL_releaseMemory(chipX,chipY)
    call CAL_releaseMemory(theta,phi)
    call CAL_releaseMemory(cc1x,cc1y,cc1z)
    call CAL_releaseMemory(cc2x,cc2y,cc2z)
    call CAL_releaseMemory(ccdNr)
    deallocate(srcdirs)
 
  end subroutine rawcoords




  end module edetect_io_ml






