! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module eml_io

  use param
  use dal
  use errorHandling
  use emldata
  use edetect_io_ml
  use edetect_gen
  use emutils_mod
!  use random_aux

  implicit none

  private

  public get_sources, co_image, co_bgima, co_expima, &
         mulmlpari, wr_stab, wr_srcmap, mlout
!!! get_smaps, get_event,  !!! photon mode (not fully implemented)

  !character(len=16), dimension(max_ima) :: instrume 

contains

  !-------------------------------------------------------------------------------
  subroutine get_sources (nsou, xx, yy, sn, src, extent, in, totrate, totcounts)
  !-------------------------------------------------------------------------------

    real, dimension(max_sou), intent(out) :: xx, yy, sn, extent
    real, dimension(max_sou), intent(out) :: totrate,totcounts 
    real, dimension(max_ima,max_sou), intent(out) :: src
    integer, dimension(max_sou), intent(out) :: in
    integer, intent(out) :: nsou

    integer  :: l1,nrow,band,isou,ipxid,j
    integer  :: ipoint   !!! IT (obsolete)   ,snrmin,ierr
    real     :: xms,yms,snr,scts,ext_in   !!! ,ran 
    real, dimension(max_ins) :: srcsum

    integer(kind=int32) :: i, is, ins, prio, iband   !!! , iseed
    logical :: nosrc, radecpos

    real(kind=single), dimension(:), pointer, save :: rp4, rp5, rp6, rp8, rp9,rp10
    real(kind=double), dimension(:), pointer, save :: dp1, dp2
    integer(kind=int32), dimension(:), pointer, save :: ip1, ip2, ip3
    real(kind=single) :: inst      !! IT 2016 "withrawrows": ID_INST may be real
    real(kind=double), dimension(:), pointer, save :: ip2r   !!! - " -
    integer :: ctype                                         !!! - " -

    real(kind=double) ::cos_pa, sin_pa, rr, dr, r0r, d0r, &
                       dra, ddec, xr, yr


    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: col(9)

    src=-1000.
    nosrc=.false. 
    totrate=0.0
    totcounts=0.0 

    ctype = 0
      
! should be read from parameter file
!!!!    snrmin = 1

! open list

    set = DataSet(c_d1list, read) 
    tab = table(set, "SRCLIST")
    nrow = numberOfRows(tab)

    l1=1
 
   
!    if (hasColumn(tab,"ML_ID_SRC") .and. .not.xidfix) then
!    use ML_ID_SRC also for "xidfix", since non-unique BOX_ID_SRC numbers upset emldetect 

    if (hasColumn(tab,"ML_ID_SRC")) then
      col(l1) = column(tab,"ML_ID_SRC",read)  
      ip1 => int32Data(col(l1))
      l1=l1+1
    else if (hasColumn(tab,"BOX_ID_SRC")) then 
      col(l1) = column(tab,"BOX_ID_SRC",read)  
      ip1 => int32Data(col(l1))
      l1=l1+1
    else if (hasColumn(tab,"SRC_NUM")) then 
      col(l1) = column(tab,"SRC_NUM",read)  
      ip1 => int32Data(col(l1))
      l1=l1+1 
    else 
      allocate(ip1(0:nrow-1))  
      do i=0,nrow-1           
        ip1(i)=i+1             
      end do            

    end if
    if (hasColumn(tab,"ID_INST")) then
      col(l1) = column(tab,"ID_INST",read)  
      !!! IT 2016 "withrawrows": ID_INST may now be real
      ctype = columnDataType(col(l1))
      if (ctype == Real64) then
         ip2r => real64Data(col(l1))
      else
         ip2 => int32Data(col(l1))
      endif
      l1=l1+1
    else 
      allocate(ip2(0:nrow-1))
      ip2 = 0
    end if
    if (hasColumn(tab,"ID_BAND")) then  
      col(l1) = column(tab,"ID_BAND",read)  
      ip3 => int32Data(col(l1))
      l1=l1+1
    else
      allocate(ip3(0:nrow-1))
      ip3 = 0
    end if 


    radecpos=.true.    !!! IT 2016/02: use RA, DEC also if X_IMA, Y_IMA are NULL
    if ( hasColumn(tab,"X_IMA") .and. hasColumn(tab,"Y_IMA") ) then
      col(l1) = column(tab,"X_IMA",read)
      rp4 => real32Data(col(l1))
      l1=l1+1 
      col(l1) = column(tab,"Y_IMA",read)
      rp5 => real32Data(col(l1))
      l1=l1+1
      !!! IT 2016/02
      if (sum(rp4).ge.0 .and. sum(rp5).ge.0) radecpos=.false.
    endif
!!!    else if ( hasColumn(tab,"RA") .and. hasColumn(tab,"DEC") ) then
    if (radecpos) then
       if ( hasColumn(tab,"RA") .and. hasColumn(tab,"DEC") ) then
          col(l1) = column(tab,"RA",read)
          dp1 => real64Data(col(l1))
          l1=l1+1 
          col(l1) = column(tab,"DEC",read)
          dp2 => real64Data(col(l1))
          l1=l1+1
          if (abs(sum(dp1)).ge.0 .and. abs(sum(dp2)).ge.0) then
             radecpos=.true.
          else
             call error("NoPositionColumn","No valid source positions found in input list") 
          endif
       else
          call error("NoPositionColumn","No source position columns found in input list") 
       endif
    endif

    if (hasColumn(tab,"RATE")) then
      col(l1) = column(tab,"RATE",read)
      rp6 => real32Data(col(l1))
      l1=l1+1
    else
      allocate(rp6(0:nrow-1))
      rp6 = 0.
      nosrc=.true.
    end if
    if (hasColumn(tab,"SIGMA")) then
      col(l1) = column(tab,"SIGMA",read)
      rp8 => real32Data(col(l1))
      l1=l1+1
    else if (hasColumn(tab,"LIKE")) then
      col(l1) = column(tab,"LIKE",read)
      rp8 => real32Data(col(l1))
      l1=l1+1
    else if (hasColumn(tab,"DET_ML")) then
      col(l1) = column(tab,"DET_ML",read)
      rp8 => real32Data(col(l1))
      l1=l1+1
    else if (hasColumn(tab,"WCORR")) then
      col(l1) = column(tab,"WCORR",read)
      rp8 => real32Data(col(l1))
      l1=l1+1
    else
      allocate(rp8(0:nrow-1))
      rp8 = 1.
      !!! IT 2016/08 show a warning if no likelihood column is found
      call warning("MissingColumn", &
           "None of the columns (LIKE, DET_ML, SIGMA, or WCORR) found in the input source list. &
           &Input detections will be used in the , thus, and not sorted by likelihood.")
    end if
!   read source extent:
    if (hasColumn(tab,"EXT")) then
      input_ext=.true.
      col(l1) = column(tab,"EXT",read)
      rp9 => real32Data(col(l1))    
    else
      input_ext=.false.
      allocate(rp9(0:nrow-1))
      rp9 = 0.0
    end if
    if (hasColumn(tab,"SCTS")) then
      col(l1) = column(tab,"SCTS",read)
      rp10 => real32Data(col(l1))    
    else
      allocate(rp10(0:nrow-1))
      rp10 = 0.0
    end if    

    is = 0
    prio = 2

    do i=0,nrow-1

      isou = ip1(i)
      if (ctype == Real64) then
         inst = ip2r(i)
      else
         inst = ip2(i)
      endif
      band = ip3(i)
      if (radecpos) then
        cos_pa = dcos(crota2 * pi180)
        sin_pa = dsin(crota2 * pi180)
        rr = dp1(i) * pi180
        dr = dp2(i) * pi180
        r0r = crval1 * pi180
        d0r = crval2 * pi180
        call spher2(rr, dr, r0r, d0r, dra, ddec)
        xr =   dra * cos_pa + ddec * sin_pa
        yr = - dra * sin_pa + ddec * cos_pa
        xms = xr * overpi180 / crdelt1 + crpix1
        yms = yr * overpi180 / crdelt2 + crpix2
      else
        xms = rp4(i)
        yms = rp5(i)
      endif
      if (rp6(i) > 0.0) then 
        scts = rp6(i)
      else
        ! catch NULL values
        scts=0.0
      endif
      if (rp8(i) > 0.0) then
        snr = rp8(i)
      else
        snr = 0.0
      endif
      if (rp9(i) .gt. 0.0) then 
        ext_in = rp9(i)
      else
        ext_in = 0.0
      endif

      if (i == 0) then
        is = is+1
        prio = 2
      else if (ip1(i) /= ip1(i-1)) then
!!!!        if (sn(is) >= snrmin) is = is+1
        is=is+1
        prio = 2
      end if  

      if ((inst == 0 .or. inst == 7 .or. &
           band == 0 .or. band == 7) .and. prio > 0 .or. &
          prio > 1) then
        in(is) = isou
        xx(is) = xms
        yy(is) = yms
        extent(is) = ext_in
        if (rp10(i) > 0) totcounts(is)=rp10(i)
        if (scts > 0) totrate(is)=scts
        if (snr >= 0) then 
          sn(is) = snr
        else
          sn(is) = 1.0
        endif
      end if
      if (xidfix) then
        if (inst == 0) src(:,is)=0.0
        do ins=1,num_instr
          do iband=1,nxidbands(expip(ins))
            if (inst == expip(ins) .and. band == xiddef(iband,expip(ins))) then
              srcsum(expip(ins))=srcsum(expip(ins))+scts
              ipoint = ip(ins, 1)
              src(ipoint,is)=src(ipoint,is)+scts
            endif
          end do 
        end do 
      else

        if (inst == 0 .or. inst == 7) then
          prio = 0
        elseif ((band == 0 .or. band == 7) .and. prio > 0) then
          prio = 1
        endif
        if (inst /= 0 .and. band /= 0 .and. &
            inst /= 7 .and. band /= 7) then
          do ins=1,num_instr
            if (inst == expip(ins)) then
              if (band>0.and.band<=max_bands) then
                ipoint = ip(ins, band)
                if (ipoint>0.and.ipoint<=max_ima) then
                  src(ipoint, is) = scts
                else
                  nosrc=.true.
                end if
              else
                nosrc=.true. 
              end if
            end if
          end do
        end if

      end if
    end do
    nsou=is 

    do i=1,num_instr
      j=expip(i)
      if (num_image(i) >= xiddef(2,j) .and. xidfla) then 
        ipxid=ip(i,num_image(i))
        do is=1,nsou 
          src(ipxid,is)=src(ip(i,xiddef(1,j)),is) + &  
                        src(ip(i,xiddef(2,j)),is)
        end do
      end if
    end do 
    do ipoint=1,nim
      if (src(ipoint,1)<-999.) nosrc=.true.
    end do         
    if (nosrc) src=0.
    

    if (nsou == 0) then
!!! IT 2012/11/01: request for 3XMM pipeline
!!!       call error("EmptySourceList","Input source list is empty")
       call warning("EmptySourceList","Input source list is empty: no output results.")
       stop
    end if

    write(errstr,*) nsou, " sources read in"
    call message(AppMsg,Verbose)

  end subroutine get_sources


!!!!!! PHOTON MODE: not fully implemented
!!!  !---------------------------------------------------------------------
!!!!!!  subroutine get_smaps (no,x_start,y_start,x_end,y_end)
!!!  subroutine get_smaps (no)
!!!  !---------------------------------------------------------------------
!!!      
!!!    integer         no,numsm,maxsm  !!!  i,
!!!    parameter       (maxsm=1000)
!!!    integer         stasm(maxsm),endsm(maxsm)
!!!    integer         listsm(20000),stasmo(20000)
!!!!!!    integer         x_start, x_end, y_start, y_end
!!!
!!!    common /smaps/ numsm,stasm,endsm
!!!
!!!! meaning of nosm ?????
!!!!    call smoke (no,'b',x_start,y_start,x_end,y_end,0,0,0,0, &
!!!!                numsm,listsm,nosm,stasmo,ierr)
!!!
!!!! achtung: was passiert, wenn nosm.eq.listsm(numsm) ???
!!!    stasm(1:numsm)=stasmo(listsm(1:numsm))
!!!    endsm(1:numsm)=stasmo(listsm(1:numsm)+1)-1
!!!
!!!  end subroutine get_smaps
!!!      
!!!  !---------------------------------------------------------------------
!!!  subroutine get_event (no,ixs,iys,ixd,iyd,iamp,end_ev)
!!!  !---------------------------------------------------------------------
!!!
!!!    integer      no,nolast,numsm,maxsm,iseq,imap
!!!    parameter    (maxsm=1000)
!!!    integer      stasm(maxsm),endsm(maxsm)
!!!    integer      ixs,iys,ixd,iyd,iamp
!!!!    integer      icol1,icol2,icol3,icol4,icol5
!!!    logical      end_ev 
!!!
!!!    common /smaps/ numsm,stasm,endsm
!!!
!!!    data nolast /-99/ 
!!!
!!!    save
!!!
!!!    if (no.ne.nolast) then
!!!      imap = 1
!!!      iseq = stasm(imap)
!!!      nolast = no
!!!    end if       
!!!
!!!    iseq = iseq+1
!!!    if (iseq.gt.endsm(imap)) then
!!!      imap = imap+1
!!!      iseq=stasm(imap)
!!!    end if
!!!
!!!    end_ev=.false.
!!!    if (iseq.gt.endsm(numsm)) then
!!!      end_ev=.true. 
!!!    end if
!!!
!!!  end subroutine get_event
!!!--- FRAMEWORK PHOTON MODE: END

  !---------------------------------------------------------------------
  subroutine mulmlpari 
  !---------------------------------------------------------------------

    integer :: i 
    !real, dimension(max_ima) :: pimax, pimin
    integer, dimension(max_ima) :: band
!!! IT (obsolete)    logical, dimension(max_ima) :: ootflag 
    integer :: necf, nxidecf, npi, nimage
    !!--JR
    logical :: withimbuff

!!! IT (obsolete 2016/12)    type(datasett) :: set
!!! IT (obsolete 2016/12)    type(tablet) :: tab
    
    c_d1list = stringParameter("boxlistset")
    outtab = stringParameter("mllistset")
    dmlmin = realParameter("mlmin")      
!    dmlsou = realParameter("dmlsoumin")
    dmlext = realParameter("dmlextmin")
    maxext =  realParameter("maxextent")
    minext = realParameter("minextent")
    max_sfit = intParameter("nmaxfit")
    nsmul = intParameter("nmulsou")
    phofla = booleanParameter("useevents")
    if (phofla) call error("IllegalFlag","Photon mode not implemented")    
    posfla = booleanParameter("fitposition")
    negfla = booleanParameter("fitnegative")
    extfla = booleanParameter("fitextent")
!    if (extfla) then
!      call warning("NoExtent","Extent fitting is disabled in this version")
!      extfla=.false.
!    end if
    expfla = booleanParameter("withexpimage")
    detfla = booleanParameter("withdetmask")
    cntfla = booleanParameter("fitcounts")       
    errfla = booleanParameter("determineerrors")
    xidfla = booleanParameter("withxidband")
    xidfix =  booleanParameter("xidfixed")
    !!--- JR -----
    withimbuff = booleanParameter("withimagebuffersize")
    psfmod = stringParameter("psfmodel")
    !! -----------
    !!! IT 2014/05: extra output rows for sources in overlapping images
    withrawrows = booleanParameter("withrawrows")
    if (xidfix) then
      ! override values for position and extent fitting
      posfla=.false.
      extfla=.false.
    endif
    twostage =  booleanParameter("withtwostage")

    hotfla = booleanParameter("withhotpixelfilter")
    smpfla = booleanParameter("withsourcemap")
    usecalpsf = booleanParameter("usecalpsf")
!    simulate = booleanParameter("simulate")
    ences  = realParameter("scut")
    encee  = realParameter("ecut") 
    extmod =  ChangeCase(stringParameter("extentmodel"),2)
    if (extmod /= "GAUSSIAN" .and. &
        extmod /= "BETA") then 
          call error("badExtentmodel","Extent model incorrect - should be GAUSSIAN or BETA")
    endif

    if ((.not. usecalpsf) .and. (extmod == "BETA")) then 
       call warning("badExtentmodel","Extent model BETA only valid with usecalpsf=yes, using model GAUSSIAN instead")
       extmod = "GAUSSIAN"
    endif    


    if (simulate) then
      fitflag = .false.
      randomflag = .true.
       call warning("SimulationMode","No fit will be performed")
    endif 

!!!!!!!!!!!!!!!!! TEST thresholds for nmulsou fitting on SCTS/rate/like
    threshfla=booleanParameter("withthreshold")
    if (threshfla) then
       threshold =  realParameter("threshold")
       threshcol =  stringParameter("threshcolumn")
    endif

    if (max_sfit*nsmul>max_fit) &
       call error("WrongParam","nmaxfit*nmulsou must be <= 10")

    nxidbands(1) = parameterCount("xidpndef") 
    nxidbands(2) = parameterCount("xidm1def")
    nxidbands(3) = parameterCount("xidm2def")

    do i=1, parameterCount("hrpndef")
      hrdef(i,1) = intParameter("hrpndef",i-1)
    end do
    do i=1,nxidbands(1)
      xiddef(i,1)= intParameter("xidpndef",i-1)
    end do

    do i=1, parameterCount("hrm1def")
      hrdef(i,2) = intParameter("hrm1def",i-1)
    end do
    do i=1,nxidbands(2)
      xiddef(i,2)= intParameter("xidm1def",i-1)
    end do

    do i=1, parameterCount("hrm2def")
      hrdef(i,3) = intParameter("hrm2def",i-1)
    end do
    do i=1,nxidbands(3)
      xiddef(i,3)= intParameter("xidm2def",i-1)
    end do
    pixim = 3600.0 * dabs(crdelt1) 


    !!-- JR ------
    if (withimbuff) then
      imbuff = intParameter("imagebuffersize")
      withimbuff = .true.
      else
      imbuff = 640
      withimbuff = .false.
    endif
    !!------------

    xoff=0.
    yoff=0.
    if (booleanParameter("withoffsets")) &
       call warning("ObsoleteParameter", &
      "Parameter withoffsets is obsolete and ignored.") !!! 2016/12
!!!    if (booleanParameter("withoffsets")) then
!!!      set = dataSet(stringParameter("mergedListset"), READ)
!!!      if (hasAttribute(set,"XOFFPNM1")) then    
!!!        xoff(2)=Real32Attribute(set,"XOFFPNM1")
!!!      else 
!!!        tab = table(set,"SRCLIST")
!!!        if (hasAttribute(tab,"XOFFPNM1")) then
!!!          xoff(2)=Real32Attribute(tab,"XOFFPNM1")
!!!        end if        
!!!      end if
!!!      if (hasAttribute(set,"XOFFPNM2")) then    
!!!        xoff(3)=Real32Attribute(set,"XOFFPNM2")
!!!      else 
!!!        tab = table(set,"SRCLIST")
!!!        if (hasAttribute(tab,"XOFFPNM2")) then
!!!          xoff(3)=Real32Attribute(tab,"XOFFPNM2")
!!!        end if
!!!      end if
!!!      if (hasAttribute(set,"YOFFPNM1")) then    
!!!        yoff(2)=Real32Attribute(set,"YOFFPNM1")
!!!      else 
!!!        tab = table(set,"SRCLIST")
!!!        if (hasAttribute(tab,"YOFFPNM1")) then
!!!          yoff(2)=Real32Attribute(tab,"YOFFPNM1")
!!!        end if
!!!      end if
!!!      if (hasAttribute(set,"YOFFPNM2")) then    
!!!        yoff(3)=Real32Attribute(set,"XOFFPNM2")
!!!      else 
!!!        tab = table(set,"SRCLIST")
!!!        if (hasAttribute(tab,"YOFFPNM2")) then
!!!          yoff(3)=Real32Attribute(tab,"YOFFPNM2")
!!!        end if
!!!      end if
!!!      call release(set)
!!!    end if

    if (phofla) then
      call get_files("bkgimagesets",c_bgima,instrume,band,exp_id,.true.)
      if (expfla) call get_files("expimagesets",c_expima,instrume,band,exp_id,.false.)
      call get_files("imagesets",c_image,instrume,band,exp_id,.false.)
    else
      call get_files("imagesets",c_image,instrume,band,exp_id,.true.)
      call get_files("bkgimagesets",c_bgima,instrume,band,exp_id,.false.)
      if (expfla) call get_files("expimagesets",c_expima,instrume,band,exp_id,.false.)
      if (detfla) call get_files("detmasksets",c_detima,instrume,band,exp_id,.false.)
      !if (smpfla) call get_files("sourceimagesets",c_srcima,instrume,band,.false.)
    end if

 
    npi=parameterCount("pimin")          !!! IT 2016/04
    if (npi.gt.max_bands*num_instr) &
         call error("TooManyBands", &
         "Up to six energy bands are allowed per instrument. &
           &Please use fewer pimin / pimax values and adjust the input images.")

    pimin=2000
    do i=1,npi 
      pimin(i)=intParameter("pimin",i-1)
    end do
    if (npi/=nim) call warning("FileMismatch", &
      "Number of PI boundaries not equal number of images") 

    npi=parameterCount("pimin")
    pimax=4500
    do i=1,npi 
      pimax(i)=intParameter("pimax",i-1)
    end do
    if (npi/=nim) call warning("FileMismatch", &
      "Number of PI boundaries not equal number of images") 
    
    energy(1:nim)=0.5*(pimin(1:nim)+pimax(1:nim))

    necf=parameterCount("ecf")
    ecf=1.0
    do i=1, necf
      ecf(i) = realParameter("ecf",i-1)
    end do
    if (necf/=nim) call warning("FileMismatch", &
      "Number of ECFs not equal number of images") 

    if (xidfla) then
      nxidecf=parameterCount("xidecf")
      do i=1, nxidecf
        xidecf(i) = realParameter("xidecf",i-1)
      end do
      if (nxidecf/=num_instr) call warning("fileMismatch", &
        "Number of XID band ECFs not equal to number of instruments")
    end if 

!!!!!! IT/GL 2012/02: obsolete
!!!!    if (usecalpsf) then
!!!    point_resp(1:num_instr) = expip(1:num_instr)
!!!!    else
!!!!      point_resp=4
!!!!    end if

    if (smpfla) then
      nimage=parameterCount("sourceimagesets")
      if (nimage/=nim) call error("FileMismatch", &
                         "Inconsistent number of output images")
      do i=1,nimage
        c_srcima(i)=stringParameter("sourceimagesets",i-1)
      end do
    end if

  end subroutine mulmlpari

  !---------------------------------------------------------------------
  subroutine wr_stab(ll,nsrc,insrc,id_cluster,xms,yms,scts,rate,ra,dec,lii,bii,&
                     chipix,chipiy,ccdinr,offax, &
                     exms,eyms,escts,erate,eradec,&
                     exi_ml,ext,ext_err,ext_ml,enfrac,&
                     ecut2,flags,srcexpos,vignetting,srcontime) 
  !---------------------------------------------------------------------

    integer, intent(in) :: nsrc
    integer, dimension(max_fit), intent(IN) :: insrc
    integer, intent(IN) :: id_cluster 
    real, dimension(max_ima), intent(in) :: enfrac
    real, intent(in) :: ecut2
    real, dimension(max_fit), intent(in) :: xms, yms, exms, eyms
    real, dimension(max_fit) :: ext, ext_err, ext_ml 
    real, dimension(max_ima,max_fit), intent(in) :: scts, escts, exi_ml  
    character(len=*), intent(in) :: flags 
    real(kind=double), dimension(max_fit), intent(in) :: ra,dec,lii,bii
    real(kind=single), dimension(max_fit), intent(in) :: eradec
    integer, dimension(max_ins,max_fit), intent(in) :: chipix,chipiy,ccdinr 
    real(kind=double),  dimension(max_ins,max_fit), intent(in) :: offax
    real, dimension(max_ima,max_fit) :: bkg, expos, flx, rate, eflx, erate
    real, dimension(max_ima,max_fit), intent(in) :: srcexpos,vignetting,srcontime
    real, dimension(max_ins,max_fit) :: sctsi, esctsi, flxi, eflxi, ontimei, &
                                        ratei, eratei, bkgi, expoi, exi_mli ,mfraci 
    real, dimension(:,:,:), allocatable :: scts0, escts0, &
                                flx0, eflx0, rate0, erate0, expos0, &
                                bkg0, exi_ml0, ext_ml0, vig, mfrac0,ontimesrc
    integer, dimension(:,:,:), allocatable :: nzero0
    real, dimension(max_fit) :: sctst, esctst, flxt, eflxt, ratet, &
                                eratet, bkgt, expot, exi_mlt,mfract
    real(kind=single), dimension(:,:), allocatable :: hr1, hr2, hr3, hr4, &
                                                    hr1_err, hr2_err, hr3_err, hr4_err
    integer, dimension(:,:), allocatable :: rawx,rawy,rawchip
    real(kind=single),  dimension(:,:), allocatable :: offaxis
    integer, dimension(:,:), allocatable :: qflag

    real(kind=single) :: cutrad, hew, ootfac   !!! , sexp, fac, dum
!!! IT (obsolete)    real(kind=double), dimension(max_ins) :: ra_cen, dec_cen
    integer :: num_instr0, num_image0(max_ima), xid_arr0(max_ima)
    integer :: nt   !!! , nflux, nzt,nzi,nz0
    integer :: ix0,iy0,i,i0,j,j0,jj,k,l,ll,ipoint,instused,num_ima   !!!,nexp,ix,iy
    integer :: iimage,iinstr,ieband,nzero,nspari
    real(kind=double) :: inst, idinstmod         !!! IT 2014/05 (-> withrawrows)
    character(len=11) :: idinsttmp               !!! IT 2014/05 (-> withrawrows)
    integer, dimension(3) :: nf, ng
    integer, save :: id_src = 0
    logical, dimension(0:4) :: instflg
    integer :: astat

! Determine rates, fluxes, sums, per band and total

    allocate (scts0(max_ins,max_bands,nsrc), stat=astat)
    if (astat /= 0) call error("notEnoughMemory", "Not enough memory for arrays in 'eml_io'")
    allocate (escts0(max_ins,max_bands,nsrc))
    allocate (flx0(max_ins,max_bands,nsrc))
    allocate (eflx0(max_ins,max_bands,nsrc))
    allocate (rate0(max_ins,max_bands,nsrc))
    allocate (erate0(max_ins,max_bands,nsrc))
    allocate (expos0(max_ins,max_bands,nsrc))
    allocate (bkg0(max_ins,max_bands,nsrc))
    allocate (exi_ml0(max_ins,max_bands,nsrc))
    allocate (nzero0(max_ins,max_bands,nsrc))
    allocate (ext_ml0(max_ins,max_bands,nsrc))
    allocate (vig(max_ins,max_bands,nsrc))
    allocate (ontimesrc(max_ins,max_bands,nsrc))
    allocate (mfrac0(max_ins,max_bands,nsrc))
    allocate (rawx(max_ins,nsrc))
    allocate (rawy(max_ins,nsrc))
    allocate (rawchip(max_ins,nsrc))
    allocate (offaxis(max_ins,nsrc))
    allocate (hr1(max_ins,nsrc))
    allocate (hr2(max_ins,nsrc))
    allocate (hr3(max_ins,nsrc))
    allocate (hr4(max_ins,nsrc))
    allocate (hr1_err(max_ins,nsrc))
    allocate (hr2_err(max_ins,nsrc))
    allocate (hr3_err(max_ins,nsrc))
    allocate (hr4_err(max_ins,nsrc))
    allocate (qflag(max_ins,nsrc))
    

    nf=0
    ng=0
    nt=0
    do j0=1,num_instr
      j=expip(j0)
      num_ima=num_image(j0)
      if (xid_arr(j0)>0) num_ima=num_ima-1
      nf(j)=nf(j)+num_ima
      ng(j)=ng(j)+1
      nt=nt+num_ima
    end do
    sctsi=0.
    esctsi=0.
    flxi=0.
    eflxi=0.
    ratei=0.
    eratei=0.
    exi_mli=0.
    exi_mlt=0.
    expoi(1:3,1:nsrc)=0.
    ontimei(1:3,1:nsrc)=0.
    expot(1:nsrc)=0.
    expos=1.0
    bkgi(1:3,1:nsrc)=0.
    bkgt(1:nsrc)=0.
    sctst(1:nsrc)=0.
    esctst(1:nsrc)=0.
    flxt(1:nsrc)=0.
    eflxt(1:nsrc)=0.
    ratet(1:nsrc)=0.
    eratet(1:nsrc)=0.
    scts0=0.
    escts0=0.
    flx0=0.
    eflx0=0.
    rate0=0.
    erate0=0.
    exi_ml0=0.
    nzero0=0
    ext_ml0=0.
    bkg0=0.
    expos0=0.
    instflg(0)=.true.
    instflg(1:4)=.false.
    num_instr0=0
    hr1=-999.
    hr1_err=-999.
    hr2=-999.
    hr2_err=-999.
    hr3=-999.
    hr4=-999.
    hr3_err=-999.
    hr4_err=-999.
    rawx=-999.
    rawy=-999.
    rawchip=-999.
    offaxis=-999.

    xid_arr0=-1
    mfrac0(:,:,:)=1.0 
    mfract(1:nsrc)=1.0
    vig(:,:,:)=0.0
    ontimesrc(:,:,:)=0.0
    mfraci(:,:)=1.0


    do i=1,nsrc
      
      if (extfla .and. ext_ml(i) .eq. 0) then
        nspari=nspar-1
      else
        nspari=nspar
      endif
      do j0=1,num_instr
        j=expip(j0)
        instflg (j)=.true.
        num_instr0=max(j,num_instr0)
        num_image0(j)=num_image(j0)
        xid_arr0(j)=xid_arr(j0)
!       in case of raster scans, last set of raw coordinates will be written:
        rawx(j,i)=chipix(j0,i)
        rawy(j,i)=chipiy(j0,i)
        rawchip(j,i)=ccdinr(j0,i)
        offaxis(j,i)=real(offax(j0,i)/60.D0)
        do l=1,num_image(j0)
          ipoint=ip(j0,l) 

!!!          ix0 = min(max(nint(xms(i)+0.5),1),npix(1)) ! max(int(xms(i)),1)
!!!          iy0 = min(max(nint(yms(i)+0.5),1),npix(2)) ! max(int(yms(i)),1)

          OOTfac=1.0 
          if ((index(instrume(ipoint),"PN")>0.or.index(instrume(ipoint),"pn")>0) .and. .not. OotCorrDone(ipoint)) then
            if (index(submode,"rime")>0.and.index(submode,"ull")>0.or. &
                index(submode,"RIME")>0.and.index(submode,"ULL")>0) OOTfac=0.9411
            if (index(submode,"Ext")>0.and.index(submode,"ull")>0.or. &
                index(submode,"ext")>0.and.index(submode,"ull")>0.or. &
                index(submode,"EXT")>0.and.index(submode,"ULL")>0) OOTfac=0.97815
          end if

          ix0 = min(max(nint(xms(i)+0.5)-ixb0(ipoint),1),xbsize(ipoint))
          iy0 = min(max(nint(yms(i)+0.5)-iyb0(ipoint),1),ybsize(ipoint))
          bkg(ipoint,i) = bg_image0(ix0,iy0,ipoint)
          mfrac0(j,l,i)=min(mfrac0(j,l,i),psffrac(i,ipoint))
!!! IT 2017/06 consider only sources within the detection mask
!!!          if (srcexpos(ipoint,i)>0.0) then
          if ((srcexpos(ipoint,i)>0.0).and.(maskzero(i,ipoint)==0)) then
            rate(ipoint,i) = rate(ipoint,i)/OOTfac
            erate(ipoint,i)= erate(ipoint,i)/OOTfac
            flx(ipoint,i)  = rate(ipoint,i)/ecf(ipoint)*1.e-11
            eflx(ipoint,i) = erate(ipoint,i)/ecf(ipoint)*1.e-11
            scts0(j,l,i) = scts0(j,l,i)+scts(ipoint,i)
            escts0(j,l,i) = escts0(j,l,i)+escts(ipoint,i)**2
            flx0(j,l,i) = flx0(j,l,i)+flx(ipoint,i)*srcexpos(ipoint,i)
            eflx0(j,l,i) = eflx0(j,l,i)+(eflx(ipoint,i)*srcexpos(ipoint,i))**2
            rate0(j,l,i) = rate0(j,l,i)+rate(ipoint,i)*srcexpos(ipoint,i)
            erate0(j,l,i) = erate0(j,l,i)+(erate(ipoint,i)*srcexpos(ipoint,i))**2
            exi_ml0(j,l,i) = exi_ml0(j,l,i)+exi_ml(ipoint,i)
!!!(below)            nzero0(j,l,i) = maskzero(i,ipoint)                  !(checked-JR)
            bkg0(j,l,i) = bkg0(j,l,i)+bkg(ipoint,i)
            expos0(j,l,i) = expos0(j,l,i)+srcexpos(ipoint,i)
            vig(j,l,i)=vignetting(ipoint,i)
            ontimesrc(j,l,i)=srcontime(ipoint,i)
          end if
          !!! IT 2016/10 derive d.o.f.s from sum over /all/ maskzero values
          nzero0(j,l,i) = nzero0(j,l,i)+maskzero(i,ipoint)
        end do
      end do

!raster scans:
       do j=1,3
         do l=1,num_image(1)
          if (expos0(j,l,i)>0.) then
            escts0(j,l,i) = sqrt(escts0(j,l,i))
            rate0(j,l,i) = rate0(j,l,i)/expos0(j,l,i)
            erate0(j,l,i) = sqrt(erate0(j,l,i))/expos0(j,l,i)
            flx0(j,l,i) = flx0(j,l,i)/expos0(j,l,i)
            eflx0(j,l,i) = sqrt(eflx0(j,l,i))/expos0(j,l,i)            
          end if
        end do
      end do

      !use minimum psffrac for summary band row (GL 21/03/06): 

      do j0 = 1, num_instr
        j=expip(j0)
        if (instflg(j)) then
         num_ima=num_image(j0)
         if (xid_arr(j0) > 0) num_ima=num_image(j0)-1        
          do l=1,num_ima
            ipoint=ip(j0,l)             
            exi_mli(j,i)=exi_mli(j,i)+exi_ml(ipoint,i)
            mfraci(j,i)=min(psffrac(i,ipoint),mfraci(j,i))
            exi_mlt(i)=exi_mlt(i)+exi_ml(ipoint,i) 
            bkgi(j,i)=bkgi(j,i)+bkg(ipoint,i)
            bkgt(i)=bkgt(i)+bkg(ipoint,i)
            expoi(j,i)=expoi(j,i)+srcexpos(ipoint,i)
            ontimei(j,i)=max(ontimei(j,i),srcontime(ipoint,i))
            expot(i)=expot(i)+srcexpos(ipoint,i)
          end do
        end if
      end do
      do j=1,3  
        if (instflg(j)) then
         num_ima=num_image(1)
         if (xid_arr(1) > 0) num_ima=num_image(1)-1        
          do l=1,num_ima
            sctsi(j,i)=sctsi(j,i)+scts0(j,l,i)
            esctsi(j,i)=esctsi(j,i)+escts0(j,l,i)**2
            ratei(j,i)=ratei(j,i)+rate0(j,l,i)
            eratei(j,i)=eratei(j,i)+erate0(j,l,i)**2
            flxi(j,i)=flxi(j,i)+flx0(j,l,i)
            eflxi(j,i)=eflxi(j,i)+eflx0(j,l,i)**2
          end do
        end if
        esctsi(j,i)=sqrt(esctsi(j,i))
        eratei(j,i)=sqrt(eratei(j,i))
        eflxi(j,i)=sqrt(eflxi(j,i))
      end do 
! raster scans (21/12/04):
      do j=1,3  
        if (instflg(j)) then
          sctst(i)=sctst(i)+sctsi(j,i)
          esctst(i)=esctst(i)+esctsi(j,i)**2
          ratet(i)=ratet(i)+ratei(j,i)
          eratet(i)=eratet(i)+eratei(j,i)**2
          flxt(i)=flxt(i)+flxi(j,i)/eflxi(j,i)**2
          eflxt(i)=eflxt(i)+1./eflxi(j,i)**2
        end if
      end do 

      esctst(i)=sqrt(esctst(i))
      eratet(i)=sqrt(eratet(i))
      flxt(i)=flxt(i)/eflxt(i)
      eflxt(i)=1./sqrt(eflxt(i))

!!!!! correct dof for images where mask was zero:
      nzero=0
      do iimage=1,max_ima
        nzero=nzero+maskzero(i,iimage) 
      end do
!!!!!
      exi_mlt(i)=gammq(real(nspari+nt-nzero)/2.,exi_mlt(i))

      do j=1,3
        if (ng(j)>0) then
!!!!!!    correct dof for images where mask was zero:
          nzero=0
          do iinstr=1,max_ins
            do ieband=1,max_bands
              if (expip(iinstr) == j .and. ip(iinstr,ieband) > 0) then
                 nzero=nzero+maskzero(i,ip(iinstr,ieband))
              endif  
            end do
          end do
!!!!!
          exi_mli(j,i)=gammq(real(nspari+nf(j)-nzero)/2.,exi_mli(j,i))

          do l=1,num_image0(j)
            !!! IT 2016/10 - in order to be consistent with former versions (in
            !!! which maskzero was checked depending on srcexpos>0):
            !!! calculate exi_ml0 only if srcexpos -- now expos0 -- >0.
            !!! Otherwise, DET_ML would become 3e-7 in some rare cases (with
            !!! only two degrees of freedom) in which it used to be 0.0.
            if (expos0(j,l,i)>0) then
               exi_ml0(j,l,i) = &
                    gammq(real(nspari+ng(j)-nzero0(j,l,i))/2.,exi_ml0(j,l,i))
            else
               exi_ml0(j,l,i) = 0.0
            endif
          end do
        end if
      end do

    end do

    cutrad = sqrt(ecut2) 

    do j0=1,num_instr
     j=expip(j0)
     do i=1,nsrc
      if (hrdef(1,j)>0.and.hrdef(2,j)>0.and. &
          max(hrdef(1,j),hrdef(2,j))<=num_image(j0)) then    
       hr1(j,i)=(max(0.,rate0(j,hrdef(2,j),i))-max(0.,rate0(j,hrdef(1,j),i)))/ &
                (max(0.,rate0(j,hrdef(2,j),i))+max(0.,rate0(j,hrdef(1,j),i)))
       hr1_err(j,i)=2.*sqrt((rate0(j,hrdef(1,j),i)*erate0(j,hrdef(2,j),i))**2+ &
                            (rate0(j,hrdef(2,j),i)*erate0(j,hrdef(1,j),i))**2)/&
                (max(0.,rate0(j,hrdef(2,j),i))+max(0.,rate0(j,hrdef(1,j),i)))**2
      end if
      if (hrdef(3,j)>0.and.hrdef(4,j)>0.and. &
          max(hrdef(3,j),hrdef(4,j))<=num_image(j0)) then    
       hr2(j,i)=(max(0.,rate0(j,hrdef(4,j),i))-max(0.,rate0(j,hrdef(3,j),i)))/ &
                (max(0.,rate0(j,hrdef(4,j),i))+max(0.,rate0(j,hrdef(3,j),i))) 
       hr2_err(j,i)=2.*sqrt((rate0(j,hrdef(3,j),i)*erate0(j,hrdef(4,j),i))**2+ &
                            (rate0(j,hrdef(4,j),i)*erate0(j,hrdef(3,j),i))**2)/&
                (max(0.,rate0(j,hrdef(4,j),i))+max(0.,rate0(j,hrdef(3,j),i)))**2
      end if
      if (hrdef(5,j)>0.and.hrdef(6,j)>0.and. &
          max(hrdef(5,j),hrdef(6,j))<=num_image(j0)) then    
       hr3(j,i)=(max(0.,rate0(j,hrdef(6,j),i))-max(0.,rate0(j,hrdef(5,j),i)))/ &
                (max(0.,rate0(j,hrdef(6,j),i))+max(0.,rate0(j,hrdef(5,j),i)))
       hr3_err(j,i)=2.*sqrt((rate0(j,hrdef(5,j),i)*erate0(j,hrdef(6,j),i))**2+ &
                            (rate0(j,hrdef(6,j),i)*erate0(j,hrdef(5,j),i))**2)/&
                (max(0.,rate0(j,hrdef(6,j),i))+max(0.,rate0(j,hrdef(5,j),i)))**2
      end if
      if (hrdef(7,j)>0.and.hrdef(8,j)>0.and. &
          max(hrdef(7,j),hrdef(8,j))<=num_image(j0)) then    
       hr4(j,i)=(max(0.,rate0(j,hrdef(8,j),i))-max(0.,rate0(j,hrdef(7,j),i)))/ &
                (max(0.,rate0(j,hrdef(8,j),i))+max(0.,rate0(j,hrdef(7,j),i)))
       hr4_err(j,i)=2.*sqrt((rate0(j,hrdef(7,j),i)*erate0(j,hrdef(8,j),i))**2+ &
                            (rate0(j,hrdef(8,j),i)*erate0(j,hrdef(7,j),i))**2)/&
                (max(0.,rate0(j,hrdef(8,j),i))+max(0.,rate0(j,hrdef(7,j),i)))**2
      end if
     end do
    end do

    hew = 0. 
 
! Write out results
    instused=0
    do i=1,4
      if (instflg(i)) instused=instused+1
    end do

! Write instrument summary in all cases (changed 20/09/2005, GL):
    i0 = 0


    do k = 1,nsrc

      id_src = id_src+1

      do i = i0,num_instr0

         inst = i      !!! IT 2016 "withrawrows": ID_INST may now be real

       if (instflg(i)) then

        if (i == 0) then
! Write instrument summary
          call mlout(1,ll,id_src,insrc(k),inst,0,id_cluster, &
               sctst(k),esctst(k),xms(k),yms(k),exms(k),eyms(k), &
               ext(k),ext_err(k),exi_mlt(k),ext_ml(k), &
               bkgt(k),expot(k),flxt(k),eflxt(k),ratet(k),eratet(k), &
               ra(k),dec(k),eradec(k),lii(k),bii(k),0,0,0,0., &
               0.,0.,0.,0.,0.,0.,0.,0., 0.,0.,0., &
               cutrad,0.,hew,flags, .FALSE.)

        else
! Energy band summary only required if more than one band
          if (num_image0(i) > 1) then
            j0 = 0
          else
            j0 = 1
          end if

          do j = j0,num_image0(i)

! Set energy band id to 0
            if (j == 0) then
              jj = 0
            else
              jj = j
            end if 
! Set energy band to 9 for XID band
            if (jj == xid_arr0(i)) jj = 9

! Write energy band summary
            if (j == 0) then
              call mlout(2,ll,id_src,insrc(k),inst,jj,id_cluster, &
                 sctsi(i,k),esctsi(i,k),xms(k),yms(k),exms(k),eyms(k), &
                 ext(k),ext_err(k),exi_mli(i,k),ext_ml(k), &
                 bkgi(i,k),expoi(i,k),flxi(i,k), &
                 eflxi(i,k),ratei(i,k),eratei(i,k),ra(k),dec(k),eradec(k), &
                 lii(k),bii(k),rawx(i,k),rawy(i,k),rawchip(i,k),offaxis(i,k), &
                 hr1(i,k),hr1_err(i,k),hr2(i,k),hr2_err(i,k), &
                 hr3(i,k),hr3_err(i,k),hr4(i,k),hr4_err(i,k),0.,0., &
                 ontimei(i,k),cutrad,mfraci(i,k),hew,flags, .FALSE.)
            end if

! Write results per band and per instrument

            if (xidfix) jj=9

            if (j /= 0) then
!              write(*,*) "c)",exi_ml0(i,j,k)
              ipoint=ip(i,j)
              call mlout(3,ll,id_src,insrc(k),inst,jj,id_cluster, &
                   scts0(i,j,k), escts0(i,j,k),xms(k),yms(k), &
                   exms(k),eyms(k),ext(k),ext_err(k), &
                   exi_ml0(i,j,k),ext_ml(k),bkg0(i,j,k), &
                   expos0(i,j,k),flx0(i,j,k),eflx0(i,j,k), &
                   rate0(i,j,k),erate0(i,j,k),ra(k),dec(k),eradec(k), &
                   lii(k),bii(k),rawx(i,k),rawy(i,k),rawchip(i,k),&
                   offaxis(i,k), 0.,0.,0.,0.,0.,0.,0.,0.,enfrac(j),vig(i,j,k),&
                   ontimesrc(i,j,k), cutrad,mfrac0(i,j,k),hew,flags, .FALSE.)
              end if 

          end do   ! j = j0,num_image0(i) -> ID_BAND
         end if      
        end if     ! if (instflg(i))
      end do       ! i = i0,num_instr0    -> ID_INST

      !!! IT 2014/05: extra output rows for sources in overlapping images:
      !!! Repeat instruments-bands (num_instr, num_image instead of
      !!! num_instr0,num_image0) loops with modfied ID_INST.
      !!! Writing every source. This means an identification of the involved
      !!! images per source.
      !!! Alternatively, it could be restricted to sources which were detected
      !!! in more than one image per band (would affect edetect_stack).
      if (withrawrows) then
         do i = 1,num_instr

            !! use real64 ID_INST and decorate it (i.e.: 0,1,2,3) with the
            !! 10 digits of OBS_ID (signed 32-bit integer: max 2147483647)
            !! (IT 2017/06: adjustr needed)
            write(idinsttmp, '(A10,I1)') adjustr(exp_id(expind(i))(1:10)), &
                                         expip(expind(i))
            ! convert to number
            read(idinsttmp, *) idinstmod

            do j = 1,num_image(expind(i))
               ! only sources within the image mask
               if (maskzero(k,ip(expind(i),j))==0) then
                  ipoint = ip(expind(i),j)
                  ! call mlout with "raw" values
                  call mlout(3,ll,id_src,insrc(k),idinstmod,j,id_cluster,     &
                       scts(ipoint,k), escts(ipoint,k),xms(k),yms(k),         &
                       exms(k),eyms(k),ext(k),ext_err(k),                     &
                       exi_ml(ipoint,k),ext_ml(k),bkg(ipoint,k),              &
                       srcexpos(ipoint,k),flx(ipoint,k),eflx(ipoint,k),       &
                       rate(ipoint,k),erate(ipoint,k),ra(k),dec(k),eradec(k), &
                       ! rawx replaced by chipix, rawy by chipiy,
                       ! rawchip by ccdinr
                       lii(k),bii(k),chipix(expind(i),k),chipiy(expind(i),k), &
                       ccdinr(expind(i),k),real(offax(expind(i),k)),          &
                       0.,0.,0.,0.,0.,0.,0.,0.,enfrac(j),                     &
                       vignetting(ipoint,k),srcontime(ipoint,k),cutrad,       &
                       psffrac(k,ipoint),hew,flags, .TRUE.)
               end if
            end do   ! j = 1,num_image0(i) -> ID_BAND
         end do      ! i = 1,num_instr0    -> ID_INST
      end if

    end do           ! k = 1,nsrc

  end subroutine wr_stab

  !---------------------------------------------------------------------
  subroutine mlout (k,ll,id_src_d2,id_src_d1,id_inst,id_band,id_cluster,&
                    scts,scts_err,x_ima,y_ima,x_ima_err,y_ima_err,    &
                    ext,ext_err,exi_ml,ext_ml,bg_map, &
                    exp_map,flux,flux_err,rate,rate_err,ra,dec,       &
                    radec_err,lii,bii,rawx,rawy,rawchip,offaxis, &
                    hr1,hr1_err,hr2,hr2_err,hr3,hr3_err,hr4,hr4_err, &
                    enfrac,vig,ontimesrc,cutrad,mfrac,hew,flags,rawflag)
  !---------------------------------------------------------------------
        
    integer, intent(IN) :: k,id_src_d2,id_src_d1, &
                           id_band,id_cluster
                              !!! IT 2016 "withrawrows": ID_INST may now be real
    real(kind=double), intent(IN) :: id_inst
    real(kind=single), intent(IN) :: scts,scts_err,x_ima,y_ima,        &
                        x_ima_err,y_ima_err,ext,ext_err,&
                        exi_ml,ext_ml, &
                        bg_map,exp_map,flux,flux_err,rate,rate_err,    &
                        radec_err,hr1,hr1_err,hr2,hr2_err,hr3,hr3_err, &
                        hr4,hr4_err, &
                        enfrac,vig,ontimesrc,cutrad,mfrac,hew
    real(kind=double), intent(IN) :: ra,dec,lii,bii
    integer, intent(in) :: rawx,rawy,rawchip
    real(kind=single) :: offaxis 
    character(len=*), intent(IN) :: flags
    integer, intent(INOUT) :: ll

    integer, parameter :: ncol = 35
    integer :: i, l1
    integer, save :: nbuff   !!! IT - used below
    character(len=1024) :: image

    type(DataSetT), save :: Set
    type(DataSetT) :: set0, set1
    type(TableT), save :: tab
    type(ColumnT), dimension(50), save :: c 
    type(ColumnT), save :: sp36   !!! IT (obsolete)   , sp37, sp38

    integer, dimension(:), allocatable, save :: ia0,ia1,ia2,ia4,ia5, &
                                                ia39,ia40,ia41 
    real(kind=single), dimension(:), allocatable, save :: ra6,ra7,ra8, &
      ra9,ra10,ra11,ra12,ra13,ra14,ra15,ra16,ra17,ra18,ra19, &
      ra20,ra21,ra24,ra25,ra26,ra27,ra28,ra29,ra30,ra31,ra32,ra33,ra37, ra38, &
      ra42, ra43, ra44, ra45
!!! ra36,
    real(kind=double), dimension(:), allocatable,save :: da22,da23,da34,da35,ia3

    integer, dimension(:), pointer :: ip1,ip2,ip3,ip4,ip5,ip39,ip40,ip41    !!! ip0,
    real(kind=single), dimension(:), pointer :: rp6, rp7, rp8, &
      rp9, rp10, rp11, rp12, rp13, rp14, rp15, rp16, rp17, rp18, rp19, &
      rp20, rp21, rp24, rp25, rp26, rp27, rp28, rp29, rp30, rp31, rp32, &
      rp33, rp37, rp38, rp42, rp43, rp44, rp45
    real(kind=double), dimension(:), pointer :: dp22, dp23, dp34, dp35, ip3r
    real(kind=double), dimension(max_ins) :: ra_cen, dec_cen
    real(kind=double) :: xr, yr, rr, dr, dra, ddec, r0r, d0r, cos_pa, sin_pa

    logical, save :: first=.true.

    character(len=12) :: valflag = "            " !!!, &
!!! IT (obsolete)                         verflag = "            "
!!!    character(len=40) :: vcmflag = "                                        "
    character(len=8) :: rakey, dekey
    character(len=16) :: hrtmp !!! IT 2015/01
    character(len=2), dimension(3) :: ins
    integer :: ii, j, astat

    logical, intent(IN) :: rawflag       !!! IT 2015/01: no calls to settoNull in "raw" rows

    if (first) then
      first =.false.
      nbuff=max_sou*(nim+num_instr+1)
      allocate(ia0(0:nbuff), stat=astat)
      if (astat /= 0) call error("notEnoughMemory", "Not enough memory for arrays in 'eml_io'")
      allocate(ia1(0:nbuff))
      allocate(ia2(0:nbuff))
      allocate(ia3(0:nbuff))
      allocate(ia4(0:nbuff))
      allocate(ia5(0:nbuff))
      allocate(ra6(0:nbuff))
      allocate(ra7(0:nbuff))
      allocate(ra8(0:nbuff))
      allocate(ra9(0:nbuff))
      allocate(ra10(0:nbuff))
      allocate(ra11(0:nbuff))
      allocate(ra12(0:nbuff))
      allocate(ra13(0:nbuff))
      allocate(ra14(0:nbuff))
      allocate(ra15(0:nbuff))
      allocate(ra16(0:nbuff))
      allocate(ra17(0:nbuff))
      allocate(ra18(0:nbuff))
      allocate(ra19(0:nbuff))
      allocate(ra20(0:nbuff))
      allocate(ra21(0:nbuff))
      allocate(ra24(0:nbuff))
      allocate(ra25(0:nbuff))
      allocate(ra26(0:nbuff))
      allocate(ra27(0:nbuff))
      allocate(ra28(0:nbuff))
      allocate(ra29(0:nbuff))
      allocate(ra30(0:nbuff))
      allocate(ra31(0:nbuff))
      allocate(ra32(0:nbuff))
      allocate(ra33(0:nbuff))
!!!      allocate(ra36(0:nbuff))
      allocate(ra37(0:nbuff))
      allocate(ra38(0:nbuff))
      allocate(da22(0:nbuff))
      allocate(da23(0:nbuff))
      allocate(da34(0:nbuff))
      allocate(da35(0:nbuff))
      allocate(ia39(0:nbuff))
      allocate(ia40(0:nbuff))
      allocate(ia41(0:nbuff))
      allocate(ra42(0:nbuff))
      allocate(ra43(0:nbuff))
      allocate(ra44(0:nbuff))
      allocate(ra45(0:nbuff))
    end if

    if (k == 0) then
 
      ! ll=-ll
      l1=0     
      set = dataSet (outtab, CREATE)
!!!      if (phofla) then   !!! photon mode (not fully implemented)
!!!        image = c_evlist(1)   
!!!      else
        image = c_image(1) 
!!!      end if
      set0 = dataSet (image, READ) 
      call cp_head (set0, set, "EPIC EXPOSURE  ML SOURCE LIST", xid_arr, 4)
      call release(set0)
       
      set1 = DataSet(c_d1list, read)
      if (hasAttribute(set1,"BOXMODE")) &
        call setAttribute(set, "BOXMODE", &
                          stringAttribute(set1, "BOXMODE"), &
                          "Mode of eboxdetect input source list" ) 
      call release(set1)  
      call setAttribute(set, "HDUCLASS", "OGIP    ", &
                              "Format conforms to OGIP/GSFC conventions")
      call setAttribute(set, "HDUCLAS1", "SRCLIST   ", &
                              "File contains a source list")

      ins(1) = "PN"
      ins(2) = "M1"
      ins(3) = "M2"
      do j=1,num_instr
        cos_pa = dcos(crota2 * pi180)
        sin_pa = dsin(crota2 * pi180)
        xr = (xcen(j)-crpix1) * crdelt1 * pi180
        yr = (ycen(j)-crpix2) * crdelt2 * pi180
        dra =  xr * cos_pa - yr * sin_pa
        ddec = xr * sin_pa + yr * cos_pa
        r0r = crval1 * pi180
        d0r = crval2 * pi180
        call antisp(dra,ddec,r0r,d0r,rr,dr)
        ra_cen(j) = rr * overpi180
        dec_cen(j) = dr *overpi180

        write (rakey,'("OARA",a,i2.2)')ins(expip(j)),j
        write (dekey,'("OADE",a,i2.2)')ins(expip(j)),j
        call setAttribute(set, rakey, ra_cen(j), "Optical axis: RA [deg]") 
        call setAttribute(set, dekey, dec_cen(j), "Optical axis: DEC [deg]")
      end do

      !!! IT 2015/01 (for stacked source detection)
      !!! save hrpndef, hrm1def, hrm2def as header keywords
      do j=1,3
         write(rakey,'("HR",a,"DEF")') ins(j)
         write(hrtmp,'(8(i2))') hrdef(1:8,j)
         call setAttribute(set, rakey, hrtmp, "HR energy bands")
      end do


      tab = addTable (set, "SRCLIST", ll)

!      l1 = l1+1
!      c(l1) = addColumn (tab, "SRC_NUM",  Integer32, " ")
!      ip0 => int32Data(c(l1))
!      ip0 = ia0(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "ML_ID_SRC",  Integer32, " ")
      ip1 => int32Data(c(l1))
      ip1 = ia1(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "BOX_ID_SRC",  Integer32, " ")
      ip2 => int32Data(c(l1))
      ip2 = ia2(0:ll-1)

      l1 = l1+1
      !!! IT 2016 "withrawrows": ID_INST becomes real to host the OBS_ID
      if (withrawrows) then
         c(l1) = addColumn (tab, "ID_INST",    Real64, " ")
         ip3r => real64Data(c(l1))
         ip3r = ia3(0:ll-1)
         write(rakey,'("TDISP",i1)') l1
         call setAttribute(c(l1), "TDISP", "F11.0", &
                                        "Display ID_INST like integers")
      else
         c(l1) = addColumn (tab, "ID_INST",    Integer32, " ")
         ip3 => int32Data(c(l1))
         ip3 = ia3(0:ll-1)
      endif

      l1 = l1+1
      c(l1) = addColumn (tab, "ID_BAND",    Integer32, " ")
      ip4 => int32Data(c(l1))
      ip4 = ia4(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "ID_CLUSTER", Integer32, " ")
      ip5 => int32Data(c(l1))
      ip5 = ia5(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "SCTS",       Real32, "counts")
      rp6 => real32Data(c(l1))
      rp6 = ra6(0:ll-1)
      do i=0,ll-1        
        if (.not.rawflag .and. (ra17(i)<=0. .or. (ra38(i) < psfthreshold .and. ia3(i) .ne. 0))) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "SCTS_ERR",   Real32, "counts")
      rp7 => real32Data(c(l1))
      rp7 = ra7(0:ll-1)
      do i=0,ll-1        
        if (.not.rawflag .and. (ra17(i)<=0. .or. ra38(i) < psfthreshold .and. ia3(i) .ne. 0 )) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "X_IMA",      Real32, "image pixels")
      rp8 => real32Data(c(l1))
      rp8 = ra8(0:ll-1) 

      l1 = l1+1
      c(l1) = addColumn (tab, "X_IMA_ERR",  Real32, "image pixels")
      rp9 => real32Data(c(l1))
      rp9 = ra9(0:ll-1)
      if (xidfix) then
        do i=0,ll-1
           call settoNull(c(l1),i)
        enddo 
      endif
  
      l1 = l1+1
      c(l1) = addColumn (tab, "Y_IMA",      Real32, "image pixels")
      rp10 => real32Data(c(l1))
      rp10 = ra10(0:ll-1) 

      l1 = l1+1
      c(l1) = addColumn (tab, "Y_IMA_ERR",  Real32, "image pixels")
      rp11 => real32Data(c(l1))
      rp11 = ra11(0:ll-1)
      if (xidfix) then
        do i=0,ll-1
            call settoNull(c(l1),i)
        end do
      endif

      l1 = l1+1
      c(l1) = addColumn (tab, "EXT",        Real32, "image pixels")
      rp12 => real32Data(c(l1))
      rp12 = ra12(0:ll-1)
      do i=0,ll-1 
        if (.not.input_ext .and. .not.extfla) call settoNull(c(l1),i)
      end do


      l1 = l1+1
      c(l1) = addColumn (tab, "EXT_ERR",    Real32, "image pixels")
      rp13 => real32Data(c(l1))
      rp13 = ra13(0:ll-1)
      do i=0,ll-1
        if (.not.input_ext .and. (.not.extfla)) call settoNull(c(l1),i)
        if (xidfix)  call settoNull(c(l1),i)
      end do


      l1 = l1+1
      c(l1) = addColumn (tab, "DET_ML",     Real32, " ")
      rp14 => real32Data(c(l1))
      rp14 = ra14(0:ll-1)
      do i=0,ll-1        
        if (.not.rawflag .and. (ra17(i)<=0. .or. ra38(i)< psfthreshold  .and. ia3(i) .ne. 0)) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "EXT_ML",     Real32, " ")
      rp15 => real32Data(c(l1))
      rp15 = ra15(0:ll-1)
      do i=0,ll-1
        if (.not.extfla .or. ra15(i)<dmlext) call settoNull(c(l1),i)
      end do
      if (xidfix) then
        do i=0,ll-1
            call settoNull(c(l1),i)
        enddo 
      endif


      l1 = l1+1
      c(l1) = addColumn (tab, "BG_MAP",     Real32, "counts/pixel")
      rp16 => real32Data(c(l1))
      rp16 = ra16(0:ll-1)      
      do i=0,ll-1        
        if (.not.rawflag .and. ra17(i)<=0.) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "EXP_MAP",    Real32, "seconds")
      rp17 => real32Data(c(l1))
      rp17 = ra17(0:ll-1)
      do i=0,ll-1        
        if (.not.rawflag .and. (ia3(i)==0.or.ia4(i)==0.or.ra17(i)<=0.)) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "FLUX",       Real32, "cgs")
      rp18 => real32Data(c(l1))
      rp18 = ra18(0:ll-1)
      do i=0,ll-1        
        if (.not.rawflag .and. (ra17(i)<=0. .or. ra38(i) < psfthreshold .and. ia3(i) .ne. 0)) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "FLUX_ERR",   Real32, "cgs")
      rp19 => real32Data(c(l1))
      rp19 = ra19(0:ll-1) 
      do i=0,ll-1        
        if (.not.rawflag .and. (ra17(i)<=0. .or. ra38(i) < psfthreshold .and. ia3(i) .ne. 0)) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "RATE",       Real32, "counts/s")
      rp20 => real32Data(c(l1))
      rp20 = ra20(0:ll-1)
      do i=0,ll-1        
        if (.not.rawflag .and. (ra17(i)<=0..or. ra38(i) < psfthreshold .and. ia3(i) .ne. 0 )) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "RATE_ERR",   Real32, "counts/s")
      rp21 => real32Data(c(l1))
      rp21 = ra21(0:ll-1) 
      do i=0,ll-1        
        if (.not.rawflag .and. (ra17(i)<=0. .or. ra38(i) < psfthreshold .and. ia3(i) .ne. 0)) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "RA",         Real64, "deg")
      dp22 => real64Data(c(l1))
      dp22 = da22(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "DEC",        Real64, "deg")
      dp23 => real64Data(c(l1))
      dp23 = da23(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "RADEC_ERR",  Real32, "arcsec")
      rp24 => real32Data(c(l1))
      rp24 = ra24(0:ll-1)
      if (xidfix) then
        do i=0,ll-1
            call settoNull(c(l1),i)
        enddo 
      endif

      l1 = l1+1
      c(l1) = addColumn (tab, "LII",        Real64, "deg")
      dp34 => real64Data(c(l1))
      dp34 = da34(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "BII",        Real64, "deg")
      dp35 => real64Data(c(l1))
      dp35 = da35(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "RAWX", Integer32, "raw pixels")
      ip39 => int32Data(c(l1))
      ip39=ia39(0:ll-1)
      call setNullValue(c(l1),-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia39(i)<=0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "RAWY", Integer32, "raw pixels")
      ip40 => int32Data(c(l1))
      ip40=ia40(0:ll-1)
      call setNullValue(c(l1),-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia40(i)<=0) call settoNull(c(l1),i)
      end do


      l1 = l1+1
      c(l1) = addColumn (tab, "OFFAX", Real32, "arcmin")
      rp45 => real32Data(c(l1))
      rp45=ra45(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ra45(i)<= 0.) call settoNull(c(l1),i)
      end do


      l1 = l1+1
      c(l1) = addColumn (tab, "CCDNR", Integer32, " ")
      ip41 => int32Data(c(l1))
      ip41=ia41(0:ll-1)
      call setNullValue(c(l1),-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia41(i)<=0) call settoNull(c(l1),i)
      end do


      l1 = l1+1
      c(l1) = addColumn (tab, "HR1",        Real32, " ")
      rp25 => real32Data(c(l1))
      rp25 = ra25(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0.or.ra25(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR1_ERR",    Real32, " ")
      rp26 => real32Data(c(l1))
      rp26 = ra26(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0.or.ra26(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR2",        Real32, " ")
      rp27 => real32Data(c(l1))
      rp27 = ra27(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0.or.ra27(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR2_ERR",    Real32, " ")
      rp28 => real32Data(c(l1))
      rp28 = ra28(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0.or.ra28(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR3",        Real32, " ")

      rp29 => real32Data(c(l1))
      rp29 = ra29(0:ll-1)
      do i=0,ll-1        
       if (ia3(i)==0.or.ia4(i)/=0.or.ra29(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR3_ERR",    Real32, " ")
      rp30 => real32Data(c(l1))
      rp30 = ra30(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0.or.ra30(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR4",        Real32, " ")

      rp42 => real32Data(c(l1))
      rp42 = ra42(0:ll-1)
      do i=0,ll-1        
       if (ia3(i)==0.or.ia4(i)/=0.or.ra42(i)<-10.0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "HR4_ERR",    Real32, " ")
      rp43 => real32Data(c(l1))
      rp43 = ra43(0:ll-1)
      do i=0,ll-1        
        if (ia3(i)==0.or.ia4(i)/=0.or.ra43(i)<-10.0) call settoNull(c(l1),i)
      end do


      l1 = l1+1
      c(l1) = addColumn (tab, "CUTRAD",     Real32, "pixels") ! unit ???
      rp31 => real32Data(c(l1))
      rp31 = ra31(0:ll-1)


      l1 = l1+1
      c(l1) = addColumn (tab, "MASKFRAC",     Real32, " ")
      rp38 => real32Data(c(l1))
      rp38 = ra38(0:ll-1)
      do i=0,ll-1        
       if (ia3(i)==0) call settoNull(c(l1),i)
      end do
      


      l1 = l1+1
      c(l1) = addColumn (tab, "EEF",     Real32, " ")
      rp33 => real32Data(c(l1))
      rp33 = ra33(0:ll-1)
      do i=0,ll-1        
        if (ia4(i)==0.or.ia3(i)==0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "VIGNETTING",     Real32, " ")
      rp37 => real32Data(c(l1))
      rp37 = ra37(0:ll-1)
      do i=0,ll-1        
        if (ia4(i)==0.or.ia3(i)==0) call settoNull(c(l1),i)
      end do

      l1 = l1+1
      c(l1) = addColumn (tab, "ONTIME",     Real32, " ")
      rp44 => real32Data(c(l1))
      rp44 = ra44(0:ll-1)
      do i=0,ll-1        
        if (.not.rawflag .and. (ia3(i)==0 .or. ra44(i).le.0.0)) call settoNull(c(l1),i)
      end do


      call neighbour(ra8,ra10,ra32,ll)

      l1 = l1+1
      c(l1) = addColumn (tab, "DIST_NN",    Real32, "arcsec") ! unit ???
      rp32 => real32Data(c(l1))
      rp32 = ra32(0:ll-1)

      l1 = l1+1
      c(l1) = addColumn (tab, "FLAG",   String, " ", (/12/) )
      sp36=c(l1)
      

!      l1 = l1+1
!      c(l1) = addColumn (tab, "VER_FLAG",   String, " ", (/12/) )
!      sp37=c(l1)

!      l1 = l1+1
!      c(l1) = addColumn (tab, "VER_COMM",   String, " " , (/40/) )
!      sp38=c(l1)

      do ii=0,ll-1
        call setStringCell(sp36,ii,valflag) 
!        call setStringCell(sp37,ii,verflag) 
!        call setStringCell(sp38,ii,vcmflag) 
      end do

          
      do i=1,l1
        call release(c(l1))
      end do

      !!! IT 2015/01 & 2016/05 - a few more informational header keywords
      ! pixel size (motivation: be consistent with eboxdetect, where input
      ! images may be in detector or image coordinates and thus have different
      ! WCS header keywords)
      call setAttribute(set, "XBINSIZE", abs(crdelt1*3600.), &
           "Pixel size of input images in arcsec")
      call setAttribute(set, "YBINSIZE", abs(crdelt2*3600.), &
           "Pixel size of input images in arcsec")
      ! minimum detection likelihood
      call setAttribute(set, "LIKEMIN", dmlmin, &
           "Minimum detection likelihood to include a source")
      ! number of detections
      if (ll > 0) then
         write(errstr,*) ia1(ll-1), " detections written to ", outtab
         call setAttribute(set, "NDETECT", ia1(ll-1), &
              "Total number of detections")
      else
         write(errstr,*) "0 detections written to ", outtab
         call setAttribute(set, "NDETECT", 0, &
              "Total number of detections")
      endif
      call message(AppMsg,Verbose)

      call release (tab)
      call release (set)

    else if (ll < nbuff) then   !!! IT: max_sou -> nbuff

      ia1(ll) = id_src_d2
      ia2(ll) = id_src_d1
      ia3(ll) = id_inst
      ia4(ll) = id_band
      ia5(ll) = id_cluster
      ra6(ll) = scts
      ra7(ll) = scts_err
      ra8(ll) = x_ima

      ra9(ll) = x_ima_err
      ra10(ll) = y_ima 
      ra11(ll) = y_ima_err
      ra12(ll) = ext
      ra13(ll) = ext_err
      ra14(ll) = exi_ml
      ra15(ll) = ext_ml
      ra16(ll) = bg_map
      ra17(ll) = exp_map
      ra18(ll) = flux
      ra19(ll) = flux_err
      ra20(ll) = rate
      ra21(ll) = rate_err
      da22(ll) = ra
      da23(ll) = dec
      ra24(ll) = radec_err
      ra25(ll) = hr1
      ra26(ll) = hr1_err
      ra27(ll) = hr2
      ra28(ll) = hr2_err 
      ra29(ll) = hr3
      ra30(ll) = hr3_err 
      ra31(ll) = cutrad
      ra32(ll) = 0.
      ra33(ll) = enfrac
      da34(ll) = lii
      da35(ll) = bii
      ra37(ll) = vig
      ra38(ll) = mfrac    
      ia39(ll) = rawx
      ia40(ll) = rawy
      ia41(ll) = rawchip
      ra42(ll) = hr4
      ra43(ll) = hr4_err
      ra44(ll) = ontimesrc  
      ra45(ll) = offaxis
      ll = ll+1 

!!!    else if (ll == max_sou) then  !!! Helpdesk id=52410
    else if (ll == nbuff) then
      call warning("BufferOverflow  ","Overflow of output sources buffer") 
    end if

  end subroutine mlout

  subroutine neighbour (xms,yms,dist,nsrc)

  real(kind=single), dimension(0:max_sou), intent(in) :: xms, yms
  integer, intent(in) :: nsrc 
  real(kind=single), dimension(0:max_sou), intent(out) :: dist
  real(kind=single) d2,d2save
  integer :: i,j
  
  do i=0,nsrc-1
    d2save=1e10
    do j=0,nsrc-1
      d2=(xms(i)-xms(j))**2+(yms(i)-yms(j))**2
      if (d2 > 0.1 .and. d2 < d2save) then
        d2save=d2
      end if  
    end do
    dist(i)=sqrt(d2save)*pixim
  end do  

  end subroutine neighbour

  subroutine wr_srcmap

      real(kind=single), dimension(:,:), pointer :: r4arr

      integer :: i, j, ipoint, num_ima    !!!, ix, iy

      character(len=1024) :: output, refima
      
      type(DataSetT) :: set, set0
      Type(ArrayT) :: arr

      do i=1,num_instr
        num_ima=num_image(i) 
        if (xid_arr(i)>0) num_ima=num_ima-1
        do j=1,num_ima
          ipoint=ip(i,j) 

          nullify(r4arr)

          output=c_srcima(ipoint)
          set = dataset(output, write)
          arr = addArray(set, "PRIMARY", Real32, npix)
          r4arr => real32array2data(arr)
!!!! randomize image
!              if (randomflag) then
!                r4arr(ix-1,iy-1)=poissonRandom(bg_image(ix,iy,ipoint))  
!              else            
!                r4arr(ix-1,iy-1)=bg_image(ix,iy,ipoint)
!              endif
          r4arr(:,:) = 1e-5       !!! IT 2016/12
          r4arr(ixb0(ipoint):ixb1(ipoint), iyb0(ipoint):iyb1(ipoint)) = &
               bg_image(1:xbsize(ipoint), 1:ybsize(ipoint), ipoint)

!!!          if (phofla) then
!!!            refima = c_evlist(1)   
!!!          else
            refima = c_image(1) 
!!!          end if
          set0 = dataSet (refima, READ) 
          call cp_head (set0, set, "EPIC SOURCE IMAGE", xid_arr, 1)
          call release(set0)

          call release(arr)
          call release(set)   

        end do
      end do
 
      end subroutine wr_srcmap

end module eml_io













