! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: emldetect_mod
!
! Authors:  Hermann Brunner - AIP
!          
! Description: Maximum likelihood EPIC source searching
! Comments:
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module emldetect_mod

  use param  
  use dal
  use edetect_io_ml
  use emldata
  use emlfit
  use eml_io
  use psf


  implicit none 
  
  private 

  public emldetect

contains 
  
  !---------------------------------------------------------------------
  subroutine cut_radius (x,y,scut2,ecut2,enfrac)
  !---------------------------------------------------------------------

    use cal

    real, intent(in)  :: x, y
    real, intent(out) :: scut2, ecut2
    real, dimension(max_ima), intent(out) :: enfrac 

    real, parameter :: sec2mm = 7500.0*pi180/3600.0
    integer :: i,j,k,ipoint,ierr
    real    :: scut0,ecut0,ecut1,scut,ecut
    real(kind=double) :: E, theta, fraction 
    logical :: first = .true. 

    type (PsfT) :: ppsf
    type(PsfValidityRangesT) :: vrange 

    save

    ierr=0
    
    scut=0.
    ecut=0.

      if ((ences .lt. 1.0) .or. (encee .lt. 1.0)) then

!      always call setState, in case values are overwritten elsewhere (30/7/04)     
!      if (first) then
         !!! IT 2012/07: 
         !!! (still) only PN (which is the default to choose one ecut/scut for
         !!! all instruments); set psfmodel correctly
         !!! IT 2014/08: introduce slew-mode PSF and use select-case syntax
         select case (psfmod)
         case ("medium")
            call CAL_setState1(instrument=EPN, AccuracyLevel=ACCURACY_MEDIUM)
         case ("slew")
            call CAL_setState1(instrument=EPN, AccuracyLevel=ACCURACY_SLEW)
         case default   !!! "ellbeta"
            call CAL_setState1(instrument=EPN, AccuracyLevel=ACCURACY_ELLBETA)
         end select

!      call CAL_setState1(instrument=EMOS1,AccuracyLevel=ACCURACY_MEDIUM)

        vrange = CAL_PsfValidityRanges()
        first = .false.
!      end if

        do i=1,num_instr
          theta=sqrt((x-xcen(i))**2+(y-ycen(i))**2)*pixim 
          theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)
          do j=1,num_image(i)
            ipoint = ip(i,j) 
            E=energy(ipoint)
            ppsf = Cal_getPSF(E, theta, 0d0) 
            fraction=ences
            if (fraction .lt. 1.0) then 
              scut0 = CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
            else
              scut0 = ences
            endif
            scut=max(scut,scut0)
            fraction=encee
            if (fraction .lt. 1.0) then 
              ecut0 = CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
            else
              ecut0 = encee
            endif
            ecut=max(ecut,ecut0)
          end do
          do j=1,num_image(i)
            ipoint = ip(i,j) 
            E=energy(ipoint)
            ppsf = Cal_getPSF(E, theta, 0d0) 
            ecut1=0. 
            do k=1,99
              fraction=k*0.01
              ecut0=ecut1 
              ecut1=CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
              if (ecut1>ecut) exit
            end do
            enfrac(j)=((k-1)*(ecut1-ecut)+k*(ecut-ecut0))*0.01/(ecut1-ecut0)
          end do
        end do
      endif      

     ! if parameter value >= 1.0, interprete input in units of image pixels

     if (encee .ge. 1.0) ecut=encee
     if (ences .ge. 1.0) scut=ences

     scut2 = scut*scut
     ecut2 = ecut*ecut

  end subroutine cut_radius

  !---------------------------------------------------------------------
  subroutine sou_clus(xms,yms,scts,ext,insrc,nsrc,scut2,xx,yy,src,extent,in, &
                      k,nsou,indx,iwork,threshold, selvalue) 
  !---------------------------------------------------------------------

    real, dimension(max_sou), intent(in)  :: xx, yy, extent, selvalue 
    real, dimension(max_ima, max_sou), intent(in) :: src
    real, intent(in) :: scut2
    integer, dimension(max_sou), intent(in) :: in, indx
    integer, intent(in) :: k, nsou

    integer, dimension(max_sou), intent(inout) :: iwork

    real, dimension(max_fit), intent(out) :: xms, yms, ext
    real, dimension(max_ima,max_fit), intent(out) :: scts   
    integer, dimension(max_fit), intent(out) :: insrc
    integer, intent(out) :: nsrc
    real, dimension(max_fit) :: value
    real, intent(in) :: threshold
    integer :: nmulcount
    real, dimension(max_ima,max_fit) :: scts_old
    real, dimension(max_fit) :: xms_old,yms_old,ext_old
    integer, dimension(max_fit) :: insrc_old

    integer :: i,j,kk,sign,m
    logical :: found      

    insrc(:)=0

    kk=indx(nsou-k+1)
    xms(1) = xx(kk)
    yms(1) = yy(kk)
    ext(1) = extent(kk)
    scts(1:nim,1)=src(1:nim,kk)
    insrc(1)=in(kk)
    value(1)=selvalue(kk)

    nsrc=1
    found = .true.

    do while (found.and.nsrc.lt.max_sfit)
      found = .false.
      do j=k+1,nsou
        kk=indx(nsou-j+1)
        if(iwork(kk).eq.0) then
          do i=1,nsrc
            if ((xms(nsrc)-xx(kk))**2+(yms(nsrc)-yy(kk))**2 &
                .le.scut2.and.nsrc.lt.max_sfit) then
              found = .true.
              nsrc=nsrc+1
              iwork(kk)=1        
              xms(nsrc)=xx(kk)
              yms(nsrc)=yy(kk)
              value(nsrc)=selvalue(kk)
              ext(nsrc) = extent(kk)
              scts(1:nim,nsrc)=src(1:nim,kk)
              insrc(nsrc)=in(kk)
              exit
            end if
          end do  
        endif
        if (found) exit              
      end do   
    end do


    if (nsmul>1) then
      xms_old=xms
      yms_old=yms
      scts_old(1:nim,:)=scts(1:nim,:)
      ext_old=ext
      insrc_old=insrc
      nmulcount=0
      do i=1,nsrc
        if (value(i) > threshold .or. .not. threshfla) then
          do j=1,nsmul
            nmulcount=nmulcount+1
!            xms(nmulcount)=xms_old(i)
!            yms(nmulcount)=yms_old(i)
! test modifying input positions:
            sign=1
            do m=1,j
              sign=sign*(-1)
            enddo
            xms(nmulcount)=xms_old(i)+min(float(j-1),2.0)*sign
            sign=-1
            do m=1,j
              sign=sign*(-1)
            enddo
            yms(nmulcount)=yms_old(i)+min(float(j-1),2.0)*sign

            ! first guess for additional source is point-like 
            ext(nmulcount)=0.0
            scts(1:nim,nmulcount)=scts_old(1:nim,i)/nsmul
            insrc(nmulcount)=insrc_old(i)
          end do
        else
          nmulcount=nmulcount+1
          xms(nmulcount)=xms_old(i)
          yms(nmulcount)=yms_old(i)
          ext(nmulcount)=ext_old(i)
          scts(1:nim,nmulcount)=scts_old(1:nim,i)
          insrc(nmulcount)=insrc_old(i)                  
        endif
      enddo
!      nsrc=nsmul*nsrc
       nsrc=nmulcount
    end if 




  end subroutine sou_clus

  !---------------------------------------------------------------------
  subroutine fill_arrs(xms,yms,nsrc,ecut2)
  !---------------------------------------------------------------------

    real, dimension(max_fit), intent(in) :: xms, yms
    real, intent(in) :: ecut2
    integer, intent(in) :: nsrc

    integer :: i,j,k,l,ipoint,ix,iy
!!! (photon mode)  ierr,ixs,iys,ixd,iyd,iamp,ns,ne,i_offax,i_posang,
!!!               x_start,x_end,y_start,y_end
!!! (photon mode)    integer, dimension(max_ins) :: no_ev(max_ins)
    integer :: cutsum, masksum
    integer :: evmax, sum, ixmax, iymax          !!! IT (obsolete)     , tempsum
    integer :: ixshift, iyshift, ixlow, ixhig, iylow, iyhig, ixtmp, iytmp

    real    :: xmax,xmin,ymax,ymin,ecut
!!! (photon mode)  ss,xd,yd,offax,posang,x,y,
!!! (photon mode)    logical :: end_ev
    logical :: imvalid     !!! IT 2017 to verify subimage coordinates

!!!    ierr=0

!!! IT 2012/07
!    xmax=xms(1)
!    xmin=xms(1)
!    ymax=yms(1)      
!    ymin=yms(1)
!
!    do k=1,nsrc
!      xmax=max(xms(k),xmax)
!      xmin=min(xms(k),xmin)
!      ymax=max(yms(k),ymax)
!      ymin=min(yms(k),ymin)        
!    end do

    xmin = minval(xms(1:nsrc))
    xmax = maxval(xms(1:nsrc))
    ymin = minval(yms(1:nsrc))
    ymax = maxval(yms(1:nsrc))

    ecut = sqrt(ecut2)
    mask%ix_max=min(int(xmax+ecut),npix_arr(1,1))
    mask%ix_min=max(int(xmin-ecut),1)
    mask%iy_max=min(int(ymax+ecut),npix_arr(2,1))
    mask%iy_min=max(int(ymin-ecut),1)


! first check whether maskfrac criterion is met for each source
! (changed 05/01/07, GL):
    ! check fraction of cutout radius cover by valid pixels,
    ! zero mask, if fraction < maskthreshold:
    
    maskzero=0
    do k=1,nsrc
       do i=1,nim
          masksum=0
          cutsum=0
          iyshift = mask%iy_min - iyb0(i)        ! subimage coordinates
          do iy=mask%iy_min,mask%iy_max
             iytmp = iy-yms(k)
             iytmp = iytmp*iytmp
             ixshift = mask%ix_min - ixb0(i)
             do ix=mask%ix_min,mask%ix_max
                ixtmp = ix-xms(k)
                ixtmp = ixtmp*ixtmp
                if ((ixtmp+iytmp).le.ecut2) then
                   cutsum=cutsum+1
                   imvalid = .false.
                   if (detfla .and. ixshift.ge.1 .and. ixshift.le.xbsize(i) &
                        .and. iyshift.ge.1 .and. iyshift.le.ybsize(i)) then
                      if (mask_image(ixshift,iyshift,i)) imvalid = .true.
                   endif
                   if (.not. detfla .or. imvalid) then
                      imvalid = .false.
                      if (expfla .and. ixshift.ge.1 .and. ixshift.le.xbsize(i) &
                           .and. iyshift.ge.1 .and. iyshift.le.ybsize(i)) then
                         if (exp_image(ixshift,iyshift,i)>0.) imvalid = .true.
                      endif
                      if (.not. expfla .or. imvalid) then
                         masksum=masksum+1
                      endif
                   endif
                endif
                ixshift = ixshift+1
             end do
             iyshift = iyshift+1
          end do
          if (cutsum > 0) maskfrac(k,i)=float(masksum)/float(cutsum)
          !         do not fit source below threshold for maskfrac: 
          if (maskfrac(k,i) .lt. maskthreshold) maskzero(k,i)=1
       end do
    end do


    mask%mask(:,:,:)=0
    evmax=0
    iylow = 0
    do iy=mask%iy_min,mask%iy_max
       iylow = iylow+1    !!!   iylow = iy-mask%iy_min+1
       ixlow = 0
       do ix=mask%ix_min,mask%ix_max
        ixlow = ixlow+1   !!!   ixlow = ix-mask%ix_min+1
        mask%mask(ixlow,iylow,:) = 0  
        do k=1,nsrc          
          if ((ix-xms(k))**2+(iy-yms(k))**2.le.ecut2) then
            sum=0
            do i=1,nim
              !!! coordinates within subimage
              if ((ix.gt.ixb0(i)).and.(ix.le.ixb1(i)+1).and. &
                  (iy.gt.iyb0(i)).and.(iy.le.iyb1(i)+1)) then
              ixshift = ix-ixb0(i)
              iyshift = iy-iyb0(i)
              ! search brightest pixel   
              if (hotfla) then 
!!! 2016 - ipoint is undefined here, isn't it??? We are using "i"! ---                sum=sum+image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint) !(JR)
                sum = sum+image(ixshift,iyshift,i)
                if (sum>evmax) then
                  evmax=sum
                  ixmax=ix
                  iymax=iy
                end if 
              end if  !!! hotfla
!              mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1) = 1
              if (detfla .or. expfla) then  
                if (maskzero(k,i) == 0) then
                  if (.not. detfla .or. mask_image(ixshift,iyshift,i)) then       !(JR)
                    if (.not. expfla .or. exp_image(ixshift,iyshift,i) > 0.) &
                      mask%mask(ixlow,iylow,i) = 1
                  endif
                endif
              else
                 mask%mask(ixlow,iylow,i) = 1
              endif
              endif !!! subimage coordinates
            end do
          end if
        end do
      end do
    end do 
    if (hotfla) mask%mask(ixmax-mask%ix_min+1,iymax-mask%iy_min+1,:) = 0


    l = 1
    do i=1,num_instr

!!!      if (num_evlist(i).eq.0) then

        do j=1,num_image(i) 
          ipoint=ip(i,j)
          iyshift = mask%iy_min-iyb0(ipoint) ! subimage coordinates
          do iy=mask%iy_min,mask%iy_max  
            iylow = iy-mask%iy_min+1
            ixshift = mask%ix_min-ixb0(ipoint)
            do ix=mask%ix_min,mask%ix_max
              ixlow = ix-mask%ix_min+1
              if (mask%mask(ixlow,iylow,ipoint).eq.1) then 
!!! now, coordinates should be within the subimage anyway: otherwise mask%mask would be zero
!!!                if (ixshift.gt.0.and.ixshift.le.xbsize(ipoint).and.iyshift.gt.0.and.iyshift.le.ybsize(ipoint)) &
                    events(l)%n=image(ixshift,iyshift,ipoint)
                events(l)%ipt=ipoint
                l=l+1
!!! IT 2016/05 Throw error, if number of photons exceeds max_events.
!!! Makes this loop somewhat slower, but makes the issue clearer to users.
                 if (l.gt.max_events) &
                     call error("TooManyPhotons", &
                     "Cannot handle so many photons in the input images.")
              end if 
              ixshift = ixshift+1
            end do
            iyshift = iyshift+1
          end do
        end do


!!! PHOTON MODE has never been fully implemented.
!!!      else
!!!
!!!! read in event lists
!!!        ns=ip(i,1)
!!!        ne=ip(i,num_image(i))
!!!
!!!! open events file and get list of small maps
!!!!cc       call pix2sky(float(ix_min)
!!!!!!        call get_smaps (no_ev(i),x_start,y_start,x_end,y_end)
!!!        call get_smaps (no_ev(i))
!!!
!!!! go through list of small maps and read one event at a time
!!!        do while (.not.end_ev) 
!!!          call get_event (no_ev(i),ixs,iys,ixd,iyd,iamp,end_ev)
!!!          call sky2pix(ixs,iys,x,y,start_arr(1,i),step_arr(1,i))
!!!          ix=nint(x)
!!!          iy=nint(y)
!!!          if (ix.ge.mask%ix_min.and.ix.le.mask%ix_max.and. &
!!!              iy.ge.mask%iy_max.and.iy.le.mask%iy_max.and. &
!!!                mask%mask(ix-mask%ix_min+1,iy-mask%iy_min+1,1).eq.1) &
!!!          then
!!!            events(l)%xs  = x
!!!            events(l)%ys  = y 
!!!            events(l)%n   = 1
!!!            events(l)%ins = i
!!!            do j=ns,ne
!!!              if (iamp.ge.ampl_low(j).and.iamp.le.ampl_high(j)) &
!!!                events(l)%ipt = j
!!!            end do 
!!!            xd=ixd-xdcen(i)
!!!            yd=iyd-ydcen(i)
!!!            offax = sqrt(xd*xd+yd*yd)*detpix(i)/60.
!!!            i_offax = min(int(offax/epsmax(i)*max_offax)+1,max_offax)
!!!            events(l)%offax  = i_offax
!!!            posang = atan2(xd,yd)
!!!            i_posang = min(int(posang/twopi*max_posang)+1,max_posang)
!!!            events(l)%posang = i_posang
!!!            hist(i)%hist(i_offax,i_posang)=hist(i)%hist(i_offax,i_posang)+1.
!!!            l = l+1 
!!!          end if
!!!        end do
!!!
!!!! accumulate off axis & position angle histogram        
!!!        hist(i)%ns_offax=max_offax  
!!!        hist(i)%ne_offax=1  
!!!        hist(i)%ns_posang=max_posang  
!!!        hist(i)%ne_posang=1  
!!!
!!!        ss=0.
!!!        do i_offax=1,max_offax
!!!          do i_posang=1,max_posang
!!!            if (hist(i)%hist(i_offax,i_posang).gt.0.) then
!!!              hist(i)%ns_offax=min(hist(i)%ns_offax,i_offax)
!!!              hist(i)%ne_offax=max(hist(i)%ns_offax,i_offax)
!!!              hist(i)%ns_posang=min(hist(i)%ns_posang,i_posang)
!!!              hist(i)%ne_posang=max(hist(i)%ns_posang,i_posang)
!!!              ss=ss+hist(i)%hist(i_offax,i_posang)
!!!            end if
!!!          end do
!!!        end do
!!!
!!!        hist(i)%hist(:,:) = hist(i)%hist(:,:)/ss       
!!!
!!!      end if       
!!!--- FRAMEWORK PHOTON MODE: END

    end do

    npts = l-1

  end subroutine fill_arrs

!!!!!! PHOTON MODE: not fully implemented
 !!!  !---------------------------------------------------------------------
!!!  subroutine sky2pix (xsky,ysky,x,y,start,step)
!!!  !---------------------------------------------------------------------
!!!
!!!    integer          xsky,ysky
!!!    real             x,y
!!!    double precision start(2),step(2)
!!!
!!!    x=(xsky-start(1)+0.5)/step(1)
!!!    y=(ysky-start(2)+0.5)/step(2)
!!!
!!!  end subroutine sky2pix
!!!
!!!  !---------------------------------------------------------------------
!!!  subroutine pix2sky(x,y,xsky,ysky,start,step)
!!!  !---------------------------------------------------------------------
!!!
!!!    real              x,y
!!!    integer           xsky,ysky
!!!    double precision  start(1),step(1)
!!!
!!!    xsky=nint(x)
!!!    ysky=nint(y)
!!!
!!!  end subroutine pix2sky
!!!--- FRAMEWORK PHOTON MODE: END

  !---------------------------------------------------------------------
  subroutine rd_data (ierr)
  !---------------------------------------------------------------------
 
    integer           i,j,ipoint,maxdim,ierr,  ixmin, iymin, ixmax, iymax
    parameter         (maxdim=2)
    integer           dattyp,filtyp,iomode, ipxid, ix, iy
    integer, dimension(max_ins) :: num_ima0
!!! IT 2014/05: withrawrows
    integer ii(1), nn                         
    integer, dimension(num_instr) :: isort
    integer, dimension(num_instr) :: isort1
!!!
    real              exposure      
    real(kind=single), dimension(3) :: xnull, ynull
    double precision  start(maxdim),step(maxdim)
    character*1024    fname
    integer           astat

    ierr = 0

    dattyp   = 10
    filtyp   = 1
    iomode   = 0
    xnull = 0.
    ynull = 0.
    xid_arr=-1
    crpix1=-2d30
    crpix2=-2d30
    crval1=-2d30
    crval2=-2d30
    crdelt1=-2d30
    crdelt2=-2d30

    nim = ip(num_instr,num_image(num_instr))

! add XID  band to pointer array ip if at least band 1-3 are present
    num_ima0 = num_image
    do i = 1,num_instr
      j=expip(i)
      if (num_image(i) >= xiddef(2,j) .and. xidfla) then
        num_image(i)=num_image(i)+1
        nim=nim+1
        ip(i,num_image(i)) = nim
        xid_arr(i)=num_image(i)
        pimin(nim)=pimin(ip(i,xiddef(1,j)))
        pimax(nim)=pimax(ip(i,xiddef(2,j)))
      end if
    end do

    do i = 1,num_instr
      do j = 1,num_ima0(i)
        ipoint=ip(i,j)  

        fname = c_image(ip(i,j))

        call co_image(fname,start,step,npix,exposure,xcen(i),ycen(i), &
                      ra_pnt(i),dec_pnt(i),pa_pnt(i),datetime, &
                      crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                      crota2,xnull,ynull,expip(i),ipoint,nim)

        pixim = 3600.d0 * dabs(crdelt1) 

        if (j.eq.1) then
          start_arr(:,i) = start
          step_arr (:,i) = step 
          npix_arr (:,i) = npix 
        end if 

        fname = c_bgima(ip(i,j))

        call co_bgima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                      xnull,ynull,expip(i),ipoint,nim)

        if (expfla) then
           
          fname = c_expima(ip(i,j))
          call co_expima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                         xnull,ynull,expip(i),ipoint, &
                         xcen(i),ycen(i),exp_center(ipoint),nim)

          !exp_center(ipoint)=exp_image(int(xcen(i)),int(ycen(i)),ipoint)
        else
          if (i==1.and.j==1) then
             allocate (exp_image(ixsize,iysize,nim), stat=astat)
             if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'exp_image'")
          end if
          exp_image=1.0          
        end if

      end do

      if (detfla) then
        fname = c_detima(i)
        call co_mask(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                     xnull,ynull,expip(i),ip(i,1),nim)
        
        do j = 2,num_image(i)
           ! copy everything that's available - but not more ...
           ixmin = max(ixb0(ip(i,1)),ixb0(ip(i,j)))+1
           iymin = max(iyb0(ip(i,1)),iyb0(ip(i,j)))+1
           ixmax = min(ixb1(ip(i,1)),ixb1(ip(i,j)))+1
           iymax = min(iyb1(ip(i,1)),iyb1(ip(i,j)))+1
           mask_image(  ixmin-ixb0(ip(i,j)):ixmax-ixb0(ip(i,j)), &
                        iymin-iyb0(ip(i,j)):iymax-iyb0(ip(i,j)), ip(i,j) ) = &
             mask_image(ixmin-ixb0(ip(i,1)) :ixmax-ixb0(ip(i,1)), &
                        iymin-iyb0(ip(i,1)) :iymax-iyb0(ip(i,1)),  ip(i,1))
        end do
      else
        if (i==1) then
           allocate (mask_image(ixsize, iysize, nim), stat=astat)
           if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'mask_image'")
        end if
        mask_image=.true.
      end if   
       
    end do

    if (crpix1<-1d30) crpix1=0d0
    if (crpix2<-1d30) crpix2=0d0
    if (crval1<-1d30) crval1=0d0
    if (crval2<-1d30) crval2=0d0
    if (crdelt1<-1d30) crdelt1=0d0
    if (crdelt2<-1d30) crdelt2=0d0

! add XID band images if band 1 - 3 are available

    do i = 1,num_instr
     j=expip(i)
     if (num_image(i) >= xiddef(2,j) .and. xidfla) then 
      ipxid=ip(i,num_image(i))
           ixmin = max(ixb0(ipxid),ixb0(xiddef(1,j)),ixb0(xiddef(2,j)))+1
           iymin = max(iyb0(ipxid),iyb0(xiddef(1,j)),iyb0(xiddef(2,j)))+1
           ixmax = min(ixb1(ipxid),ixb1(xiddef(1,j)),ixb1(xiddef(2,j)))+1
           iymax = min(iyb1(ipxid),iyb1(xiddef(1,j)),iyb1(xiddef(2,j)))+1
      image(      ixmin-ixb0(ipxid)            :ixmax-ixb0(ipxid), &
                  iymin-iyb0(ipxid)            :iymax-iyb0(ipxid), &
                  ipxid) = &
       image(     ixmin-ixb0(ip(i,xiddef(1,j))):ixmax-ixb0(ip(i,xiddef(1,j))), &
                  iymin-iyb0(ip(i,xiddef(1,j))):iymax-iyb0(ip(i,xiddef(1,j))), &
                    ip(i,xiddef(1,j))) + &
       image(     ixmin-ixb0(ip(i,xiddef(2,j))):ixmax-ixb0(ip(i,xiddef(2,j))), &
                  iymin-iyb0(ip(i,xiddef(2,j))):iymax-iyb0(ip(i,xiddef(2,j))), &
                  ip(i,xiddef(2,j)))
      bg_image(   ixmin-ixb0(ipxid)            :ixmax-ixb0(ipxid), &
                  iymin-iyb0(ipxid)            :iymax-iyb0(ipxid), &
                  ipxid) = &
       bg_image(  ixmin-ixb0(ip(i,xiddef(1,j))):ixmax-ixb0(ip(i,xiddef(1,j))), &
                  iymin-iyb0(ip(i,xiddef(1,j))):iymax-iyb0(ip(i,xiddef(1,j))), &
                  ip(i,xiddef(1,j))) + &
       bg_image(  ixmin-ixb0(ip(i,xiddef(2,j))):ixmax-ixb0(ip(i,xiddef(2,j))), &
                  iymin-iyb0(ip(i,xiddef(2,j))):iymax-iyb0(ip(i,xiddef(2,j))), &
                  ip(i,xiddef(2,j)))
      exp_image(  ixmin-ixb0(ipxid)            :ixmax-ixb0(ipxid), &
                  iymin-iyb0(ipxid)            :iymax-iyb0(ipxid), &
                  ipxid) = &
       exp_image( ixmin-ixb0(ip(i,xiddef(1,j))):ixmax-ixb0(ip(i,xiddef(1,j))), &
                  iymin-iyb0(ip(i,xiddef(1,j))):iymax-iyb0(ip(i,xiddef(1,j))), &
                  ip(i,xiddef(1,j))) + &
       exp_image( ixmin-ixb0(ip(i,xiddef(2,j))):ixmax-ixb0(ip(i,xiddef(2,j))), &
                  iymin-iyb0(ip(i,xiddef(2,j))):iymax-iyb0(ip(i,xiddef(2,j))), &
                  ip(i,xiddef(2,j)))
      mask_image( ixmin-ixb0(ipxid)            :ixmax-ixb0(ipxid), &
                  iymin-iyb0(ipxid)            :iymax-iyb0(ipxid), &
                  ipxid) = &
       mask_image(ixmin-ixb0(ip(i,xiddef(1,j))):ixmax-ixb0(ip(i,xiddef(1,j))), &
                  iymin-iyb0(ip(i,xiddef(1,j))):iymax-iyb0(ip(i,xiddef(1,j))), &
                  ip(i,xiddef(1,j))) .and. &
       mask_image(ixmin-ixb0(ip(i,xiddef(2,j))):ixmax-ixb0(ip(i,xiddef(2,j))), &
                  iymin-iyb0(ip(i,xiddef(2,j))):iymax-iyb0(ip(i,xiddef(2,j))), &
                  ip(i,xiddef(2,j)))

      exp_center(ipxid)=(exp_center(ip(i,xiddef(1,j)))+ &
                         exp_center(ip(i,xiddef(2,j))))/2.0
      energy(ipxid)=(energy(ip(i,xiddef(1,j)))+energy(ip(i,xiddef(2,j))))/2.0 
      ecf(ipxid)=xidecf(i)
     end if
    end do

    allocate (bg_image0(ixsize,iysize,nim), stat=astat)
    if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'bg_image0'")
    bg_image0 = bg_image       
 
     !!! IT 2014/05: prepare for extra output rows for sources in overlapping
     !!! images: indices sorted sort by exp_id and by inst_id
     ! sorting: successively identify array minima and set them to huge values
     if (withrawrows) then
        nn = sum(num_image)
        expind=0
        isort=0
        isort1=0
        isort = expip(1:num_instr)
        ! sort by expip (inst_id)
        do i=1,num_instr
           ii = minloc(isort)
           expind(i) = ii(1)
           isort(ii(1)) = huge(ii(1))
        end do
        ! sort by exp_id - use first 8 digits
        isort=0
        do i=1,num_instr
           read(exp_id(expind(i))(1:8), *) isort(i)
        end do
        isort1 = expind(1:num_instr)
        do i=1,num_instr
           ii = minloc(isort)
           expind(i) = isort1(ii(1))
           isort(ii(1)) = huge(isort(ii(1)))
        end do
     end if
     !!!!!!!!!!!!!!!

  end subroutine rd_data

  !---------------------------------------------------------------------
  subroutine sorind(a,w,ind,indw,n,nw,ierr)
  !---------------------------------------------------------------------

    integer :: n,nw,ierr,ind(n),indw(nw)  !!!,i,j
    real :: a(n),w(nw)

    ierr=0

! index the array ind such that a is in ascending order
    ind(1:n) = 0
    indw(1:nw) = 0

    if (n.gt.1) then
      call indexr(n,a,ind)
    else if (n.eq.1) then
      ind(1) = 1
    else if (n.eq.0) then
      ind(1) = 0
    endif

! sort a according to ind
    w(1:n) = a(ind(1:n))
    a(1:n) = w(1:n)
    indw(1:n) = ind(1:n)

  end subroutine sorind    


  !---------------------------------------------------------------------
  subroutine emldetect
  !---------------------------------------------------------------------

!  author : GRH              date: 15-APR-1987 15:48
!  update : GRH              date: 3-DEC-1992 09:13  
!  update : PTK              date: 05-JAN-1993 17:27 EXSAS implementation
!  update : PTK              date: 14-JUN-1993 20:17 table output / correct
!                                                    bugs
!  update : EXSAS            date: 15-OCT-1993 08:52
!  update : PTK              date: 08-NOV-1993 10:41 POSFLA, new parameter
!                                                    IPOS in CGAUSN
!  update : PTK              date: 18-JAN-1994 09:55 NELEM1,NELEM2
!  update : EXSAS            date: 19-JAN-1994 16:43
!  update : HEB              date: 27-MAY-1996

!     character*80  table
!     character*2   io
!     character*70  d2par

      integer ::    k, ll, ierr, &
                    indx(max_sou),insrc(max_fit), &
                    iwork(max_sou),kk,nsou, &
                    in(max_sou), nsrc, istat, id_clu 
      real ::       scut2, ecut2, xx(max_sou), yy(max_sou), extent(max_sou)
      real ::       totcounts(max_sou),totrate(max_sou)
      real(kind=single), dimension(max_fit) :: xms, yms, exms, eyms, eradec 
      real(kind=single), dimension(max_ima, max_fit) :: scts, escts, rate, erate
!!! IT (obsolete)      real(kind=single), dimension(max_ima, max_fit) :: detcounts
      real(kind=single), dimension(max_ima, max_fit) :: srcexpos,vignetting,srcontime
      real(kind=single), dimension(max_ima, max_sou) :: src

      real(kind=single), dimension(max_sou) :: sn, rwork
      real(kind=single), dimension(max_sou) :: selvalue
      real(kind=double), dimension(max_fit) :: ra, dec, lii, bii
      integer, dimension(max_ins,max_fit) :: chipix,chipiy,ccdinr   !!! ,istate1 
      real(kind=double), dimension(max_ins,max_fit) :: offax
 
      real, dimension(max_ima,max_fit) :: exi_ml 
      real, dimension(max_fit) :: ext, ext_err, ext_ml 
      real, dimension (max_ima) :: enfrac
      character(len=5) :: flags
      logical :: first, fitdone                    
      integer :: nsmul_old

      integer :: dtvalues(1:8)  !!! IT 2017 - informational

      call mulmlpari

! Read input images
      call rd_data (ierr)

      call get_sources (nsou,xx,yy,sn,src,extent,in,totrate,totcounts)

      
! set threshold for nmulsou fitting:
      if (threshfla) then
         if (threshcol == "LIKE" .or. threshcol == "like") selvalue=sn
         if (threshcol == "RATE" .or. threshcol == "rate") selvalue=totrate
         if (threshcol == "SCTS" .or. threshcol == "scts") selvalue=totcounts
      endif      

! Sort the array, key is SN
      call sorind(sn, rwork, indx, iwork, nsou, max_sou, istat)


      iwork=0
      id_clu=0
      ll=0
       
! Do loop, start from brightest source 
      nsmul_old=nsmul 
      do k=1,nsou

        fitdone = .false.
!             write(*,*) k, " sources processed"
        if (mod(k,10) == 0) then
           call date_and_time(values=dtvalues)    !!! IT 2017 - informational
!!!           write(errstr,*)  k, " sources processed"
           write(errstr, &
            '(i4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2,".",i3.3,i8,a)') &
            dtvalues(1:2), &
            dtvalues(3)+(isign(1,dtvalues(5)-dtvalues(4)/60)-1)/2, &
            mod(dtvalues(5)-dtvalues(4)/60 + 24,24), dtvalues(6:8), &
            k, " sources processed"
           call message(AppMsg,Verbose)
        end if
        kk=indx(nsou-k+1)
        if (iwork(kk).eq.0) then
          iwork(kk)=1
          id_clu=id_clu+1





! Calculate cut radii
          call cut_radius (xx(kk),yy(kk),scut2,ecut2,enfrac)
         ! initialize nsrc and ext_ml(1): (3/8/05)
         nsrc=1
         ext_ml(1)=0.0

         if (twostage) then
           first=.true.
           nsmul=1

           ! Identify source cluster
           call sou_clus (xms,yms,scts,ext,insrc,nsrc,scut2,xx,yy,src,extent,in, &
                          k,nsou,indx,iwork,threshold,selvalue) 

           ! Fill data arrays
           call fill_arrs (xms,yms,nsrc,ecut2)
           ! change "exit" to  "if then" (5/5/06, GL)
           if (npts.gt.((nim+3)*nsrc+1)) then
             ! Perform multi-source ML fit
              call mulsou(nsrc,xms,yms,ext,scts,rate,ra,dec,lii,bii, &
                          chipix,chipiy,ccdinr,offax, &
                          exms,eyms,ext_err,escts,erate,eradec, &
                          exi_ml,ext_ml,ecut2,srcexpos,vignetting,srcontime,insrc,first)
              fitdone = .true.
           endif
         endif

         if (.not. twostage .or. (maxval(ext_ml(1:nsrc)) .gt. 0)) then
            nsmul=nsmul_old
            first=.false.
            call sou_clus (xms,yms,scts,ext,insrc,nsrc,scut2,xx,yy,src,extent,in, &
                           k,nsou,indx,iwork,threshold,selvalue) 
            call fill_arrs (xms,yms,nsrc,ecut2)
           ! change "exit" to  "if then" (5/5/06, GL)
           if (npts.gt.((nim+3)*nsrc+1)) then
              call mulsou(nsrc,xms,yms,ext,scts,rate,ra,dec,lii,bii, &
                          chipix,chipiy,ccdinr,offax, &
                          exms,eyms,ext_err,escts,erate,eradec, &
                          exi_ml,ext_ml,ecut2,srcexpos,vignetting,srcontime,insrc,first)
              fitdone = .true.
           endif
         endif



! Set dummy output values
            flags="     "
           
! write source info to table

!!! IT 2015/01
!!!          if (fitdone) then
          if (fitdone .and. nsrc.gt.0) then
            call wr_stab(ll,nsrc,insrc,id_clu,xms,yms,scts,rate,ra,dec,lii,bii,&
                         chipix,chipiy,ccdinr,offax, &
                         exms,eyms,escts,erate,eradec,exi_ml,ext,ext_err,&
                         ext_ml,enfrac,ecut2,flags,srcexpos,vignetting,srcontime)
          endif
        endif

      end do

! Close output table       
      ! ll = -ll
      call mlout(0,ll,k,0,0d0,0,0,0.,0.,0.,0.,0., &
                 0.,0.,0.,&
                 0.,0.,0.,0.,0.,0.,0.,0., &
                 0d0,0d0,0.,0d0,0d0,0,0,0,0.,0.,0.,0.,0.,0.,0.,0.,0., &
                 0.,0.,0.,0.,0.,0.,flags, withrawrows) 


       

! write out source map
      if (smpfla) call wr_srcmap

    end subroutine emldetect
  
end module emldetect_mod








