! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module emlfit

  use oal
  use cal
  use param
  use dal
  use sla
  use edetect_io_ml
  use edetect_gen
  use emldata
  use psf
  use fftpsf

  implicit none

  private

  public mulsou, model, mlfit, funct, like, ordr, pordr


contains

  !---------------------------------------------------------------------
      subroutine mulsou (nsrc,sxms,syms,ssigma,sscts,rate,ra,dec,lii,bii, &
                         chipix,chipiy,ccdinr,offax, &
                         exms,eyms,esigma,escts,erate,eradec, &
                         exi_ml,ext_ml,ecut2,srcexpos,vignetting,srcontime,insrc,first)
  !---------------------------------------------------------------------

      real, intent(in) :: ecut2
      integer, intent(inout) :: nsrc
      real, dimension(max_fit), intent(inout) :: sxms, syms, ssigma,  &
                                                 exms, eyms, esigma, ext_ml

      real(kind=double), dimension(max_fit), intent(out) :: ra, dec, lii, bii
      real(kind=double), dimension(max_fit) :: xms, yms ,sigma
      real(kind=double), dimension(max_ima,max_fit) :: scts
      real(kind=single), dimension(max_fit) ::  eradec
      real, dimension(max_ima,max_fit), intent(inout) :: sscts, escts, &
                                                         rate, erate,exi_ml
      real, dimension(max_ima,max_fit) :: srcexpos
      real(kind=single),  dimension(max_ima,max_fit) :: vignetting,srcontime
      integer, dimension(max_ins,max_fit), intent(out) :: chipix,chipiy,ccdinr 
      real(kind=double), dimension(max_ins,max_fit)  :: offax,azim
      real :: effarea_cen,effarea_src
      real :: distsq,extdiff,mindist=1.5,minextdiff=1.0
      logical :: multflag, maskbad
      real ::       yfit1(npts),yfit2(nim), &
                    yfit3,dml(max_fit), &
                    fe(max_par),df, &
                    felhi,fello, flamda, sc,rcal, x1,y1,exp,expn,expj,ene

      real(kind=double) :: like00,like0,like1,likold,lik_arr(nim), &
                           like00arr(nim),lik_arr_save(nim,0:nsrc), &
                           likesave(0:nsrc),delfit(max_par), &
                           a(max_par),deltaa(max_par),deltae(max_par),sigmaa(max_par),afit(max_par), &
                           sigfit(max_par),atemp(max_par) !,amerk(max_par),likmerk,lik_amerk(max_ima),like1_arr(max_ima)
   
      ! diagnostic variable:
      real :: likediff
      !
      real(kind=double) :: asave(max_par,0:nsrc)
!      real(kind=double) :: astart(max_par)
      real :: like2dof(0:nsrc),likecorr
      real :: dmlthresh   !,bestlike
      integer :: bestfit=0
      logical :: oldfit
!      integer :: ordsave(max_par)
      integer :: ntermf0,ntfix0,ndof
      integer :: nn
      !
      real(kind=double) :: xr,yr,dra,ddec,r0r,d0r,rr,dr,sin_pa,cos_pa, &
                    lr,br !,fac
      integer ::    ierr,ioff0,ioff1,ioff,noff,i,ifel,ix0,iy0, &
                    npsou,ipoint,iordr(max_par),iordr0(max_par), &
                    j,insrc(max_fit),ical,ical2,nzero, &
                    ix,iy,l,m,n,nlauf,ntermf,ntfix,nffix,np,nima,nband, &
                    ixshift, iyshift, ixlow, ixhig, iylow, iyhig, &
                    xtmp0, ytmp0, ytmp1
 !,iorde(max_par),iimage,masksum,cutsum
      logical       done,first
      character(len=6) :: minmeth 
        

      integer, parameter :: irad = 120, irad2=14400, maxfun=1000, iprint=0
      real, parameter :: delt = 1e-6
      real(kind=double), parameter ::  rhobeg=0.1, rhoend=1.D-6
!---JR
!      real :: testsum
  
!     integer:: istate1    
       minmeth='mrqmin'
!      minmeth='powell' 

     likold = 0.0

     if (testoutput) write(*,*) "start MULSOU:",nsrc
!     write(*,*) "input scts:", scts(1:6,1)
!     if (nsrc == 2) write(*,*) "scts:", scts(1:15,2)
     if (testoutput) write(*,*) "input x1,y1:", xms(1),yms(1)
     if (testoutput .and. nsrc == 2) write(*,*) "x2,y2:", xms(2),yms(2)

      ! read new CAL PSF for new cluster:
      newcal = .true.

      oldfit = .false.
      
!      convert fit parameters to double precision (23/06/06 GL)      
      xms(1:max_fit) = dble(sxms(1:max_fit))
      yms(1:max_fit) = dble(syms(1:max_fit))
      sigma(1:max_fit)=dble(ssigma(1:max_fit))
      scts(1:max_ima,1:max_fit)=dble(sscts(1:max_ima,1:max_fit))


!      call system_clock(clock0,clock2,clock3)


      count_psf=0

      ierr=0
      ntfix = 0
      nlauf=0

!      ical=45/pixim
!      ical=60/pixim
!---------------------------------------------------------------------------------
!--- (JR) ------------------------------------------------------------------------
    psfmod = stringParameter("psfmodel")
    !!! IT 2014/08 introducing slew-mode PSF: removed boolean variable modpsf;
    !!! same values for slew and ellbeta
    if (psfmod .eq. "medium") then
       ical=60./pixim
    else    !!! "ellbeta" or "slew"
       ical=280./pixim
    end if


      rcal=ical
      ical2 = ical*ical
      nspar=3
      npsou=nim+nspar
      np=npsou*nsrc
      do i=1,nsrc
        ioff=npsou*(i-1)
        sc=0.0
        do j=1,nim
!!!          expj=exp_image(ixb(nint(xms(i)),j),iyb(nint(yms(i)),j),j)  !(JR)
!!!!          expj=exp_image(max(1,min(npix(1),nint(xms(i)))), &
!!!!                           max(1,min(npix(2),nint(yms(i)))),j)
!!!          expj=max(expj,1.0)
          ixshift = nint(xms(i))-ixb0(j)
          iyshift = nint(yms(i))-iyb0(j)
          !!! detection coordinates xms / yms could be outside the subimage
          !!! range of one (out of several) images
          if (ixshift.ge.1 .and. iyshift.ge.1 .and. &
              ixshift.le.xbsize(j) .and. iyshift.le.ybsize(j)) then
              expj = max( 1.0, exp_image(ixshift, iyshift, j) )
          else
             expj = 1.0  !!! fake input for deltaa entries that should be never
                         !!! addressed because of maskzero
          endif
          if (randomflag .and. .not. scts(j,i) > 0.) then
             scts(j,i)=0.0
          endif
          if (scts(j,i)<=0. .and. .not. randomflag) then 
!            scts(j,i)=100./expj
!           try 10 counts (04/05/06)
             scts(j,i)=10./expj
          else if (.not. scts(j,i) > 0.) then
!           catch NULL values from input list           
            scts(j,i)=0.0
          endif
          a(ioff+3+j)=scts(j,i)
          deltaa(ioff+3+j)=max(sqrt(scts(j,i)*expj)/expj,1.D-5)
          deltae(ioff+3+j)=deltaa(ioff+3+j)
          sc=sc+a(ioff+3+j)*expj
        end do
!       assume 10 counts, if all input values were 0 or NULL:  
        sc=max(sc,10.)
        a(ioff+1)=xms(i)
        deltaa(ioff+1)=2.0/sqrt(sc)
        deltae(ioff+1)=deltaa(ioff+1)
        a(ioff+2)=yms(i)
        deltaa(ioff+2)=2.0/sqrt(sc)
        deltae(ioff+2)=deltaa(ioff+2)
!   use larger delta for nmulsou>1:
        if (nsmul>1) then
          deltaa(ioff+1)=1.0
          deltaa(ioff+2)=1.0
        endif
!   different options to initialize extent parameter
        if (input_ext) then
          ! use extent from input list
          a(ioff+3)=sigma(i) 
        else
          if (extfla) then 
            a(ioff+3)=5.0
          else
            a(ioff+3)=0.
          endif
        endif

        deltaa(ioff+3)=1.0 
!        (use counts dependent stepsizes for felsu, GL 29/06/06):
        deltae(ioff+3)=10.0/sqrt(sc)
      end do
      !end loop over nsrc


    done = .false.
    do while (.not. done .and. fitflag)


!    write(*,*) "nspar,npsou,np:",nspar,npsou,np

!-----------------------------------------------------------

      iordr(1)=0
      call ordr(np,np,iordr) 
      iordr0=iordr                                          

      nffix=0
      nspar=3
      ntermf=np
! optionally fix sigmas in fit
      if (.not. extfla) then
        do i=1,nsrc
          ioff=npsou*(i-1)
          call ordr(np,ioff+3,iordr)
        end do
        ntermf=ntermf-nsrc
        ntfix=nsrc
        if (.not. xidfix) nspar=nspar-1
      end if

! optionally fix positions in fit
      if (.not. posfla) then
        do n=1,nsrc
          ioff=npsou*(n-1)
          call ordr(ntermf,ioff+1,iordr)
          call ordr(ntermf,ioff+2,iordr)
        enddo
        ntermf=ntermf-2*nsrc
        ntfix=ntfix+2*nsrc
        if (.not. xidfix) nspar=nspar-2
      endif

!  set number of input images nband (06/11/05) 
      nband = sum(num_image(1:num_instr))

! fix count rate at zero if mask in entire source region is zero:
 

!    set flux to zero, if maskzero=1    
      do n=1,nsrc
        ioff=npsou*(n-1)
        do j=1,nim         
          if (maskzero(n,j) == 1) then
            a(ioff+3+j)=0.0
            deltaa(ioff+3+j)=0.0
            call ordr(ntermf,ioff+3+j,iordr)
            ntermf=ntermf-1
            ntfix=ntfix+1
          endif
        end do
      end do
      iordr0=iordr
      ntermf0=ntermf
      ntfix0=ntfix

!-------------------------------------------

!  write(*,*) "start, ntermf,ntfix:",ntermf,ntfix
!  write(*,*) "start, iordr:",iordr(1:6)
!  write(*,*) "start, maskzero:", maskzero(1,1:3)

!*********************************************************************************************

     if (xidfix) then
       ! simply fit fluxes for all sources: 
       call pordr(np,a,afit,iordr,1)
       call pordr(np,deltaa,delfit,iordr,1)
       call pordr(np,sigmaa,sigfit,iordr,1)
       do nlauf=1,50
         flamda=1e-2    
         call mlfit(ntermf,afit,delfit,sigfit,iordr,xms,yms,ecut2, &
                    flamda,yfit1,like0,lik_arr, nsrc, ntfix, ierr)
         if(abs(likold-like0).le.delt) then
           exit
         endif
         likold=like0
       end do
       call pordr(np,a,afit,iordr,-1)                                   
       call pordr(np,deltaa,delfit,iordr,-1)                            
       call pordr(np,sigmaa,sigfit,iordr,-1)
       likesave(0)=like0
       lik_arr_save(:,0)=lik_arr
       asave(:,0)=a
       bestfit=0 
!       write(*,*) "xidfix params:", a(1:6) 
     endif


!      call system_clock(clock1,clock2,clock3) 
!      write(*,*) "clock:",clock1-clock0 

     if (.not. xidfix) then
       ! calculate likelihood value for null hypothesis (all fluxes zero)
       atemp=a
       iordr(1)=0
       call ordr(np,np,iordr)
       atemp(3:npsou*(nsrc-1)+3:npsou)=0.0
       do i=1,nsrc
         ioff=npsou*(i-1)
!!!         atemp(ioff+3)=0.0
!!!         do j=1,nim
!!!           atemp(ioff+3+j)=0.0
!!!         enddo
         atemp(ioff+4:ioff+3+nim)=0.0
       end do

       call funct(atemp,iordr,yfit1,yfit2,yfit3,nsrc,np,xms,yms,ecut2)
       call like(like00,like00arr,yfit1,yfit2)
       if (testoutput) write(*,*) "like00:",like00
       iordr=iordr0
       !!!

       bestfit=0
       ! first point source fit: fix all sigmas at zero:
!       write(*,*) "Nsrc:",nsrc
       do i=1,nsrc
          ioff=npsou*(i-1)
          a(ioff+3)=0.0
          call ordr(np,ioff+3,iordr)
       end do
       if (extfla) then
         ntermf=ntermf-nsrc
         ntfix=ntfix+nsrc 
         nspar=nspar-1
       endif

!       write(*,*) "2 pnt iordr:", iordr(1:36)
!       write(*,*) "2 pnt ntermf,ntfix,nspar:",ntermf,ntfix,nspar
!      write(*,*) "iordr f"
       call pordr(np,a,afit,iordr,1)
       call pordr(np,deltaa,delfit,iordr,1)
       call pordr(np,sigmaa,sigfit,iordr,1)
         do nlauf=1,50
           flamda=1e-2    
           call mlfit(ntermf,afit,delfit,sigfit,iordr,xms,yms,ecut2, &
                                   flamda,yfit1,like0,lik_arr, nsrc, ntfix, ierr)
!           write(*,*) "likediff:",likold-like0
           if(abs(likold-like0).le.delt) then
             exit
           endif
           likold=like0
         end do
       call pordr(np,a,afit,iordr,-1)                                   
       call pordr(np,deltaa,delfit,iordr,-1)                            
       call pordr(np,sigmaa,sigfit,iordr,-1)
       likesave(0)=like0
       lik_arr_save(:,0)=lik_arr

       nzero = sum(maskzero(1:nsrc,1:max_ima))
       if (testoutput) write(*,*) "(like2dof) nsrc,nspar,nband,nzero:", nsrc,nspar,nband,nzero
       likediff=like00-like0
       if (testoutput) write(*,*) likediff
       like2dof(0)=gammq(real(nsrc*(nspar+nband)-nzero)*0.5,likediff)
       asave(:,0)=a
       if (testoutput) write(*,*) "2 pnt,nlauf, like0:",nlauf,like0
       if (testoutput) write(*,*) "2 pntsrc:a(1:9)", a(1:9)
!       write(*,*) "lik_arr(1:6)",lik_arr(1:6)
       if (testoutput) write(*,*) "2 pntsrc:a(19:36)", a(19:36)

!      call system_clock(clock1,clock2,clock3) 
!      write(*,*) "2 pnt clock:",clock1-clock0     
     
     endif
     ! end (.not. xidfix)
   
     if (extfla) then

       ! fit only one free sigma in each run
         ntermf=ntermf+1 
!         ntfix=nsrc-1
!        (GL 07/05/2007) changed to:
         ntfix=ntfix-1
         nspar=nspar+1
       do n=1,nsrc
          noff=npsou*(n-1)
!         iordr(1)=0
!         call ordr(np,np,iordr)
          iordr=iordr0
         !fix all ext values to 0, only fit ext(n)
         do nn=1,nsrc
           if (nn .ne. n) then
             ioff=npsou*(nn-1)
             a(ioff+3)=0.0
             call ordr(np,ioff+3,iordr)
           endif
         enddo  
!         write(*,*) "n,1 pnt iordr:", n,",",iordr(1:28)
!         write(*,*) "n,1 pnt ntermf,ntfix:",n,ntermf,ntfix
         call pordr(np,a,afit,iordr,1)
         call pordr(np,deltaa,delfit,iordr,1)
         call pordr(np,sigmaa,sigfit,iordr,1)
           do nlauf=1,50
             flamda=1e-2    
             call mlfit(ntermf,afit,delfit,sigfit,iordr,xms,yms,ecut2, &
                                     flamda,yfit1,like0,lik_arr, nsrc, ntfix, ierr)
             if(abs(likold-like0).le.delt) then
               exit
             endif
             likold=like0 
           end do
         call pordr(np,a,afit,iordr,-1)                                   
         call pordr(np,deltaa,delfit,iordr,-1)                            
         call pordr(np,sigmaa,sigfit,iordr,-1)
         if (testoutput) write(*,*) "nlauf b)",nlauf
         if (testoutput) write(*,'(3F8.3,15F10.7)') a(1:18)
         if (testoutput) write(*,'(3F8.3,15F10.7)') a(18:36)
!        call system_clock(clock1,clock2,clock3) 
!        write(*,*) "ext 1 clock:",clock1-clock0     
         if (testoutput) write(*,*) "n,a(3),a(21),like0 ",n,a(3),a(21),like0
         asave(:,n)=a
         likesave(n)=like0
         lik_arr_save(:,n)=lik_arr

         ext_ml(n)=likesave(0)-like0
         ext_ml(n)=gammq(1.5,ext_ml(n))
!         New computation of ext_ml and its correction for N_dof (GL 17/06/05):
          if (testoutput)  write(*,*) "old ext_ml:",ext_ml(n)
          likediff=like00-like0

          nzero = nzero+sum(maskzero(1:nsrc,1:max_ima))
          if (testoutput) then
            write(*,*) "corr. of ext_ml: ntermf,nffix,ntfix,nzero, nspar,nsrc:",ntermf,nffix,ntfix,nzero,nspar,nsrc
          endif
          ndof=ntermf
          likecorr=gammq(ndof/2.,likediff)
          ext_ml(n)=likecorr-like2dof(0)
          if (testoutput) write(*,*) "new definition ext_ml:",ext_ml(n)
          like2dof(n) = likecorr


 
         if (ext_ml(n) < dmlext .or. a(noff+3) < minext) then
           if (testoutput) write(*,*) "extent too small, ignore"
           ext_ml(n)=0
         endif

!!!!!!!!!  test: 09/06/2005 (GL): update  point source fit, if likelihood is better and  ext=0:
         if (like0 < likesave(0)  .and. a(noff+3) == 0.0) then
           asave(:,0)=a
           likesave(0)=like0
           lik_arr_save(:,0)=lik_arr

           nzero = sum(maskzero(1:nsrc,1:max_ima))
           if (testoutput) write(*,*) "(replace like2dof) nsrc,nspar-1,nband,nzero:", nsrc,nspar-1,nband,nzero
           likediff=like00-like0
           like2dof(0)=gammq(real(nsrc*(nspar-1+nband)-nzero)*0.5,likediff)
         endif 

         if (ext_ml(n) < dmlext .or. a(noff+3) < minext) then
           if (testoutput) write(*,*) "no extent"
           ext_ml(n)=0
           !(05/04/2005 GL:)
           a(noff+3)=0.0
           !!!!!
         else
           if (n == 1) then 
             bestfit=1
             if (testoutput) write(*,*) "bestfit=1, a:",a(1:3),a(19:21)
           else if (like0 < likesave(bestfit)) then
             bestfit=n
             if (testoutput) write(*,*) "bestfit=n, n,a:",n,a(1:3),a(19:21)
           endif
         endif
!         write(*,*) "ext likelihood:",likesave(n)
       enddo
!      (nsrc)
!      (05/04/2005 GL:)
       do n=1,nsrc
         noff=npsou*(n-1)
!         write(*,*) "ext:",asave(noff+3,n)
         if (asave(noff+3,bestfit) < minext) then
           ext_ml(n)=0.0
         endif
       enddo
     if (testoutput) write(*,*) "end ext:",nsrc,a(1),a(2),a(3),ext_ml(1), ext_ml(2) 
!!!!!!!!
     endif 
!    (extfla)
!     call system_clock(clock1,clock2,clock3) 
!     write(*,*) "end extent clock:",clock1-clock0     


      a=asave(:,bestfit)     
      ntermf=ntermf0
      ntfix=ntfix0
! now calculate the likelihood of each of the sources
      do n=nsrc,1,-1
        iordr=iordr0
        ioff=npsou*(n-1)
        multflag=.false.
        if(nsrc.eq.1)then
! simply calculate likelihood for one source only
!!!          do j=1,nim
!!!            a(3+j)=0.
!!!          end do
          a(4:nim+3) = 0.
          call pordr(np,a,afit,iordr,1)
          call funct(afit,iordr,yfit1,yfit2,yfit3,nsrc,np,xms,yms,ecut2)
          call pordr(np,a,afit,iordr,-1)
          call like(like0,lik_arr,yfit1,yfit2)
        else
! check, if 2 sources have same position and extent:
          if (posfla) then
            do l=n-1,1,-1
              ioff1=npsou*(l-1)
              distsq=((a(ioff+1)-a(ioff1+1))**2+(a(ioff+2)-a(ioff1+2))**2)
              extdiff=abs(a(ioff+1)-a(ioff1+1))
              if (distsq < mindist .and. extdiff < minextdiff) then
                multflag=.true.
!                write(*,*) "source #, multflag:",n,multflag
                exit 
              endif
            enddo
          endif 
           
! zero flux of relevant component
          nffix=0
          if (posfla) then
            call ordr(ntermf,ioff+1,iordr)
            call ordr(ntermf,ioff+2,iordr)
              nffix=nffix+2
          end if
!          if (extfla .and. ext_ml(n) >= dmlext) then
          ! changed 15/11/04 GL
          if (extfla) then
            call ordr(ntermf,ioff+3,iordr)
            nffix=nffix+1
          end if
          do j=1,nim
            a(ioff+3+j)=0.
!  only call ordr, if maskzero=0 (changed 09/07/04)
            if (maskzero(n,j) .eq. 0) then  
              call ordr(ntermf,ioff+3+j,iordr)
              nffix=nffix+1
            endif
          end do
! fix parameters 


          call pordr(np,a,afit,iordr,1)
          call pordr(np,deltaa,delfit,iordr,1)
          call pordr(np,sigmaa,sigfit,iordr,1)
            do nlauf=1,50
              flamda=1e-2 
              call mlfit(ntermf-nffix,afit,delfit,sigfit,iordr,xms,yms, &
                         ecut2,flamda,yfit1,like0,lik_arr,nsrc,ntfix+nffix,ierr)
              if(abs(likold-like0).le.delt) exit                   
              likold=like0                                                    
            end do
            if (testoutput) write(*,*) "c) nlauf,like0:",nlauf,like0
!          endif  
          call pordr(np,a,afit,iordr,-1)
          call pordr(np,deltaa,delfit,iordr,-1)
          call pordr(np,sigmaa,sigfit,iordr,-1)                           
        endif
!  write(*,*) "iordr n"
   
!      write(*,*) "parameters c)"
!      write(*,'(3F8.3,6F15.7)') a(1:9)
!      write(*,*) "lik_arr(1:6):",lik_arr(1:6)
!      write(*,'(3F8.3,10F15.7)') a(19:36)
!      call system_clock(clock1,clock2,clock3) 
!      write(*,*) "clock c:",clock1-clock0      

        !dml(n)=like0-likmerk
!!!        do j=1,nim 
!!!!old:
!!!!          exi_ml(j,n)=lik_arr(j)-lik_amerk(j)
!!!!new:
!!!           exi_ml(j,n)=lik_arr(j)-lik_arr_save(j,bestfit)
!!!!           write(*,*) "exi_ml(j,n) , 1:", exi_ml(j,n)
!!!        end do
        exi_ml(1:nim,n)=lik_arr(1:nim)-lik_arr_save(1:nim,bestfit)
!        fac=1.0

        nband=0
        dml(n)=0.         
        do j=1,num_instr
          nima=num_image(j)
          if (xid_arr(j)>0) nima=xid_arr(j)-1
!!!          do i=1,nima
!!!            ipoint=ip(j,i)
!!!            dml(n)=dml(n)+exi_ml(ipoint,n)
!!!            nband=nband+1
!!! !            fac=fac*nband
!!!          end do
          dml(n) = dml(n) + sum(exi_ml(ip(j,1:nima),n))
          nband = nband+nima
        end do

        nzero = sum(maskzero(n,1:max_ima))

        likediff=like00-like0
        if (testoutput) write(*,*) "correction like0 dof:",ntermf,nffix,ntfix,nzero
        if (testoutput) write(*,*) "corrected like0:",ntermf,nffix,gammq(real(ntermf-nffix)*0.5,likediff)
        if (testoutput) write(*,*) "corrected bestfit:",like2dof(bestfit)  
        

!        Calculate dml(n) from Ndof corrected likelihoods compared to like00:
         dml(n)=like2dof(bestfit)-gammq(real(ntermf-nffix)*0.5,likediff) 
         if (testoutput) write(*,*) "new dml(n):",n,dml(n)
!!!!{
! 20/06/05
!        write(*,*) "detml raw:",dml(n)
!        dml(n)=gammq(real(nspar+nband-nzero)/2.,dml(n))
!        write(*,*) "detml corr:",n, nspar, nband, nzero, dml(n)
!!!!!}
        noff=npsou*(n-1)

!        write(*,*) "a:",a(noff+1),a(noff+2),a(noff+3)
!        special case: decision between 2 pnt sources and one extended source:
!        do not use dmlmin threshold in order to avoid bias towards extended source:
         dmlthresh=dmlmin
        if (extfla .and. nsrc == 2 .and.  bestfit == 0 .and. n == 2 .and. a(3) > minext) then
          likediff=like00-like0
          if (testoutput) write(*,*) "correction 1:",ntermf,ntfix,nffix,nspar,nband,nzero
          likecorr=gammq(real(nspar+nband-nzero)*0.5,likediff)
          if (testoutput) write(*,*) "likecorr,like2dof(0):",likecorr,like2dof(0)
          if (likecorr < like2dof(0)) then
            dmlthresh=0.0
            if (testoutput) write(*,*) "set dmlthresh = 0, likecorr", likecorr
          endif
        endif
        if (extfla .and. nsrc == 2 .and.  bestfit == 0 .and. n == 1 .and. a(npsou+3) > minext) then
          likediff=like00-like0
          likecorr=gammq(real(nspar+nband-nzero)*0.5,likediff)
          if (testoutput) write(*,*) "likecorr,like2dof(0):",likecorr,like2dof(0)
          if (likecorr < like2dof(0)) then
            dmlthresh=0.0
            if (testoutput) write(*,*) "set dmlthresh = 0, likecorr", likecorr
          endif
        endif
          

        if ((dml(n).lt. dmlthresh .or. multflag) .and. .not. xidfix) then
             if (testoutput) write(*,*) " throw out component:"
             if (testoutput) write(*,*) "a(1),a(2),a(3)",a(1),a(2),a(3)
             if (testoutput) write(*,*) "a(19),a(20),a(21)",a(19),a(20),a(21)
             if (testoutput) write(*,*) "n,dml(n):",n,dml(n)
! throw out component
!!!            do l=n+1,nsrc
!!!              ioff0=npsou*(l-2)
!!!              ioff1=npsou*(l-1)
!!!              do j=1,3
!!!                a(ioff0+j)=a(ioff1+j)
!!!              end do          
!!!              do j=1,nim
!!!                a(ioff0+3+j)=a(ioff1+3+j)
!!!                exi_ml(j,l-1)=exi_ml(j,l)
!!!              enddo
!!!              dml(l-1)=dml(l)
!!!              ext_ml(l-1)=ext_ml(l)
!!!              insrc(l-1)=insrc(l)
!!!            enddo
            do l=n+1,nsrc
              ioff0=npsou*(l-2)
              ioff1=npsou*(l-1)
              a(ioff0+1:ioff0+3)=a(ioff1+1:ioff1+3)
              a(ioff0+4:ioff0+3+nim)=a(ioff1+4:ioff1+3+nim)
            enddo
            exi_ml(1:nim,n:nsrc-1)=exi_ml(1:nim,n+1:nsrc)
            dml(n:nsrc-1)=dml(n+1:nsrc)
            ext_ml(n:nsrc-1)=ext_ml(n+1:nsrc)
            insrc(n:nsrc-1)=insrc(n+1:nsrc)
!!!
            nsrc=nsrc-1
            if(nsrc.eq.0)then
!              call system_clock(clock1,clock2,clock3)
!              write(*,*) "clock mulsou return1:", clock1-clock0
!               write(*,*) "return 1"
              return
            endif
          np=nsrc*npsou
          exit
        else  

! get back best-fit parameters
!          write(*,*) "get params from bestfit", bestfit
!!!          do m=1,np
!!! !            a(m)=amerk(m)
!!!             a(m)=asave(m,bestfit)
!!!          enddo
           a(1:np) = asave(1:np,bestfit)
!          like0=likmerk
           like0=likesave(bestfit)
!          calculate final detection likelihoods for source (02/12/04, GL):
!          zero fluxes of source n:
           ioff=npsou*(n-1)
!!!           do j=1,nim
!!!             a(ioff+3+j)=0.
!!!           end do
           a(ioff+4:ioff+3+nim) = 0.
           if (testoutput) write(*,*) "d) n",n,np
           if (testoutput) write(*,'(3F8.3,15F10.7)') a(1:18)
           if (testoutput) write(*,'(3F8.3,15F10.7)') a(19:36) 
                    

           call pordr(np,a,afit,iordr,1)
           call funct(afit,iordr,yfit1,yfit2,yfit3,nsrc,np,xms,yms,ecut2)
           call pordr(np,a,afit,iordr,-1)
           call like(like1,lik_arr,yfit1,yfit2)
           if (testoutput) write(*,*) "like1",like1
!!!           do j=1,nim 
!!!             exi_ml(j,n)=lik_arr(j)-lik_arr_save(j,bestfit)
!!!           end do
           exi_ml(1:nim,n) = lik_arr(1:nim)-lik_arr_save(1:nim,bestfit)
           nband=0
           dml(n)=0.         
           do j=1,num_instr
             nima=num_image(j)
             if (xid_arr(j)>0) nima=xid_arr(j)-1
!!!             do i=1,nima
!!!               ipoint=ip(j,i)
!!!               dml(n)=dml(n)+exi_ml(ipoint,n)
!!!               nband=nband+1
!!!             end do
             dml(n) = dml(n) + sum(exi_ml(ip(j,1:nima),n))
             nband = nband+nima
           end do
           if (extfla .and. ext_ml(n) == 0.0) then
             dml(n)=gammq(real(nspar-1+nband-nzero)*0.5,dml(n))
           else
             dml(n)=gammq(real(nspar+nband-nzero)*0.5,dml(n))
           endif
           if (testoutput) write(*,*) "d2)", n,nspar,nband,nzero,dml(n)
           do m=1,np
             a(m)=asave(m,bestfit)
           enddo           
          if (n == 1) done = .true.
        endif
      enddo
     end do 

!   write(*,*) "final params:" 
!   write(*,'(3F8.3,15F10.7)') a(1:18)
!   write(*,'(3F8.3,15F10.7)') a(19:36) 

!     write(*,*) "dml(1),dml(2):",dml(1),dml(2)

!  assign output parameter values before error loop (GL 12/11/2004):
 
!!!   do i=1,nsrc
!!!    ioff=npsou*(i-1)
!!!    xms(i)=a(ioff+1)
!!!    yms(i)=a(ioff+2)
!!!    sigma(i)=a(ioff+3)
!!!    do j=1,nim
!!!      scts(j,i)=a(ioff+3+j)
!!!    end do
!!!   enddo  
   xms(1:nsrc) = a(1:npsou*(nsrc-1)+1:npsou)
   yms(1:nsrc) = a(2:npsou*(nsrc-1)+2:npsou)
   sigma(1:nsrc) = a(3:npsou*(nsrc-1)+3:npsou)
   do i=1,nsrc
      ioff=npsou*(i-1)
      scts(1:nim,i)=a(ioff+4:ioff+3+nim)
   enddo


  
  ! second stage will be run, return to main program:
!  write(*,*) "first, ext:",first,ext_ml(1:nsrc)
  if (first .and. maxval(ext_ml(1:nsrc)) .gt. 0.0) then
!    call system_clock(clock1,clock2,clock3)
!    write(*,*) "clock mulsou 2:", clock1-clock0    
    return
  endif

!    call system_clock(clock1,clock2,clock3)
!    write(*,*) "clock mulsou end fitting:", clock1-clock0   


! call "model" to calculate PSF weighted mask fraction:
      psffrac=0.0
      iordr(1)=0
      call ordr(np,np,iordr)
      call pordr(np,a,afit,iordr,1)
      ! call funct with scts set to 1.0:
      do i=1,nsrc
       ioff=npsou*(i-1)
!!!        do j=1,nim
!!!         afit(ioff+3+j)=1.0
!!!        end do
       afit(ioff+4:ioff+3+nim)=1.0
      enddo        
      call funct(afit,iordr,yfit1,yfit2,yfit3,nsrc,np,xms,yms,ecut2)
      call model(yfit1,yfit2,yfit3,xms,yms,sigma,scts,nsrc,ecut2,.true.) 


! throw out sources with PSF weighted mask fraction psffrac < psfthreshold
  do n=nsrc,1,-1
    nn=0
    ioff=npsou*(i-1)
    do i=1,num_instr
      ! check for psffrac values below the threshold:
      maskbad=.FALSE.
      do j=1,num_image(i)
        ipoint=ip(i,j)
        if (psffrac(n,ipoint) < psfthreshold .and. .not. xidfix) maskbad=.TRUE.
      enddo
      do j=1,num_image(i)   
        ipoint=ip(i,j) 
        if (.not. maskbad) then
          nn=nn+1  
        else
          ! throw out source, if psffrac < psfthreshold in all energy bands:
          scts(ipoint,n)=0.0
          exi_ml(ipoint,n)=0.0
          if  (maskzero(n,ipoint) == 0) then 
            nzero=nzero+1 
            maskzero(n,ipoint)=1
          endif
        endif
      end do
    end do
!   check likelihood once more:


    dml(n)=0.0
    do j=1,nim
      if (maskzero(n,j) == 0) dml(n)=dml(n)+exi_ml(j,n)
    enddo
    if (extfla .and. ext_ml(n) == 0.0) then
      dml(n)=gammq(real(nspar-1+nband-nzero)*0.5,dml(n))
    else
      dml(n)=gammq(real(nspar+nband-nzero)*0.5,dml(n))
    endif
!    write(*,*) xms(1),yms(1),nn
    if ((nn == 0 .or. dml(n) .lt. dmlmin) .and. .not. xidfix) then
    ! throw out component:
!!!       do l=n+1,nsrc
!!!         ioff0=npsou*(l-2)
!!!         ioff1=npsou*(l-1)
!!!         do j=1,3
!!!           a(ioff0+j)=a(ioff1+j)
!!!         end do
!!!         xms(l-1)=xms(l)
!!!         yms(l-1)=yms(l)
!!!         sigma(l-1)=sigma(l)          
!!!         do j=1,nim
!!!           a(ioff0+3+j)=a(ioff1+3+j)
!!!           exi_ml(j,l-1)=exi_ml(j,l)
!!!           scts(j,l-1)=scts(j,l)
!!!           ! also copy psffrac and maskzero (GL 02/08/06): 
!!!           psffrac(l-1,j)=psffrac(l,j)
!!!           maskzero(l-1,j)=maskzero(l,j)
!!!         enddo
!!!         dml(l-1)=dml(l)
!!!         ext_ml(l-1)=ext_ml(l)
!!!         insrc(l-1)=insrc(l)
!!!       enddo
       do l=n+1,nsrc
         ioff0=npsou*(l-2)
         ioff1=npsou*(l-1)
         a(ioff0+1:ioff0+3) = a(ioff1+1:ioff1+3)
         a(ioff0+4:ioff0+3+nim)=a(ioff1+4:ioff1+3+nim)
       enddo
       xms(n:nsrc-1)=xms(n+1:nsrc)
       yms(n:nsrc-1)=yms(n+1:nsrc)
       sigma(n:nsrc-1)=sigma(n+1:nsrc)          
       exi_ml(1:nim,n:nsrc-1)=exi_ml(1:nim,n+1:nsrc)
       scts(1:nim,n:nsrc-1)=scts(1:nim,n+1:nsrc)
       psffrac(n:nsrc-1,1:nim)=psffrac(n+1:nsrc,1:nim)
       maskzero(n:nsrc-1,1:nim)=maskzero(n+1:nsrc,1:nim)
       dml(n:nsrc-1)=dml(n+1:nsrc)
       ext_ml(n:nsrc-1)=ext_ml(n+1:nsrc)
       insrc(n:nsrc-1)=insrc(n+1:nsrc)
!!!
       nsrc=nsrc-1
       if (nsrc.eq.0) return       
    endif
  enddo



! error loop begins   
      if (errfla) then

        iordr(1)=0
        call ordr(np,np,iordr)                                           
 
! now position errors for position fixing
!       npsou=nim+nspar
!        write(*,*) "calc. errors" 
        do i=1,nsrc
          do j=1,npsou
             if (j >= 4) then
               ipoint=j-3 
             else
               ipoint=1
             endif 
             if (j <= 2 .and. posfla .or. &
                j == 3 .and. extfla .and. ext_ml(i)>=dmlext .or. &
                j >= 4 .and.  maskzero(i,ipoint) == 0) then

              ifel=npsou*(i-1)+j

              call ordr(np,ifel,iordr)

              flamda=1.0e-2
              call pordr(np,a,afit,iordr,1)
! test smaller delta values for pos parameters (22/06/06)
!              call pordr(np,deltaa,delfit,iordr,1)
              call pordr(np,deltae,delfit,iordr,1)
              call pordr(np,sigmaa,sigfit,iordr,1)
              call felsu(np,afit,delfit,sigfit,iordr,xms,yms, &
                         flamda,nsrc,like0,fello,felhi,np,ecut2)
              call pordr(np,a,afit,iordr,-1)
!              call pordr(np,deltaa,delfit,iordr,-1)
              call pordr(np,deltae,delfit,iordr,-1)
              call pordr(np,sigmaa,sigfit,iordr,-1)
              if (fello .le. 0) then
                ! if lower error is set to 0, use felhi as error
                fe(ifel)=max(felhi,0.)
              else                
                fe(ifel)=0.5*(max(fello,0.)+max(felhi,0.))
              endif
            end if
          end do
        end do
      
      else
        fe = 0.
      end if

                                                                     
! assign output values
      exms(1:nsrc) = fe(1:npsou*(nsrc-1)+1:npsou)
      eyms(1:nsrc) = fe(2:npsou*(nsrc-1)+2:npsou)
      where( sigma(1:nsrc) > 0 )
         esigma(1:nsrc) = fe(3:npsou*(nsrc-1)+3:npsou)
      elsewhere
         esigma(1:nsrc) = 0.
      endwhere
      eradec(1:nsrc)  = sqrt(exms(1:nsrc)**2+eyms(1:nsrc)**2) * pixim

      do i=1,nsrc
        ioff=npsou*(i-1)
!!!!        xms(i)=a(ioff+1)
!!!        exms(i)=fe(ioff+1)
!!!!        yms(i)=a(ioff+2)
!!!        eyms(i)=fe(ioff+2)
!!!!        sigma(i)=a(ioff+3)
!!!        if (sigma(i) .gt. 0.) then
!!!          esigma(i)=fe(ioff+3)
!!!        else
!!!          esigma(i)=0.
!!!        end if 
!!!        do j=1,nim
!!!!          scts(j,i)=a(ioff+3+j)
!!!!         set counts error to 0.0, if source was thrown out due to psffrac limit (GL 27/02/06) 
!!!          if (maskzero(i,j) == 0) then
!!!            escts(j,i)=fe(ioff+3+j)
!!!          else
!!!            escts(j,i)=0.0
!!!          endif
!!!        end do
        where (maskzero(i,1:nim) == 0)
           escts(1:nim,i) = fe(ioff+4:ioff+3+nim)
        elsewhere
           escts(1:nim,i) = 0.0
        endwhere
  
        cos_pa = dcos(crota2 * pi180)
        sin_pa = dsin(crota2 * pi180)
        xr = (xms(i)-crpix1) * crdelt1 * pi180
        yr = (yms(i)-crpix2) * crdelt2 * pi180
        dra =  xr * cos_pa - yr * sin_pa
        ddec = xr * sin_pa + yr * cos_pa
        r0r = crval1 * pi180
        d0r = crval2 * pi180
        call antisp(dra,ddec,r0r,d0r,rr,dr)
        ra(i) = rr * overpi180
        dec(i) = dr *overpi180
!!!        eradec(i)  = sqrt(exms(i)**2+eyms(i)**2) * pixim



        call sla_EQGAL(rr,dr,lr,br)
        lii(i) = lr * overpi180
        bii(i) = br * overpi180

      enddo

!      write(*,*) "final coords:"
!       write(*,'(4F8.3)') xms(1),yms(1),xms(2),yms(2)



! get raw coordinates
      call rawcoords(ra,dec,nsrc,datetime, &
                     chipix, chipiy, ccdinr, offax, azim)
!---------- JR; testing ---------------------------------
!      write(*,*)ra(1),dec(1),offax(1,1),azim(1,1)
!--------------------------------------------------------
!      write(*,*) "offax:",offax(1:3,1)
      srcontime=-99.9
      do n=1,nsrc
        do i=1,num_instr
          do j=1,num_image(i)
            ipoint=ip(i,j)
!!!            l=expip(i)
             if (ccdinr(i,n) > 0 .and. maskzero(n,ipoint) == 0) then 
               srcontime(ipoint,n)=ontime(ccdinr(i,n),ipoint)
             else
               srcontime(ipoint,n)=-99.9
             endif
          enddo
        enddo
      enddo   

  
vignetting(:,:)=0.0
effarea_cen=0.0
effarea_src=0.0
! calculate vignetting:            (accuracy level shouldn't be relevant)
      do n=1,nsrc
         do i=1,num_instr
            !!! IT 2014/08: switched to select-case syntax
            select case (expip(i))
            case (2)
               call CAL_setState1(instrument=EMOS1)
            case (3)
               call CAL_setState1(instrument=EMOS2)
            case default ! (1)
               call CAL_setState1(instrument=EPN)
            end select
            do j=1,num_image(i)
               ipoint=ip(i,j)
               ene=energy(ipoint)
!           effarea_cen=CAL_getEffectiveArea(dble(ene),0d0,0d0)
!           write(*,*) "effarea:",effarea_cen
               vignetting(ipoint,n)=CAL_getVignettingFactor(dble(ene),offax(i,n),azim(i,n))
            enddo
         enddo
     enddo


! update background map
!      write(*,*) "bg_map, nsrc:", nsrc
      if (testoutput) write(*,*) "bg_map, src1:",xms(1),yms(1),sigma(1),ext_ml(1),scts(1:5,1)
      if (testoutput) write(*,*) "bg_map, src2:",xms(1),yms(2),sigma(2),ext_ml(2),scts(1:5,2)


       l=1
      do i=1,num_instr
                                     !!! (instead of point_resp)
        if (expip(i) .gt. 4) call error("WrongPSF","Wrong PSF number") 

        do j=1,num_image(i)
          ipoint=ip(i,j)
!          ene=energy(ipoint)
          do n=1,nsrc
            ! avoid extended source in bg map, where pnt source is written to source list (15/10/04):
            if (ext_ml(n) .eq. 0.0 .and. .not. (xidfix .or. randomflag)) sigma(n) = 0 
!!! IT 2012/01: equivalent to int(xms(n)), int(yms(n)), i.e. "floor"
!            ix0=xms(n)
!            iy0=yms(n)
            ix0=int(xms(n))
            iy0=int(yms(n))

            exp=0.
            expn=0. 
!!-- test JR --------------------------
!            testsum =0.
!!-------------------------------------           
            ixlow = max(ix0-irad, ixb0(ipoint)+1)    ! image coordinates
            iylow = max(iy0-irad, iyb0(ipoint)+1)
            ixhig = min(ix0+irad, ixb1(ipoint)+1)
            iyhig = min(iy0+irad, iyb1(ipoint)+1)
            iyshift = iylow - iyb0(ipoint)   ! subimage coordinates
            do iy=iylow, iyhig
               y1=iy     ! floating point
               ytmp0 = iy-iy0
               ytmp0 = ytmp0 * ytmp0
               ytmp1 = y1-yms(n)
               ytmp1 = ytmp1 * ytmp1
               ixshift = ixlow - ixb0(ipoint)
               do ix=ixlow, ixhig
               xtmp0 = ix-ix0
               xtmp0 = xtmp0*xtmp0
               !!! outer circle
               if (xtmp0+ytmp0<irad2) then
                  x1=ix
                  !!! inner circle
                  if ((x1-xms(n))**2+ytmp1<ical2) then
!!! obsolete          if (point_resp(i) .ge. 1 .and. point_resp(i) .le. 4) then
!!! IT 2012/02
!!! Replaced former psf1, psf2, psf3 routines by a single "bpsf" ("background"):
!!! same "bpsf" call for pn, MOS1, and MOS2 - bpsf checks for the instrument.
!!! Case "point_resp(i) .eq. 4" - calling pn PSF via psf1 in the former code
!! version - is now handled within the bpsf routine.
!!!----->
                       df = bpsf(energy,ipoint,ix,iy, xms(n),yms(n), &
                            sigma(n),rcal,offax,azim,n)
                       bg_image(ixshift,iyshift,ipoint) = &
                            bg_image(ixshift,iyshift,ipoint) + &
                         scts(ipoint,n) * df * &
                         exp_image(ixshift,iyshift,ipoint)
!!! IT/GL 2012/02 Calculate exposure (for averaging) only within ecut radius
                       if (xtmp0+ytmp0<ecut2) then
                          exp = exp + df * &
                               exp_image(ixshift,iyshift,ipoint)
                          expn = expn + df
                       endif
!!!<-----
!!! obsolete                     else if (point_resp(i) .gt. 4) then
!!! obsolete                        call error("WrongPSF","Wrong PSF number") 
!!! obsolete                     endif 
!!!                    select case (point_resp(i))       
!!!                      case(1)
!!!                      bg_image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint) = &        !(JR)
!!!                         bg_image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint) + &     !(JR)
!!!                         scts(ipoint,n) * &
!!!                         psf1(energy,ipoint,ix,iy, &
!!!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!!!                         sigma(n),rcal,offax,azim,n) * &
!!!                         exp_image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint)        !(JR)
!!!                       exp = exp + & 
!!!                         psf1(energy,ipoint,ix,iy, &
!!!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!!!                         sigma(n),rcal,offax,azim,n) * &
!!!                         exp_image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint)        !(JR)
!!!                       expn = expn + &
!!!                         psf1(energy,ipoint,ix,iy, &
!!!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!!!                         sigma(n),rcal,offax,azim,n)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!               
!!!                  else 
!!!                   bg_image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint) = &
!!!                   bg_image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint) + &
!!!                       scts(ipoint,n)*psf4(energy,j,x1,y1, &
!!!                       xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!!!                       xcen(i),ycen(i),sigma(n),rcal,offax(i,n)) * &
!!!                       exp_image(ixb(ix,ipoint),iyb(iy,ipoint),ipoint)
                  !!! outer annulus
                  else 
                   bg_image(ixshift,iyshift,ipoint) = &
                   bg_image(ixshift,iyshift,ipoint) + &
                       scts(ipoint,n)*&
                       psf4(x1,y1,xms(n),yms(n), &
                            xcen(i),ycen(i),sigma(n),offax(i,n)) * &
                       exp_image(ixshift,iyshift,ipoint)
                  end if     !!! "if ((x1-xms(n))**2+(y1-yms(n))**2<ical**2)"
!!!                end if       !!! "if (ix>=1.and.ix<=npix(1).and. [...]"
               end if        !!! "if ((ix-ix0)**2+(iy-iy0)**2<irad**2)"
               ixshift = ixshift+1
              end do         !!! "do ix=ix0-irad,ix0+irad"
               iyshift = iyshift+1
            end do           !!! "do iy=iy0-irad,iy0+irad"
!!-- test JR -------------------------------------------------------------------------------
!            write(*,*) "Normalization for source:", n , testsum
!!------------------------------------------------------------------------------------------
            rate(ipoint,n)=real(scts(ipoint,n))
            erate(ipoint,n)=escts(ipoint,n)
            scts(ipoint,n)=scts(ipoint,n)*exp/expn     
            escts(ipoint,n)=escts(ipoint,n)*exp/expn
            srcexpos(ipoint,n)=0.0
            if (expn > 0.0) srcexpos(ipoint,n)=exp/expn
!            write(*,*) "bgimage",srcexpos(ipoint,n)
          end do 
         end do
      end do

!   copy fit parameters to single precision output parameters
      sxms(1:max_fit)=sngl(xms(1:max_fit))
      syms(1:max_fit)=sngl(yms(1:max_fit))
      ssigma(1:max_fit)=sngl(sigma(1:max_fit))
      sscts(1:max_ima,1:max_fit) = sngl(scts(1:max_ima,1:max_fit))

   


!  call system_clock(clock1,clock2,clock3)
!  write(*,*) "clock end mulsou :", clock1-clock0



!      write(*,*) "count_model:", count_model
!       write(*,*) " count_psf:", count_psf                                    
      end subroutine mulsou   



  !----------------------------------------------------------------------
  subroutine ordr(nterms,nfix,iordr)
  !----------------------------------------------------------------------

    integer     i,iordr(max_par),io(max_par),nfix,nnfix,nterms

!!!    if(iordr(1).eq.0)then
!!!      do i=1,max_par
!!!        iordr(i)=i
!!!      end do
!!!    endif
!!!
!!!    do i=1,max_par
!!!      if(iordr(i).eq.nfix) exit
!!!      io(i)=iordr(i)
!!!    end do
!!!
!!!    nnfix=i
!!!
!!!    do i=nnfix,nterms-1
!!!      io(i)=iordr(i+1)
!!!    end do
!!!
!!!    io(nterms)=nfix
!!!
!!!    if(nterms.ne.max_par) then
!!!      do i=nterms+1,max_par
!!!        io(i)=iordr(i)
!!!      end do
!!!    end if 
!!!
!!!    iordr(1:max_par)=io(1:max_par)


    if(iordr(1).eq.0) iordr = (/ (i,i=1,max_par,1) /)

    nnfix = minloc(abs(iordr-nfix),1)

    io(1:nnfix-1)=iordr(1:nnfix-1)
    io(nnfix:nterms-1) = iordr(nnfix+1:nterms)
    io(nterms)=nfix
    io(nterms+1:max_par)=iordr(nterms+1:max_par)

    iordr(1:max_par)=io(1:max_par)


  end subroutine ordr

  !----------------------------------------------------------------------
  subroutine pordr(nterms,a,b,iordr,mode)
  !----------------------------------------------------------------------

    real(kind=double) :: a(max_par),b(max_par)
    integer       mode,nterms,iordr(max_par)  !, i

    if(mode.eq.1) then
       b(1:nterms) = a(iordr(1:nterms))
    else

! andere richtung
       a(iordr(1:nterms)) = b(1:nterms)
    end if

  end subroutine pordr

  !----------------------------------------------------------------------
  function cage(array,n,vallo,valhi)
  !----------------------------------------------------------------------

    real(kind=double) ::        array(max_par),cage,value
    real ::                     vallo,valhi
    integer ::     ierr,n 

    ierr=0
!
! used to cage fit parameters to reasonable values
    value=array(n)

    if (value.lt.vallo) then
      array(n)=vallo
!      write(6,*)' allowed values ',vallo,valhi
    else if (value.gt.valhi) then
      array(n)=valhi
!      write(6,*)' allowed values ',vallo,valhi
    endif
    cage=array(n)

  end function cage    

!!! IT 2012/02 - replaced      
!!!  !----------------------------------------------------------------------
!!!  subroutine cagexy (x1,y1,x0,y0,rcut2)
!!!  !----------------------------------------------------------------------
!!!
!!!    real(kind=double) :: x0,y0,x1,y1
!!!    real     rcut2  
!!!    real     ds
!!!
!!!    ds=sqrt(((x1-x0)**2+(y1-y0)**2)/rcut2)
!!!    if (ds.gt.1.) then
!!!      x1 = x0+(x1-x0)/ds
!!!      y1 = y0+(y1-y0)/ds
!!!    end if
!!! 
!!!  end subroutine cagexy

  !----------------------------------------------------------------------
  subroutine like(lik,lik_arr,yfit1,yfit2)
  !----------------------------------------------------------------------

    real, dimension(npts), intent(in) :: yfit1
    real, dimension(nim), intent(in) :: yfit2

    real(kind=double), dimension(nim), intent(out) :: lik_arr
    real(kind=double), intent(out) :: lik

    integer :: k,ipoint 

! accumulate statistics
    lik_arr(1:nim)=0.
    !loop over image pixels
    do k=1,npts
      ipoint=events(k)%ipt
      lik_arr(ipoint)=lik_arr(ipoint)+events(k)%n*log(max(yfit1(k),1e-30)) &
!!! try to prevent sources running into negative fluxes: 
!test      if (yfit1(k) < 0.0) lik_arr(ipoint)=lik_arr(ipoint)-1.
!!!!!!! 
           - 0.5*(1.-sign(1.,yfit1(k)))
    end do

!!!!!!!    lik_arr(1:nim)=lik_arr(1:nim)-yfit2(1:nim)
!!! (below)    lik_arr(1:nim)=lik_arr(1:nim)-max(yfit2(1:nim),0.0)

!    lik_arr(1:nim)=-2.*lik_arr(1:nim)
!    remove factor 2 from ML => ML=-ln(P)
!!! (below)     lik_arr(1:nim)=-lik_arr(1:nim)

     lik_arr(1:nim) = max(yfit2(1:nim),0.0)-lik_arr(1:nim)

     lik = sum(lik_arr(1:nim))

   end subroutine like



  !----------------------------------------------------------------------
  subroutine mlfit(nterms,a,deltaa,sigmaa,iordr,xms,yms,ecut2,flamda, &
                   yfit1,like0,lik_arr,is,nfel,ierr)
  !----------------------------------------------------------------------

    integer :: i,j,k   !,minmethod !!! obsolete
    integer :: ntermf,nterms,nfree,nfel,is,ierr,iordr(max_par)
    real(kind=double) :: det
    real(kind=double) :: a(max_par),deltaa(max_par),sigmaa(max_par),b(max_par)
    real(kind=double) :: xms(max_fit),yms(max_fit)
    real :: yfit1(npts),ytemp1(npts), tmpval
    real :: yfit2(nim),yfit3,ytemp3
    real(kind=double) :: lik_arr(nim),like0,like1,like2,like3, like12
    !! real :: derv1(max_par,max_events),derv3(max_par)
    real :: derv3(max_par),testsig
    real, allocatable :: derv1(:,:)
    real :: aj,ak,delta,delta2,deltak,ecut2,flamda
    real(kind=double) ::array(max_par,max_par),alpha(max_par,max_par), &
                        beta(max_par)      

    real, parameter :: small = 1e-6
!    real :: test1,test2

!   small=tiny(small)  

    !allocate(derv1(max_par,max_events))
    allocate(derv1(nterms,npts))

    ntermf=nterms+nfel
    nfree=npts-nterms
    if (nfree.le.0) &
      call error("FitFailed","degrees of freedom less or equal 0")




!!!!!!!!!! insert 9/12/2003 !!!
!!!meanwhile obsolete.    minmethod = 2
!!!meanwhile obsolete.
!!!meanwhile obsolete.
!!!meanwhile obsolete.    if (minmethod .eq. 1) then 
!!!meanwhile obsolete.
!!!meanwhile obsolete.       testsig=1.0
!!!meanwhile obsolete.
!!!meanwhile obsolete.       beta(1:nterms)=0.
!!!meanwhile obsolete.       alpha(1:nterms,1:nterms)=0. 
!!!meanwhile obsolete.
!!!meanwhile obsolete.       call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
!!!meanwhile obsolete.       call like(like0,lik_arr,yfit1,yfit2)        
!!!meanwhile obsolete.       do j=1,nterms
!!!meanwhile obsolete.         aj=a(j)
!!!meanwhile obsolete.         delta=deltaa(j)
!!!meanwhile obsolete.         delta2=2.*deltaa(j) 
!!!meanwhile obsolete.         a(j)=aj+delta 
!!!meanwhile obsolete.         call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
!!!meanwhile obsolete.         call like(like1,lik_arr,yfit1,yfit2)
!!!meanwhile obsolete.         a(j)=aj-delta 
!!!meanwhile obsolete.         call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
!!!meanwhile obsolete.         call like(like2,lik_arr,yfit1,yfit2)
!!!meanwhile obsolete.         like12=like1-like2
!!!meanwhile obsolete.         a(j)=aj      
!!!meanwhile obsolete.         beta(j)=testsig*(-1.)*like12/delta2 
!!!meanwhile obsolete.         alpha(j,j)=testsig*(like12-2.*like1+like0)/(delta*delta)
!!!meanwhile obsolete.   ! derivatives del^2(like)/(del_ak del_aj) = &
!!!meanwhile obsolete.   !    ( like(aj + daj,ak+dak) - like(aj + daj,ak) - like(aj, ak+dak) + like(aj,ak) )/(daj*daj)
!!!meanwhile obsolete.         do k=1,j-j 
!!!meanwhile obsolete.            ak=a(k)
!!!meanwhile obsolete.            deltak=5*deltaa(k)
!!!meanwhile obsolete.            a(k)=ak+deltak
!!!meanwhile obsolete.            a(j)=aj+delta
!!!meanwhile obsolete.            call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
!!!meanwhile obsolete.            call like(like2,lik_arr,yfit1,yfit2)
!!!meanwhile obsolete.            a(j)=aj
!!!meanwhile obsolete.            call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
!!!meanwhile obsolete.            call like(like3,lik_arr,yfit1,yfit2)
!!!meanwhile obsolete.            alpha(j,k)=testsig*(like2-like1-like3+like0)/(delta*deltak)
!!!meanwhile obsolete.            a(k)=ak
!!!meanwhile obsolete.         end do
!!!meanwhile obsolete.         do  k=1,j-1
!!!meanwhile obsolete.           alpha(k,j)=alpha(j,k)
!!!meanwhile obsolete.         end do
!!!meanwhile obsolete.       end do
!!!meanwhile obsolete.     else 
!!!meanwhile obsolete.     !minmethod > 1

!!!!!!!!!! end insert 9/12/2003

!!!!!!! remove 9/12/2003 and replace with numerical alpha/ beta matrices"

    do j=1,nterms
      aj=a(j)

      delta=deltaa(j)
      delta2=2.0*delta

      a(j)=aj+delta

      call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)

      a(j)=aj-delta
      call funct(a,iordr,ytemp1,yfit2,ytemp3,is,ntermf,xms,yms,ecut2)

      derv1(j,1:npts)=(yfit1(1:npts)-ytemp1(1:npts))/delta2
      derv3(j)=(yfit3-ytemp3)/delta2
!      write(*,*) "j,derv1,derv3,delta2",j,derv1(j,1:npts),derv3(j),delta2
!      write(*,*) "a(1:7)",a(1:7) 

      a(j)=aj
    end do
!    
    beta(1:nterms)=0.
    alpha(1:nterms,1:nterms)=0.          
!
! write(*,*) "mlfit i)"
    call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
!
     do i=1,npts
        if (yfit1(i) .ne.  0.0) then
           tmpval = events(i)%n/yfit1(i)-1
           do j=1,nterms
!!!              beta(j)=beta(j)+(events(i)%n-yfit1(i))/yfit1(i)*derv1(j,i)
              beta(j)=beta(j)+tmpval*derv1(j,i)
              alpha(j,1:j)=alpha(j,1:j)+derv1(j,i)*derv1(1:j,i)/yfit1(i)
           end do
        end if
    end do

    do j=1,nterms
      do k=1,j
        alpha(k,j)=alpha(j,k)
      end do
    end do

!!!meanwhile obsolete.  endif

!!!! 9/12/2003
! write(*,*) "mlfit j)"
    call funct(a,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
    call like(like1,lik_arr,yfit1,yfit2)
! write(*,*) "mlfit j) done"
    do
      
      do j=1,nterms
        do k=1,nterms
          if(j.eq.k)then
            array(j,k)=1.0d0
          else 
            array(j,k)=alpha(j,k)/dsqrt(dabs(alpha(j,j))+small) &
                                 /dsqrt(dabs(alpha(k,k))+small)
          endif
        end do
        array(j,j)=1.0d0+flamda
      end do
      call mati(array,nterms,det)

!      b(1:nterms)=a(1:nterms)
      b = a
      
      do j=1,nterms
        do k=1,nterms
          b(j)=b(j)+beta(k)*array(j,k)/dsqrt(dabs(alpha(j,j))+small) &
                                      /dsqrt(dabs(alpha(k,k))+small)
!        write(*,*) "matrix:",j,k,beta(k),array(j,k),alpha(j,j),alpha(k,k)
        end do 
      end do

      if (.not. (beta(1) .gt. -1.E10)) then
!         write(*,*) array(1:nterms,1:nterms)
         call error("InternalError", "Unexpected value in fit.")   !!! IT 2017 gfortran debugging
         stop
      endif

! write(*,*) "k)"
!!!! test output write(*,*):
!      if (.not.  ((b(1) >= 0.0) .or. (b(1) < 0.0))) then
!         write(*,*) "b:", b(1:10)
!         write(*,*) "beta:", beta(1:10)
!      do j=1,nterms
!        do k=1,nterms
!          write(*,*) "array:",j,k,array(j,k)
!          write(*,*) "alpha:",j,j,alpha(j,j)
!         write(*,*) "alpha:",k,k,alpha(k,k)
!          write(*,*) "sqrt:", j,k, dsqrt(dabs(alpha(j,j))+small) * dsqrt(dabs(alpha(k,k))+small)
!          write(*,*) "derv1:",j,k,derv1(j,k),yfit1(j)
!        write(*,*) "matrix:",j,k,beta(k),array(j,k),alpha(j,j),alpha(k,k)
!        end do 
!      end do
!      stop
!      endif 
      

!!!!!!!!!!!!!!!!!! end test output

      call funct(b,iordr,yfit1,yfit2,yfit3,is,ntermf,xms,yms,ecut2)
      call like(like0,lik_arr,yfit1,yfit2)

      if (like0-like1.lt.0.) then
        exit
      else
        flamda=10.*flamda
        if (flamda.ge.1e10) exit
      end if

    end do

    a(1:nterms)=b(1:nterms)
    do j=1,nterms
      sigmaa(j)=dsqrt(dabs(array(j,j))/(dabs(alpha(j,j))+small))
    end do
    flamda=flamda*0.1

    deallocate (derv1)


 


  end subroutine mlfit

  !----------------------------------------------------------------------
  subroutine funct(afit,iordr,yfit1,yfit2,yfit3,is,np,xms,yms,ecut2)
  !------------------------------------------------------------------------
      
    real :: yfit1(npts),yfit2(nim),yfit3, &
            ecut2,ecut

    real(kind=double) :: a(max_par),afit(max_par), x0(max_fit),y0(max_fit), &
                         sig(max_fit),scts(max_ima,max_fit), &
                         xms(max_fit),yms(max_fit),ds

    integer :: i,n,iordr(max_par),is,np, &
               ndat,npars       !!!,nn,ierr

!!!    ierr = 0

    call pordr(np,a,afit,iordr,-1)                               

    ecut = sqrt(ecut2)

! cage parameters
    npars=nim+3

    x0(:) = 0.0
    y0(:) = 0.0

!!! IT 2012/02: replace cagexy (etc.) to avoid unnecessary calls
! (n-1)*npars+1, n=1..is -> 1..nn
!    nn = (is-1)*npars+1
!    ds(1:is) = sqrt( (a(1:nn:npars)-xms(1:is))**2 &
!                      + (a(2:nn+1:npars)-yms(1:is))**2 ) * 0.125
!    where ( ds(1:is) .gt. 1. )
!       a(1:nn:npars)   = xms(1:is)+(a(1:nn:npars)-xms(1:is))/ds(1:is)
!       a(2:nn+1:npars) = yms(1:is)+(a(2:nn+1:npars)-yms(1:is))/ds(1:is)
!    endwhere

!!!!    write(*,*) '*****a:',a(1:14)
    do n=1,is 
      ndat=(n-1)*npars+1
!! -- rcut2 too large!!.. changed to 64. (12/04/10 JR)
!!!      call cagexy (a(ndat),a(ndat+1),xms(n),yms(n),64.)
      ds = sqrt(((a(ndat)-xms(n))**2+(a(ndat+1)-yms(n))**2)/64.)
      if (ds.gt.1.) then
         a(ndat) = xms(n)+(a(ndat)-xms(n))/ds
         a(ndat+1) = yms(n)+(a(ndat+1)-yms(n))/ds
      end if
      x0(n)=a(ndat)
      y0(n)=a(ndat+1)
      ndat=(n-1)*npars+3
      sig(n)=cage(a,ndat,0.,maxext)
      do i=1,nim
        ndat=(n-1)*npars+i+3
!!!!!!!!!!!!!!!!!!!!!!!!
        if (negfla) then 
          scts(i,n)=cage(a,ndat,-1.0,1e10)
        else
           scts(i,n)=cage(a,ndat,0.0,1e10) 
        endif
!         if  (scts(i,n) .eq. 0) then
!           write(*,*)  i,n, scts(i,n), x0(n),y0(n)
!         endif
!         scts(i,n)=cage(a,ndat,-1.,1e10)
!!!!!!!!!!!!!!!!!!!!!!
      end do
    end do 


! calculate expectation value of model + background

    call model(yfit1,yfit2,yfit3,x0,y0,sig,scts,is,ecut,.false.)

    call pordr(np,a,afit,iordr,1)                                

  end subroutine funct
           
  !----------------------------------------------------------------------
  subroutine model(yfit1,yfit2,yfit3,x0,y0,sig,scts,is,ecut,last)
  !----------------------------------------------------------------------
    
   
    real, dimension(npts), intent(out) :: yfit1
    real, dimension(nim), intent(out) :: yfit2
    real, intent(out) :: yfit3
!    real, dimension(max_ima,max_fit,max_mask,max_mask), SAVE :: psf_save
!!! IT 2012/02: obsolete
!!!    real, dimension(:,:,:,:), allocatable, SAVE :: psf_save
    real, dimension(:,:,:,:), allocatable, SAVE :: ff_save
!!! IT/GL 2012/03 ff_save & modelima allocatable and smaller
!!!    real, dimension(max_mask,max_mask) :: modelima
    real, dimension(:,:), allocatable, SAVE :: modelima
    integer :: astat

    real(kind=double), dimension(max_fit),  intent(in) :: x0, y0, sig
    real(kind=double), dimension(max_ima,max_fit), intent(in) :: scts
!    logical, dimension(max_ima,max_fit) :: scts_ch 
!    logical, dimension(max_fit) ::  x0_ch, y0_ch, sig_ch

    real(kind=double), dimension(max_ima,max_fit), SAVE :: scts_save
    real(kind=double), dimension(max_ima,max_fit) :: scts_scale
    real(kind=double), dimension(max_fit), SAVE  :: x0_save, y0_save, sig_save
!    real, dimension(max_ima,max_fit), intent(out) :: detcounts, srcexpos
!    real :: src_rate
    real, intent(in) :: ecut
    integer, intent(in) :: is
!    integer, SAVE :: filenum=0
!    character*4 :: filestring    

    real :: totmodel,maskmodel
  
    real :: small
    real(kind=double) :: xremain,yremain
!    real :: ene,x1,y1,ene_old
    real :: s1,s2,s3,fudge !,stemp
    integer :: i,j,l,n,ix,iy,ix0,ix1,iy0,iy1,ixx,iyy,ipoint,tempsum,xx0,yy0   !,ii,jj,jpoint,ifound
    integer :: ixshift, iyshift    !!! IT 2016/12
    integer,SAVE :: isize  !!! IT 2014/07 - "SAVE" was missing
    integer :: isizehalf,ixs,iys,ix_mask,iy_mask,lf
    integer :: ixf1,ixf2,iyf1,iyf2,ixm1,ixm2,iym1,iym2,nxmask,nymask
    logical, dimension(max_ima,max_fit) :: scalepsf
    logical, save :: first = .true.
    logical :: last
    integer, save :: is_old
!    integer :: kk,istate1
!    character(len=5) :: istr
!    character(len=20) :: modfile
!    real, dimension(100) :: randnum

!!! IT/GL 2012/03 needed for array allocation
    real,    SAVE :: ecut_save, nxmask_save, nymask_save
    integer, SAVE :: xsize,ysize
    integer       :: xsizehalf,ysizehalf

!    write(*,*) "model:",num_instr,num_image(1),num_image(2),x0(1),y0(1),sig(1),scts(1:4,1)
 
!   call system_clock(clock0,clock2,clock3)


!   write(*,*) "model, scts(1:6):",scts(1:6,1)

!   write(*,*) "model is,num_instr,num_image(1:3):",is,num_instr,num_image(1:3)


!!!        do i=1,num_instr
!!!         do j=1,num_image(i)  
!!!           ipoint=ip(i,j)
!!!         enddo
!!!        enddo

!!!!! !! aperture coordinates
    ix0 = mask%ix_min
    ix1 = mask%ix_max 
    iy0 = mask%iy_min
    iy1 = mask%iy_max
    nxmask=ix1-ix0+1
    nymask=iy1-iy0+1 

   if (last) then
!    calculate fraction of source flux on valid mask pixels:
     do n=1,is
       do i=1,num_instr
         do j=1,num_image(i)
           ipoint=ip(i,j)
           totmodel=0.0
           maskmodel=0.0
!!!           yy0=0
!!!           do iy = iy0, iy1
           do iy = 1,nymask
!!!             yy0=yy0+1     !!! iyshift=iy-iy0+1 => iyshift=yy0
!!!             xx0=0
!!!             if (n == 1 .and. ipoint == 1) then
!               write(1,*) ff_save(:,yy0,n,ipoint)
!               write(2,*) (mask%mask(:,iy-iy0+1,ipoint)
!!!             endif
!!!             do ix = ix0, ix1
             do ix = 1, nxmask
!!!               xx0=xx0+1    !!! ixshift=ix-ix0+1 => ixshift=yy0
               totmodel=totmodel+ff_save(ix,iy,n,ipoint)
!!!               if (mask%mask(ix-ix0+1,iy-iy0+1,ipoint).eq.1)then
               if (mask%mask(ix,iy,ipoint).eq.1) &
                    maskmodel=maskmodel+ff_save(ix,iy,n,ipoint)
!                 write(*,*) "on mask:",ix,iy,ff_save(xx0,yy0,n,ipoint)
!!!               else
!                 write(*,*) "off mask:",ix,iy,ff_save(xx0,yy0,n,ipoint)  
!!!               endif 
             enddo
           enddo
           if (totmodel > 0.0) psffrac(n,ipoint)=maskmodel/totmodel
!           write(*,*) maskmodel,totmodel,psffrac(n,ipoint)
         enddo
       enddo
     enddo
!   stop  
     return
   endif


!!! IT/GL 2012/03 ff_save & modelima allocatable and smaller -> below
!    if (first) then
!!!! IT 2012/02: obsolete
!!!!      allocate(psf_save(nim,is,max_mask,max_mask))
!      allocate(ff_save(max_mask,max_mask,is,nim))
!      is_old=is
!    else if (is .ne. is_old) then
!!      deallocate(psf_save)
!!      allocate(psf_save(nim,is,max_mask,max_mask))
!      deallocate(ff_save)
!      allocate(ff_save(max_mask,max_mask,is,nim))
!      is_old=is
!    endif
 
    small=1.E-10


    if (first) then
!      psf_save(1:max_ima,1:max_fit,1:max_mask,1:max_mask) = 0.0
!      psf_save(1:nim,1:is,1:max_mask,1:max_mask) = 0.0
      scalepsf(1:max_ima,1:max_fit) = .false.
    else
!   check which parameters changed since last call: 
      do n=1,is
        if ((x0(n) .ne. x0_save(n)) .or. (y0(n) .ne. y0_save(n)) .or. (sig(n) .ne. sig_save(n))) then 
          scalepsf(1:max_ima,n) = .false.
          count_model=count_model+1
!          write(*,*) "parameters changed:", is,x0(n),x0_save(n),y0(n),y0_save(n),sig(n),sig_save(n)
        else
          do i=1,num_instr
            do j=1,num_image(i)
              ipoint=ip(i,j)
              if (scts(ipoint,n) .eq. 0.0) then
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .true.
              else if (scts(ipoint,n) .eq. scts_save(ipoint,n)) then
                scts_scale(ipoint,n) = 1.0
                scalepsf(ipoint,n) = .true.
              else if (abs(scts_save(ipoint,n)) .gt. small) then
                scts_scale(ipoint,n) =  scts(ipoint,n) / scts_save(ipoint,n)
!                write(*,*) ipoint,n,scts_scale(ipoint,n)
                scalepsf(ipoint,n) = .true.
              else
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .false.
                count_model=count_model+1
              endif
            enddo
          enddo
        endif
      enddo
    endif


!!!! switch off scaling for testing purposes:
!          scalepsf(1:max_ima,1:max_fit) = .false.
!!!!!!
    

    l=1
    lf=1
    yfit1(:)=0.
    yfit2(:)=0.
    yfit3=0.
    fudge=1.0

!!! IT/GL 2012/03 ff_save & modelima allocatable and smaller
!!! (via nxmask, i.e. via ix_min, ix_mask, i.e. via ecut)
   if (first) then
      is_old=is
      ! allocate new arrays
      xsize = nxmask
      ysize = nymask
      allocate(ff_save(xsize,ysize,is,nim), stat=astat)
      if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'ff_save'")
      ff_save(:,:,:,:) = 0.0
      allocate(modelima(xsize,ysize), stat=astat)
      if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'modelima'")
      modelima(:,:) = 0.0
   else
      if (is .ne. is_old  .or. &
           nxmask .ne. nxmask_save .or. nymask .ne. nymask_save) then
         ! allocate new arrays
         xsize = nxmask
         ysize = nymask
         if (allocated(ff_save)) deallocate(ff_save)
         allocate(ff_save(xsize,ysize,is,nim), stat=astat)
         if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'ff_save'")
         ff_save(:,:,:,:) = 0.0
         if (allocated(modelima)) deallocate(modelima)
         allocate(modelima(xsize,ysize), stat=astat)
         if (astat /= 0) call error("notEnoughMemory", "Not enough memory for array 'modelima'")
      modelima(:,:) = 0.0
      endif
      if (is .ne. is_old) is_old=is
   endif


    do i=1,num_instr

        s3=0.
        do j=1,num_image(i)

          ipoint=ip(i,j)

!obsolete.!          write(*,*) "nim:",nim
!obsolete.!          write(*,*) "num_instr,num_image(i),i,j,ipoint(i,j):",num_instr,num_image(i),i,j,ip(i,j)
!obsolete.          ipoint=ip(i,j)
!obsolete.!          ene=energy(ipoint)
!obsolete.!                    check whether same energy has been calculated for other instrument:
!obsolete.!                    CAL PSFs are identical for all EPIC inst. (as of 28/01/04),
!obsolete.!                    therefore it is pointless to call PSF again for same energy
!obsolete.!VVVVVVVVVVVVVVVVVV  remove, if separate PSF for EPIC instruments should be used
!obsolete.! removed, since fitting was slowed down (GL 03/03/04) 
!obsolete.!
!obsolete.!                    ifound=0
!obsolete.!                    do ii=1,i-1 
!obsolete.!                      do jj=1,num_image(ii)
!obsolete.!                         jpoint=ip(ii,jj)
!obsolete.!                         if (ene .eq. energy(jpoint)) then
!obsolete.!!                           write(*,*)  "found same energy:",ipoint,jpoint
!obsolete.!                           ifound=jpoint
!obsolete.!                           do n=1,is
!obsolete.!                             if  (.not. scalepsf(ipoint,n)) then
!obsolete.!                                
!obsolete.!!                               write(*,*)  "found same energy:",ipoint,jpoint
!obsolete.!                               scalepsf(ipoint,n) = .true.
!obsolete.!                               psf_save(ipoint,n,:,:)=psf_save(jpoint,n,:,:)
!obsolete.!                               scts_scale(ipoint,n) = 1.0
!obsolete.!                             endif
!obsolete.!                           end do
!obsolete.!                           exit
!obsolete.!                         endif
!obsolete.!                      end do
!obsolete.!                      if (ifound > 0) exit
!obsolete.!                   end do
!obsolete.!^^^^^^^^^^^^^^^^^^^
         
          s2=0.

          tempsum=0
          yy0=0




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!  new fft psf call:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!         if (.true.) then  

          modelima(:,:)=0.0

          do n=1,is

            ixs=nint(x0(n))
            iys=nint(y0(n))
            ix_mask=ixs-ix0
            iy_mask=iys-iy0
            ! source offset from pixel center         
!            xremain=x0(n)-float(ixs)
!            yremain=y0(n)-float(iys)
             xremain=x0(n)-dble(ixs)
             yremain=y0(n)-dble(iys)

            if (scalepsf(ipoint,n) .and. (scts_scale(ipoint,n) .ne. 0.0)) then
!              write(*,*) "scaling"

!!! IT/GL 2012/03 Copy only what is necessary ...
              xsizehalf=min(xsize/2,isize/2)
              ysizehalf=min(ysize/2,isize/2)
              ixm1=max(1,ix_mask-xsizehalf+1)
              iym1=max(1,iy_mask-ysizehalf+1)
              ixm2=min(nxmask,ix_mask+xsizehalf)
              iym2=min(nymask,iy_mask+ysizehalf)
              if (scts_scale(ipoint,n) .ne. 1.0) then
!!!             ff_save(:,:,n,ipoint)=ff_save(:,:,n,ipoint)*scts_scale(ipoint,n)
                 ff_save(ixm1:ixm2,iym1:iym2,n,ipoint) = &
                      ff_save(ixm1:ixm2,iym1:iym2,n,ipoint) * &
                      real(scts_scale(ipoint,n))
              endif
!!!              modelima(:,:)=modelima(:,:)+ff_save(:,:,n,ipoint)
              modelima(ixm1:ixm2,iym1:iym2) = &
                   modelima(ixm1:ixm2,iym1:iym2) &
                   + ff_save(ixm1:ixm2,iym1:iym2,n,ipoint)

            else
               call fpsf(energy,ipoint,xremain,yremain,x0(n),y0(n),sig(n),ecut,isize,n)

!!! IT/GL 2012/03: New ranges

              ! copy fima into modelima:
              isizehalf=isize/2
!              ixf1=max(1,isizehalf-ix_mask+1)
!              iyf1=max(1,isizehalf-iy_mask+1)
!              ixf2=min(isize,isizehalf-ix_mask+nxmask)
!              iyf2=min(isize,isizehalf-iy_mask+nymask)
!              ixm1=max(1,ix_mask-isizehalf+1)
!              iym1=max(1,iy_mask-isizehalf+1)
!              ixm2=min(nxmask,ix_mask+isizehalf)
!              iym2=min(nymask,iy_mask+isizehalf)
              xsizehalf=min(xsize/2,isizehalf)
              ysizehalf=min(ysize/2,isizehalf)
              ixm1=max(1,ix_mask-xsizehalf+1)
              iym1=max(1,iy_mask-ysizehalf+1)
              ixm2=min(nxmask,ix_mask+xsizehalf)
              iym2=min(nymask,iy_mask+ysizehalf)
              ixf1=max(1,isizehalf-ix_mask+ixm1)
              iyf1=max(1,isizehalf-iy_mask+iym1)
              ixf2=min(isize,isizehalf+ixm2-ix_mask)
              iyf2=min(isize,isizehalf+iym2-iy_mask)

              ff_save(:,:,n,ipoint)=0.0
              ff_save(ixm1:ixm2,iym1:iym2,n,ipoint) = &
                   real(scts(ipoint,n))*fima(ixf1:ixf2,iyf1:iyf2)
!!!              modelima(:,:)=modelima(:,:)+ff_save(:,:,n,ipoint)
              modelima(ixm1:ixm2,iym1:iym2) = &
                   modelima(ixm1:ixm2,iym1:iym2) &
                   + ff_save(ixm1:ixm2,iym1:iym2,n,ipoint)

!             if (ipoint == 6) then
!               open(unit=1,file='modelim.dat',status='unknown')
!               write(1,*) modelima(:,:)
!               close(1)
!               stop
!             endif

! call system_clock(clock1,clock2,clock3)
!  write(*,*) "clock scaling:", clock1-clock0

            endif
          end do



! call system_clock(clock0,clock2,clock3)
          iyshift = iy0-iyb0(ipoint)  ! subimage coordinates
!!!          do iy = iy0, iy1            ! image coordinates
          do iy = 1,nymask            ! image coordinates / aperture
            ixshift = ix0-ixb0(ipoint)
!!!            do ix = ix0, ix1
            do ix = 1,nxmask
!!!              if (mask%mask(ix-ix0+1,iy-iy0+1,ipoint).eq.1) then
              if (mask%mask(ix,iy,ipoint).eq.1) then
                !!! coordinates should be within the subimage anyway:
                !!! otherwise mask%mask would be zero
!!!                ixx=min(max(ixshift,1),xbsize(ipoint))
!!!                iyy=min(max(iyshift,1),ybsize(ipoint))
                if (expfla) then
!!!                  s1=modelima(xx0,yy0)*exp_image(ixx,iyy,ipoint)
                   s1=modelima(ix,iy)*exp_image(ixshift,iyshift,ipoint)
                else
                  s1=modelima(ix,iy)
                endif
!!!                s1=s1+bg_image(ixx,iyy,ipoint)
                s1=s1+bg_image(ixshift,iyshift,ipoint)
                yfit1(lf)=s1
                s2=s2+s1
                lf=lf+1
              endif
              ixshift = ixshift+1
            end do   ! ix
            iyshift = iyshift+1
          end do   ! iy
          yfit2(ipoint)=s2
          s3=s3+s2



        end do        ! end loop over num_image      
        yfit3=yfit3+s3

    end do  !   end loop over num_instr


    first = .false.
    x0_save=x0
    y0_save=y0
    sig_save=sig
    scts_save=scts

!!! IT/GL 2012/03 needed for array allocation
    ecut_save = ecut
    nxmask_save = nxmask
    nymask_save = nymask


! done.
!------------------------------------------------------------------------------
!obsolete.! call system_clock(clock1,clock2,clock3)
!obsolete.!  write(*,*) "clock copying:", clock1-clock0
!obsolete.!       endif
!obsolete.
!obsolete.!      write(*,*) "yfit2:",yfit2(ipoint) 
!obsolete.         
!obsolete.!      open(unit=1,file='modelim.dat',status='unknown')
!obsolete.!      write(1,*) ff_save(:,:,1,1)
!obsolete.!      close(1)
!obsolete.!      stop
!obsolete.
!obsolete.
!obsolete.!       endif
!obsolete.
!obsolete.
!obsolete.
!obsolete.       !end fft psf
!obsolete.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!obsolete.
!obsolete.! call system_clock(clock1,clock2,clock3)
!obsolete.!  write(*,*) "clock scaling:", clock1-clock0
!obsolete.
!obsolete.
!obsolete.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!obsolete.!  conventional convolution PSF:
!obsolete.!  do not use this part 22/7/04
!obsolete.!      if (.false.) then
!obsolete.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!obsolete.!          do n=1,is
!obsolete.!            if (.not. (scalepsf(ipoint,n) .and. (scts_scale(ipoint,n) .ne. 0.0))) then 
!obsolete.!              count_psf=count_psf+1
!obsolete.!            endif
!obsolete.!          end do
!obsolete.!          do iy = iy0, iy1
!obsolete.!            yy0=yy0+1
!obsolete.!            xx0=0
!obsolete.!            do ix = ix0, ix1
!obsolete.!              xx0=xx0+1
!obsolete.!              if (mask%mask(ix-ix0+1,iy-iy0+1,ipoint).eq.1) then  
!obsolete.!                tempsum=tempsum+1  
!obsolete.!                x1=ix ! +0.5
!obsolete.!                y1=iy ! +0.5
!obsolete.!                s1=0.                  
!obsolete.!                do n=1,is
!obsolete.!                  if (scalepsf(ipoint,n) .and. (scts_scale(ipoint,n) .ne. 0.0)) then
!obsolete.!                    s1=s1+psf_save(ipoint,n,xx0,yy0)*scts_scale(ipoint,n)
!obsolete.!                    psf_save(ipoint,n,xx0,yy0)=psf_save(ipoint,n,xx0,yy0)*scts_scale(ipoint,n)
!obsolete.!                    else
!obsolete.!                   select case (point_resp(i))       
!obsolete.!                     case(1)
!obsolete.!                        stemp=scts(ipoint,n)*psf1(energy,ipoint, &
!obsolete.!                              ix,iy,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
!obsolete.!                             xcen(i),ycen(i),sig(n), ecut)
!obsolete.!                        s1=s1+stemp
!obsolete.!                        psf_save(ipoint,n,xx0,yy0)=stemp
!obsolete.!                     case(2)
!obsolete.!                        stemp=scts(ipoint,n)*psf2(energy,ipoint, &
!obsolete.!                              ix,iy,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
!obsolete.!                             xcen(i),ycen(i),sig(n), ecut)
!obsolete.!                        s1=s1+stemp
!obsolete.!                        psf_save(ipoint,n,xx0,yy0)=stemp
!obsolete.!                     case(3)
!obsolete.!                        stemp=scts(ipoint,n)*psf3(energy,ipoint, &
!obsolete.!                              ix,iy,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
!obsolete.!                             xcen(i),ycen(i),sig(n), ecut)
!obsolete.!                        s1=s1+stemp
!obsolete.!                        psf_save(ipoint,n,xx0,yy0)=stemp
!obsolete.!                     case(4)
!obsolete.!                        stemp=scts(ipoint,n)*psf4(energy,ipoint, &
!obsolete.!                              x1,y1,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
!obsolete.!                             xcen(i),ycen(i),sig(n), ecut)
!obsolete.!                        s1=s1+stemp
!obsolete.!                        psf_save(ipoint,n,xx0,yy0)=stemp
!obsolete.!                     case default
!obsolete.!                       call error("WrongPSF","Wrong PSF number") 
!obsolete.!                  end select  
!obsolete.! 
!obsolete.!                  endif
!obsolete.!                end do
!obsolete.!                if (expfla) then
!obsolete.!                  s1=s1*exp_image(ix,iy,ipoint)
!obsolete.!                end if
!obsolete.!                ixx=min(max(ix,1),npix(1))
!obsolete.!                iyy=min(max(iy,1),npix(2))       
!obsolete.!                s1=s1*fudge+bg_image(ixx,iyy,ipoint) 
!obsolete.!                yfit1(l)=s1
!obsolete.!                s2=s2+s1
!obsolete.!                l=l+1
!obsolete.!    
!obsolete.!              end if
!obsolete.!
!obsolete.!            end do
!obsolete.!          end do
!obsolete.!          yfit2(ipoint)=s2   
!obsolete.!          s3=s3+s2
!obsolete.!        endif
!obsolete.!     end of removed part 22/7/04
!obsolete.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!obsolete.!        write(*,*) "ipoint,yfit2(ipoint):",ipoint,yfit2(ipoint)        
!obsolete.
!obsolete.        end do
!obsolete.        ! end loop over num_image      
!obsolete.        yfit3=yfit3+s3
!obsolete.!        write(*,*) "lf:",lf
!obsolete.!        write(*,*) "yfit3:",yfit3
!obsolete.
!obsolete.
!obsolete.
!obsolete.!  call system_clock(clock1,clock2,clock3)
!obsolete.!  write(*,*) "clock model:", clock1-clock0
!obsolete.
!obsolete.
!obsolete.    end do  
!obsolete.!   end loop over num_instr
!obsolete.!   write(*,*) "model scts(1:nim,1) :", scts(1:nim,1)
!obsolete.!   write(*,*) "model scts(1:nim,2) :", scts(1:nim,2)
!obsolete.!   write(*,*) "model yfit2:", yfit2(1:nim)
!obsolete.
!obsolete.    ! end loop over num_instr
!obsolete.!    write(*,*) "yfit1:",yfit1(900),yfit1(2036)
!obsolete.
!obsolete.(moved)    first = .false.
!obsolete.(moved)    x0_save=x0
!obsolete.(moved)    y0_save=y0
!obsolete.(moved)    sig_save=sig
!obsolete.(moved)    scts_save=scts
!obsolete.(moved)
!obsolete.(moved)!!! IT/GL 2012/03 needed for array allocation
!obsolete.(moved)    ecut_save = ecut
!obsolete.(moved)    nxmask_save = nxmask
!obsolete.(moved)    nymask_save = nymask
!obsolete.
!obsolete.!open(unit=1,file='yfit.dat',status='unknown')   
!obsolete.!write(1,*) yfit1
!obsolete.!close(1)
!obsolete.
!obsolete.!stop
!obsolete.
!obsolete.
!obsolete.!          write(*,*) "scts:",is,ipoint,scts(1,1),scts(1,2),scts(2,1),scts(2,2)
!obsolete.!          write(*,*) "sig:",sig(1),sig(2)
!obsolete.!          write(*,*) "x,y:",x0(1),x0(2),y0(1),y0(2)


  end subroutine model

!obsolete.!  function getEffArea(inst_id,energy,theta,phi)
!obsolete.!
!obsolete.!  real :: getEffArea
!obsolete.!  real(kind=double) :: energy,theta,phi
!obsolete.!  integer :: inst_id
!obsolete.!
!obsolete.!          if (inst_id == 1) then
!obsolete.!             call CAL_setState(instrument=EPN,AccuracyLevel=ACCURACY_MEDIUM)
!obsolete.!           endif
!obsolete.!          if (inst_id == 2) then
!obsolete.!            call CAL_setState(instrument=EMOS1,AccuracyLevel=ACCURACY_MEDIUM)
!obsolete.!          endif
!obsolete.!          if (inst_id == 3) then
!obsolete.!            call CAL_setState(instrument=EMOS2,AccuracyLevel=ACCURACY_MEDIUM)
!obsolete.!          endif  
!obsolete.! 
!obsolete.!          getEffArea=CAL_getEffectiveArea(energy,theta,phi)
!obsolete.!  
!obsolete.!  end function getEffArea

end module emlfit



  










