C ESA (C) 2000-2018 
C
C This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
C
C    SAS is free software: you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation, either version 3 of the License, or
C    (at your option) any later version.
C
C    SAS is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.
C
C    You should have received a copy of the GNU General Public License
C    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
      subroutine EQUGAL(DJD,LDIR,DLOLD,DBOLD,DLNEW,DBNEW)
C
CC    Converts coordinates between equatorial and new galactic system
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 17:02
C
CA  author : KOD        date: 19-JUL-1989 19:45
CU  update : KOD        date: 20-MAR-1990 14:56
CU  update : EXSAS              date: 03-AUG-1993 15:47
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert coordinates between the equatorial and the
CG  new galactic system (LII,BII).
CG  Parameters extracted from:
CG     F.C.Bertiau, 'Programmes for pocket calculators HP-67
CG                   and HP-97 in the field of theoretical and
CG                   observational astronomy',
CG                   1977 Leuven University Press, 'ASTR. COOR. 2'
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Day for the conversion
CP
CP  LDIR                L4  I   transformation direction:
CP                                LDIR = .true.  : equ --> gal II
CP                                LDIR = .false. : equ <-- gal II
CP
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  RCOMMON:CGENL.CMN          general parameter common block
CI  RCOMMON:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  EULER1              SR      general coordinate transformation
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX                              
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c     INCLUDE 'RCOMMON:CGENL.CMN'
c     INCLUDE 'RCOMMON:IGENL.INC'
C
      DOUBLE PRECISION
     &        DJD, DLOLD, DBOLD, DLNEW, DBNEW, D11, D21, D31,
     &              D12, D22, D32, D1, D2, D3, DJD50, DPHI1, DTHETA1,
     &              DPHI2, DTHETA2, DT, DPID2
ccc      REAL        HFLAG
ccc      INTEGER     IOS
      INTEGER     IERR, ISTAT
      LOGICAL     LDIR
c     CHARACTER*8   RTNAME
C
C     julian day for 1950.01.01 / 00:00:00
      PARAMETER ( DJD50 = 2433282.5D0 )
C
      PARAMETER ( DPID2 = 0.5D0*3.14159265358979323846D0)
      PARAMETER ( D11 = 4.92619181375D0 )
      PARAMETER ( D21 = 1.09257611174D0 )
      PARAMETER ( D31 = 5.70722665400D0 )
C
      PARAMETER ( D12 = 5.82362268000D-7 )
      PARAMETER ( D22 = 2.59857718339D-7 )
      PARAMETER ( D32 = 0.64223753852D-7 )
C
      SAVE
c     DATA          RTNAME /'EQUGAL'/
C
      IERR = 0
C
c      if (DLOLD.lt.0.D0 .or. DLOLD.GT.D2PI) then
c        IERR = 1
c        call WRERR(RTNAME,IERR,'Warning: unusual value for DLOLD')
c      endif
C
c      if (dabs(DBOLD).gt.DPID2) then
c        IERR = max(IERR,2)
c        call WRERR(RTNAME,IERR,'Error in input parameter DBOLD')
c      endif
C
c      if (IERR.lt.2) then
C
        DT = DJD - DJD50
C
c        ISTAT = 0
c        if (dabs(DT).gt.5.D4) ISTAT = 1
c        if (dabs(DT).gt.5.D6) ISTAT = 2
c        if (ISTAT.gt.1) then
c          IERR = max(IERR,ISTAT)
c          call WRERR(RTNAME,IERR,'Error in input parameter DJD')
c        endif
C
c      endif
C
c      if  (IERR.lt.2)  then
C
        D1 = D11 + D12 * DT
        D2 = D21 + D22 * DT
        D3 = D31 + D32 * DT
C
        DPHI1   = DLOLD
        DTHETA1 = DPID2 - DBOLD
C
        if (LDIR) then
          call EULER1(D1,D2,D3,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        else
          call EULER1(-D3,-D2,-D1,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        endif
C
c        if (ISTAT.gt.1) then
c          IERR = max(IERR,ISTAT)
c          call ERRSTC(RTNAME,10)
c        else
          DLNEW = DPHI2
          DBNEW = DPID2 - DTHETA2
c        endif
C
c      endif
C
C
      RETURN
      END
